(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

grtF`screenArgs[ argListIn_, reqCalced_, reqUncalced_] :=
 Module[ { a, actual, actual2, start, argList, realDef, item, itemH},
  (*
   * This is a good place to check that the metric exists!
   *)
  If[ grDebug, Print["screenArgs[",argListIn,",",reqCalced,",",reqUncalced." ]"] ];

  If[ grtG`metricSet === {},
	 Write[ stream, "Load a metric or basis first!"];
	 Abort[]
  ,(*else*)
	 If[ !MemberQ[ grtG`metricSet, grtG`metricName],
	   Write[ stream, "Metric ",grtG`metricName," has not been loaded."];
	   Abort[]
	 ]
  ];
  (*
   * recall of last object list
   *)
  If[ argListIn === {$},
      argList = grtG`lastArgList;
  ,(*else*)
      grtG`lastArgList = argListIn;
      argList = argListIn;
  ];
  (*
   * if the index names get assigned then BAD things will happen
   *)
  If[ MemberQ[ Map[ ValueQ, {up,dn,cup,cdn,pup,pdn,bup,bdn}], True],
	Write[ grt`stream,
	    "A GRTensor reserved function (e.g. up/dn) has been assigned!"];
	Abort[];
  ];
  (*
   * now build a list of the actual objects to be checked
   * So if user said CM or Ricci change to object names,
   *
   * PASS 1: expand compound objects
   *)

  actual = {};
  Do[
    (*
     * check that the object exists only in the grt` context.
     * If it exists in multiple contexts then we want the grt`
     * context one.
     *)
     item =  argList[[a]];
     If[ Head[item] =!= Symbol, 
         itemH = Head[item]
     ,(*else*)
         itemH = item
     ];
     If[ Context[ Evaluate[itemH] ] === Context[Global`anyVar],
       (* 
        * object exists in more than one context, force into grt` 
        * (this assumes it is in the Global context and so we don't
        * need to strip off its context)
        *)
       item = ToExpression[ ToString[ SequenceForm[ "grt`", item]]];
       If[ grDebug, Print["Converted to ", item] ];C
     ];
    (*
     * do a non-abortive check if defined to force creation
     * of objects (& hence their dependencies) as well
     * as ensuring autoLoading takes place
     *)
     If[ ListQ[ grtG`subList[ item ] ],
	(*
	 * MACRO - use entry in subList
	 *)
	actual = Join[ actual, grtG`subList[ item ] ];
      ,(*else *)
	actual = Join[ actual, {item} ];
     ];
   ,{a,Length[argList]}];

  actual2 = {};
  Do[
    (*
     * do an abortive check if defined to force creation
     * of objects (& hence their dependencies) as well
     * as ensuring autoLoading takes place
     *)
       realDef = grtF`findRealDef[ actual[[a]] ]; 
       grtF`checkIfDefined[ realDef ];
       actual2 = Join[ actual, { realDef } ]
  ,{a,Length[actual]}];

   If[ grDebug, Print["screenArgs: created actual = ",actual]];

  (*
   * if calc-ing then expand the list to include all the
   * dependencies
   *)
   If[ reqUncalced,
     actual = grtF`buildDependsList[actual2]
   ];
   If[ grDebug, Print["screenArgs: with deps actual = ",actual]];
  (*
   * now check that each of these objects is defined, if checking for
   * display, then remove from actual any which havn't been calc'd
   *)
  actual2 = {};
  Do[
     calced = grtF`checkIfAssigned[ actual[[a]] ];
     Which[ 
       reqCalced && !calced,
	  (* this object gets dropped from the list *)
	  Write[ grt`stream, actual[[a]], " has not been not calculated."],

       reqUncalced && calced,
	  (* do nothing - so this object gets droped from the list *)
	  Null,
       True,
           actual2 = Join[ actual2, {actual[[a]]} ]
     ]
  ,{a,Length[actual]} ];
  Return[actual2];
] (* end grtF`screenArgs *)



grtF`buildDependsList[ objectList_] :=
Module[ {a, depList, dups},

  grtG`dependsCache = {};
  (*
   * build a list of the dependencies of the objects in 
   * the list (plus their dependencies etc.)
   *)

  depList = Flatten[ Map[ grtF`expandDepends, objectList]];
 
  (*
   * now remove duplicates from the list, do this
   * by replacing duplicates with {} and then flatten
   * at the end
   *)
  
  Do[
    If[ depList[[a]] =!= {},
       dups = Position[ depList, depList[[a]] , Heads->False];
       If[ Length[dups] > 1, 
         depList = ReplacePart[ depList, {}, Delete[ dups, 1] ]
       ];
    ]
  ,{a,Length[depList]}];
   
  Return[ Flatten[ depList]];

];

grtF`expandDepends[ objectParm_] :=
Module[ {retList, object},

  object = grtF`findRealDef[ objectParm];
  If[ grDebug, Print["expandDepends: ",objectParm, "Real def: ", object] ];

  (*
   * this expansion is very slow due to redundancy
   * so cache values to speed it up
   *)
  If[ MemberQ[ grtG`dependsCache, object],
	Return[{}]
  ,(*else*)
	grtG`dependsCache = Union[ grtG`dependsCache, {object}];
  ];
  (*
   * If the object has been calculated, then we don't
   * need to worry about it
   *)
  If[ grtF`checkIfAssigned[ object], 
      Return[{}] 
  ];
  retList = {};
  (*
   * checkIfDefined to allow auto-creation of objects &
   * hence their dep lists. Allow it to abort if it
   * can't do this
   *)
   If[ !ListQ[ grtG`ObjDef[object][grtC`depends]],
      grtF`checkIfDefined[ object];
   ];

  If[ Length[ grtG`ObjDef[ object][grtC`depends] ] =!= 0,
    retList =  
	 Map[ grtF`expandDepends, grtG`ObjDef[object][grtC`depends] ];
  ];
  If[ grDebug, Print["expandDepends: done: ",Join[retList, {object}] ] ];
  Return[ Join[ retList, {object}]];

];



grtF`checkIfDefined[ object_, abort_:True] :=
 Module[ { objRoot, found, a, b,heads, ok, baseObject},
 (*
  * see if the object is defined. If not see if something
  * with the same number of indices is defined, if so
  * construct object by altering indices
  * Print["checking ",object,StringQ[object]]; 
  *)
 If[ grDebug, Print["checkIfDefined [",object,",", abort," ]"] ];

 ok = False;

 If[ !ListQ[ grtG`ObjDef[object][grtC`depends] ],

   (*
    * check the autoload lists, to see if we're supposed to load
    * this object
    *)
   If[ !ok && StringQ[ grtG`autoLoad[ object] ],
	 ok = True;
	 Message[grii::autoload, grtG`autoLoad[ object] ];
	 Needs[ grtG`autoLoad[ object] ];
   ];

   If[ !ok && grtF`findCousins[object] =!= {},
       (*
	* there is a cousin, so use it to create a definition
	* for the new index configuration
	*)
	ok = True;
	grtF`newIndices[ object ]
   ,(*else*)
     If[ Length[object] =!= 0,
       (*
	* may be able to create the object by diff-ing a base object
        * check this by stripping off one derivative index from
	* the end and seeing if that object can be defined (recursivly) 
	*)
	baseObject = object;
        a = Length[object];
        If[ MemberQ[{cup,cdn,pup,pdn}, object[[a]] ],
	  baseObject = Delete[ baseObject, a];
	  If[ Length[ baseObject] === 0 ,
	    baseObject = Head[baseObject] 
          ]; 
	  If[ ListQ[ grtG`ObjDef[ baseObject][grtC`depends] ] ||
	      grtF`checkIfDefined[ baseObject, False],
	    (*
	     * we can add a derivative index to this object
             * ( makeD does the pup/cup raise )
	     *)
	     grtF`makeD[ object];
             ok = True;
          ];
        ]; 
     ];
   ];
   If[ !ok && abort,
       Message[ grii::nodef, object];
       Abort[];
   ];
 ,(*else*)
   ok = True; 
 ];

 If[ grDebug, Print["checkIfDefined: Returned ", ok] ];
 Return[ok];

];



grtF`findCousins[ object_ ] :=

Module[ {cousins, a, b, index, ok, candidate},


  cousins = {};
  If[ Length[object] > 0,
   Do[
    If[ grtG`allObjects[[a]][[0]] === object[[0]] &&
	Length[ grtG`allObjects[[a]] ] == Length[ object],
	  (* head and number of indices match -> might be a cousin *)
          ok = True;
          candidate = grtG`allObjects[[a]];
          For[ b=1, b <= Length[object], b++,
             index = object[[b]];
             If[ MemberQ[{cup,cdn}, index] && 
                 !MemberQ[{cup,cdn}, candidate[[b]] ],
                ok = False;
             ];
             If[ MemberQ[{pup,pdn}, index] && 
                 !MemberQ[{pup,pdn}, candidate[[b]] ],
                ok = False;
             ];    
          ]; (* For *)
	  If[ ok,  cousins = Join[ cousins, { grtG`allObjects[[a]] }] ];
    ],
   { a, Length[ grtG`allObjects]}
   ];
 ];
 If[ grDebug, Print["Found cousins for ",object, cousins] ];

 Return[ cousins];

]; 



grtF`checkIfAssigned[ object_, metric_:grtG`metricName ] :=
Module[ {newObject},
  newObject = object;

  If[ ValueQ[ grtG`ObjDef[ object][grtC`displayName] ],
    newObject = grtG`ObjDef[ object][grtC`displayName]
  ];
  Return[ MemberQ[ grtG`record[metric], newObject] ];

];


grtF`findRealDef[ object_] := 
Module[ {b, rval, testObj}, 
    rval = object;
    If[ ListQ[ grtG`multipleDef[ object ] ],

         Do[
            testObj = grtG`multipleDef[object ][[b]];
           If[ grDebug, Print["findRealDef: checking ", testObj,
                      " useWhen=", 
                       grtG`ObjDef[testObj ][grtC`useWhen][]] ];
	   If[ (grtG`ObjDef[ testObj ][grtC`useWhen][]),
	      rval = testObj;
              Break[];
           ];
         ,{b,Length[ grtG`multipleDef[ object]]}];
     ];
     Return[ rval];
]; (* End *)


