(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

grdefine[ newObject_, sym_, defString_:Automatic] :=
 Module[ {a,b,c,
	defExpr,
	defListing,
	defDummies,
	junk,
	objectName,
	objectRoot,
	objectListing,
	shortRoot,
	tmp
	 }, 
  (*
   * check the sym parameter is valid
   *)
   If[ sym =!= {}, 
      If[ !ValueQ[ grtG`ObjDef[sym][grtC`symmetry] ], 
         Print["Cannot find symmetry of object ", sym];
         Abort[];
      ];
   ];

   If[ StringPosition[ newObject, "{"] =!= {}, 
       (* non-scalar *)
      junk = grtF`ToTensorExpr[newObject];
      objectName = ToExpression[ ToString[ SequenceForm[  
			grtG`defineDependsSet[[1]] ]]];
      shortRoot = ToExpression[ ToString[ SequenceForm[  
		 Head[objectName], Apply[SequenceForm, objectName] ]]];
      objectRoot = ToString[ SequenceForm[ "grtG`", 
		 Head[objectName], Apply[SequenceForm, objectName] ]];
      (* get the listing indices *)
      tmp = grtF`checkIndices[];
      objectListing = tmp[[1]];   
   ,(*else*)
      (*
       * scalar
       *)
	objectName = ToExpression[ ToString[ SequenceForm[ 
                        "grt`", newObject]]];
	shortRoot = ToExpression[ newObject]; 
	objectRoot = ToString[ SequenceForm[
			"grtG`", newObject ]];
        (* a scalar has no listing indices *)
        objectListing = {}; 
   ]; 
   (* get rid of grt` in objectRoot so we don't get a grtG`grt`fdndn e.g.*)
   objectRoot = StringReplace[ objectRoot, "grt`" -> ""];

   If[ grDebug, Print["define: objectRoot = ", objectRoot ] ];
   If[ grtF`checkIfDefined[ objectName, False], 
      Print["There is an existing definition for this object"];
      Abort[];
   ];
   If[ StringPosition[ newObject, "{"] =!= {},
        grtG`ObjDef[objectName][grtC`rootStr] = ToString[ Head[objectName]]; 
        grtG`ObjDef[objectName][grtC`indexList] = 
           ReplaceAll[ objectName, Head[objectName]->List];
   ,(*else*)
      (*
       *  scalar.
       *)
        grtG`ObjDef[objectName][grtC`rootStr] = ToString[ objectName]; 
        grtG`ObjDef[objectName][grtC`indexList] = {}; 
   ]; 

   
  If[ grDebug, Print["name=",objectName," root=",objectRoot] ];
  (*
   * now start to look at the definition.
   * (if there is one)
   *)
   If[ !SameQ[defString, Automatic],
      (*
       * HANDLE THE DEFINITION
       *)
      defExpr = grtF`ToTensorExpr[ defString];
      tmp = grtF`checkIndices[];
      defListing = tmp[[1]];
      defDummies = tmp[[2]];
      (*
       * check that listing indices match those in the newObject
       *)
      If[ Length[ objectListing] =!= Length[defListing], 
	Print["Number of listing indices in object and definition not equal."];
	Abort[];
      ];
      If[ Complement[ objectListing, defListing] =!= {}, 
	Print["Listing indices in object and definition differ."];
	Abort[];
      ];
      (*
       * check there are no summations nested inside functions
       *)
       grtF`SumInFunc[ defExpr, defDummies];
      (*
       * now generate the calc function for the expression
       *)
       grtG`ObjDef[objectName][grtC`depends] = 
	   grtG`defineDependsSet;
       grtG`ObjDef[objectName][grtC`calcFn] = 
	   grtF`buildCalcFn[ shortRoot, defListing, defDummies, defExpr];
   ,(*else*)
     (*
      * no definition provided, so the user must enter components
      *)
       grtG`ObjDef[objectName][grtC`depends] = {}; 
       grtG`ObjDef[objectName][grtC`calcFn] = grtF`enterComponents;


   ]; (* end if definition *)
   (*
    * now fill in all the mundane fields in the object definition
    *)
    grtG`ObjDef[objectName][grtC`header] = ToString[objectName]; 
    grtG`ObjDef[objectName][grtC`root] = ToExpression[ objectRoot];
   (*
    * symmetry assignment
    *)
    Which[ 
      sym =!= {},
         grtG`ObjDef[objectName][grtC`symmetry] = 
	    grtG`ObjDef[sym][grtC`symmetry],

      Length[ objectListing] === 0,
        grtG`ObjDef[objectName][grtC`symmetry] = grtF`symscalar, 
    
      Length[ objectListing] === 1,
        grtG`ObjDef[objectName][grtC`symmetry] = grtF`symvector, 
      
      True,
	grtG`ObjDef[objectName][grtC`symmetry] = 
	   ToExpression[ StringJoin[ "grtF`symnosym", 
			ToString[ Length[ objectListing ]]]]
    ];
    (* 
     * add this object to our internal book-keeping lists
     *)
    grtF`addObject[ objectName];

    Message[ grii::create,objectName];
    Return[Null];
 ]


 
grtF`buildCalcFn[ objectRoot_, listing_, dummy_, defExpr_] := 
Module[ {
	atoms,
	dummyCnt,
	maxDummy,
	localVars,
	loopTerms,
	loopGuts,
	expr,
	listingReplace,
	procString,
	replaceList,
	termString,
	a,b,c,
	},

 Do[ loopTerms[i] = 0 , {i,0,15}];

  Clear[grtG`a1, grtG`a2, grtG`a3, grtG`a4, grtG`a5, grtG`a6,
	grtG`a7, grtG`a8];
  Clear[ grtG`metricName];
 (*
  * build a replacement list for listing indices
  *)
  listingReplace = {};
  Do[ 
    listingReplace = Append[ listingReplace,
	   ToExpression[ ToString[SequenceForm[ listing[[a]],"->","grtG`a",a]]]];
  ,{a,Length[listing]}];


  maxDummy = 0;
  expr = Expand[defExpr];
  If[ Head[expr] =!= Plus,
     expr = {expr}; ];
  Do[
    atoms = Level[ expr[[a]], {-1}];
    replaceList = listingReplace;
    dummyCnt = 0;
    (*
     * find dummies which are atoms, As we find them
     * add to the replace list so we can substitite for them
     * later
     *)
    Do[
       If[ MemberQ[ atoms, dummy[[b]] ],
	 (*
	  * found a dummy
	  *)
          replaceList = Append[ replaceList,
            ToExpression[ ToString[
		 SequenceForm[ dummy[[b]],"->","s",dummyCnt+1]]]];
          dummyCnt++;
        ]; (* end if *)
    ,{b,Length[dummy]}];
    (*
     * now convert the indices and store at the appropriate level
     *)
    loopTerms[dummyCnt] += ReplaceAll[ expr[[a]], replaceList];
    maxDummy = Max[maxDummy, dummyCnt];
  ,{a,Length[expr]}];

 (*
  * so now we can build the guts of the procedure.
  * The release holds are to free up any explicit index functions
  * which were wrapped in Hold[]
  *)

  loopGuts = " ";
  localVars = "";
  For[ a = maxDummy, a > 0, a--,
     localVars = StringJoin[ localVars, ",s", ToString[a] ];
     If[ loopTerms[a] =!= 0,
	termString = StringJoin[ "s = s + ",
		     ToString[InputForm[loopTerms[a] ]],";" ];
     ,(*else*)
	termString = "";
     ];
     loopGuts = StringJoin["Do[ ", termString, loopGuts, 
			 ",{s" ,ToString[a],",grtG`Ndim[grtG`metricName]}] "]
  ]; (* end For *)

  If[ !SyntaxQ[loopGuts],
    Print["grtF`buildCalcFn bug!!: ", loopGuts]; Abort[]; ];

  (*
   * now tack on initialization stuff etc.
   *
   * High Wierdness: We need a "s;" before the Return[], otherwise
   * when there are loops we'll get Null returned. I have no idea
   * what's up with this.
   *)
   procString = ToString[
	SequenceForm[ "grtF`calc",objectRoot,"[object_, iList_] := ",
	    "Module[{ s",localVars,"}, ",
            "s = ReleaseHold[",ToString[InputForm[loopTerms[0]]],"];", 
	    loopGuts, "s;", "Return[s]; ]" ]];

  procString = StringReplace[ procString, ";,"->","];

  If[ !SyntaxQ[procString],
    Print["grtF`buildCalcFn bug!!"]; Abort[]; ];
  (*
   * Now declare the new calc function
   *)
  If[ grDebug, Print[procString] ];
  ToExpression[ procString] ;

  Return[ ToExpression[ ToString[ SequenceForm[
	    "grtF`calc", objectRoot ]]]];

]
 


grtF`checkIndices[] := 
Module[ {
	dummy,
	listing},

 (*
  * make use of the global grtG`defineIndices to
  * build a list of dummy indices and do some
  * error checking
  *)
  (* first some error checking *)
  If[ Intersection[ grtG`defineIndices[dn],
		       grtG`defineIndices[bup] ] =!= {},
       Print["Contraction of a tensor and basis index is not allowed."];
       Abort[];
  ];
  If[ Intersection[ grtG`defineIndices[bdn],
		       grtG`defineIndices[up] ] =!= {},
       Print["Contraction of a tensor and basis index is not allowed."];
       Abort[];
  ];
  dummy = Union[ Intersection[ grtG`defineIndices[dn],
		       grtG`defineIndices[up] ],
                 Intersection[ grtG`defineIndices[bdn],
		       grtG`defineIndices[bup] ] ];
  listing = Complement[ Union[ grtG`defineIndices[up],
	                       grtG`defineIndices[dn],
	                       grtG`defineIndices[bdn],
	                       grtG`defineIndices[bup] ],
                        dummy];

  Return[{listing,dummy}];

] (* end dummyIndices *)




grtF`enterComponents[ object_, indices_] :=
Module[ {},
  Return[ grtF`myInput[ StringJoin["Enter value for ",
	   ToString[grtF`dispStr[ object, indices]]," "]] ];

];

(*
 * Utility routine
 *)

 dumpDef[ object_] := 
 Module[ {entries,a},

 entries = {grtC`header, grtC`root, grtC`rootStr, grtC`indexList,
	    grtC`preCalcFn, grtC`calcFn, grtC`calcFnParms, 
	    grtC`symmetry, grtC`depends};
  Do[
    If[ ValueQ[ grtG`ObjDef[ object][ entries[[a]] ] ],
      Print[entries[[a]]," = ",
         grtG`ObjDef[ object][ entries[[a]] ] ];
    ]; (* end if *) 
 ,{a,Length[entries]}];

 ]




grtF`explicitIndex[ index_] := Module[ {a,val, metric},

  metric = grtG`defaultMetric;
  val = 0;
  Do[
    If[ grtG`xup[metric, a] === index,
      val = a
    ];
  ,{a,grtG`Ndim[metric]}];

  If[ val === 0, 
    Print["Explicit coordinate ", index, " does not exist in metric ",
	metric];
    Abort[];
  ];

  Return[val];
];
(*
 * Kronecker Delta definitions
 *)
grtG`kdeltaupdn[ gname_, i_Integer, i_Integer] := 1;
grtG`kdeltaupdn[ gname_, i_Integer, j_Integer] := 0;
grtG`kdeltadnup[ gname_, i_Integer, i_Integer] := 1;
grtG`kdeltadnup[ gname_, i_Integer, j_Integer] := 0;



grtF`SumInFunc[ x_, dummies_] := 
Module[{i,t,bodies,headies},
  t = Table[ Level[ x,{i}],{i,Depth[x]}];
  bodies = Join[ {x}, Flatten[ t] ];
  headies = Flatten[Map[ Head, bodies]];
(* Print[" b = ", bodies, " h = ", headies]; *)
  Table[ grtF`BadFunction[ headies[[i]], bodies[[i]], dummies],
     {i,Length[t]}];
];

(*
 * allow summations inside Times, Plus and Symbol
 *)
grtF`BadFunction[ Times, a_, d_] := Null;
grtF`BadFunction[ Plus, a_, d_] := Null;
grtF`BadFunction[ Symbol, a_, d_] := Null;

grtF`BadFunction[ a_, b_, dummies_ ] :=
  If[ MemberQ[ Attributes[a], Protected] &&
      Intersection[ dummies, Flatten[Level[b,{-1}]]] =!= {},
	 (*
	  * sum in a function. Cannot create def'n
	  *)
	  Print["Summation in function ", a," encountered."];
	  Print["Please define the arguments of the function as"];
	  Print["a seperate object first."];
	  Abort[];
   ];



