(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

grtF`ToTensorExpr[ sexpr_] :=

 Module[ {a,b,
	  openItal, closeItal, 
	  openRound, closeRound,
	  openSquare, closeSquare,
	  delList,
	  delSet,
	  expr,
	  exprChars,
	  inItal,
	  newEntry,
	  rootName,
	  indexChars}, (* do NOT declare tensorArray *)
  (*
   * First, get lists of the start of "blatent" tensors
   * i.e. those with {}'s
   * (we'll deal with scalars a bit later)
   *)

   openItal = Union[ Flatten[ StringPosition[ sexpr,"{"]]];
   closeItal = Union[ Flatten[ StringPosition[ sexpr,"}"]]];
  
   If[ Length[openItal] =!= Length[closeItal],
     Print["Unmatched {}'s. Please check the expression entered"];
     Abort[];
   ];

   (*
    * check overall bracket balancing first - reassign these later
    *)
   openRound = Union[ Flatten[ StringPosition[ sexpr,"("]]];
   closeRound = Union[ Flatten[ StringPosition[ sexpr,")"]]];
   If[ Length[openRound] =!= Length[closeRound],
     Print["Unmatched ()'s. Please check the expression entered"];
     Abort[];
   ];

   openSquare = Union[ Flatten[ StringPosition[ sexpr,"["]]];
   closeSquare = Union[ Flatten[ StringPosition[ sexpr,"]"]]];
   If[ Length[openSquare] =!= Length[closeSquare],
     Print["Unmatched []'s. Please check the expression entered"];
     Abort[];
   ];
   (*
    * reset the globals
    *)
    grtG`defineDependsSet = {};
    grtG`defineIndices[up] = {};
    grtG`defineIndices[dn] = {};
    grtG`defineIndices[bup] = {};
    grtG`defineIndices[bdn] = {};
   (*
    * Compress alpha stuff into a single entry
    * e.g. a,b, -> ab,
    *)
   exprChars = grtF`compressAlpha[ Characters[sexpr]];
  
  (*
   * SCALAR DETECTION
   *)
   (* 
    * tack on a whitespace at the end so the If below doesn't
    * come to grief
    *)
   exprChars = Append[ exprChars, " "];
   inItal = False;
   Do[ (* for each char in the string *)
    If[ exprChars[[b]] === "{", inItal = True];
    If[ exprChars[[b]] === "}", inItal = False];
    If[ grDebug, Print["Testing for scalar: ",exprChars[[b]] ] ];
    If[ !inItal && grtF`LetterDigitQ[ exprChars[[b]] ] && exprChars[[b+1]] =!= "{" ,
    (*
     * Any text not followed by a "{" is potentially a
     * GRT scalar (but may be a benign constant)
     * To distinguish these we check to see if a scalar with
     * the name found has been defined. If so add {} to it.
     *)
       If[ grtF`checkIfDefined[ ToExpression[ exprChars[[b]] ], False ],
           exprChars = ReplacePart[exprChars, 
		 StringJoin[ exprChars[[b]],"{}"] , b];
       ];
     ]; (* end If *)
   ,{b,Length[exprChars]}];
   exprChars = grtF`compressAlpha[ Characters[ StringJoin[ exprChars ]]];  

   openItal = Flatten[ Position[ exprChars,"{"]];
   closeItal = Flatten[ Position[ exprChars,"}"]];
 
  (*
   * IDENTIFY TENSORS
   *
   * Build lists of indices and their types. Keep them indexed
   * by where the tensor open bracket is located in the string
   *)

   Do[ (* for each set of {}'s *)
      indexChars = Part[ exprChars, Range[openItal[[a]],closeItal[[a]] ]];
      tensorArray[a] = grtF`makeIndexList[indexChars];
      (* store the tensor name for latter *)
      tensorName[a] = exprChars[[ openItal[[a]]-1 ]];
      (* 
       * add a grtG` prefix to the tensor name in the exprChars
       * entry (since the internals are in this private context
       *)
      exprChars = ReplacePart[ exprChars, 
		  StringJoin[ "grtG`", exprChars[[ openItal[[a]] -1 ]] ],
		  openItal[[a]]-1 ]; 
  ,{a,Length[openItal]}]; (* end for each {}'s *)
   
  (*
   * Now process symmetries.
   * Two passes {symmetric & antisymmetric}
   *)

  exprChars = grtF`insertSymCalls[ exprChars, openItal, closeItal, 
			SymStart,SymStop]; 
  exprChars = grtF`insertSymCalls[ exprChars, openItal, closeItal, 
			AsymStart,AsymStop];
  
  If[ grDebug, Print["ToTensorExpr: post symm: ",exprChars] ];

  (* 
   * change the name of the object to match internal naming 
   * conventions.
   *
   * Delete SymStart's etc. first 
   *)
   delSet = {BachBar, SymStart, SymStop, AsymStart, AsymStop};
   Do[
   delList = {};
     Do[
       If [ MemberQ[ delSet, tensorArray[a][[1]][[b]] ],
         delList = Append[ delList, b];
       ];
     ,{b,Length[ tensorArray[a][[1]] ]}];
     If[ delList =!= {},
       tensorArray[a] = { Delete[ tensorArray[a][[1]], Partition[delList,1] ],
                        Delete[ tensorArray[a][[2]], Partition[delList,1] ]};
     ];
   ,{a,Length[openItal]}];

   delList = {};
   Do[

      If[ tensorName[a] =!= "kdelta",
         If[ tensorArray[a][[2]] === {}, 
           (* scalar *)
             grtG`defineDependsSet = Union[ grtG`defineDependsSet,
	        {ToExpression[ToString[SequenceForm[
		 "grt`", tensorName[a] ]]]
		 //. List -> Sequence} ];
         ,(*else*)
           grtG`defineDependsSet = Union[ grtG`defineDependsSet,
	      {ToExpression[ToString[SequenceForm[
		 "grt`", tensorName[a][ tensorArray[a][[2]] ] ]]]
		 //. List -> Sequence} ]; 
         ];
      ];
      (* 
       * form the root name e.g. T & {dn,up,dn} -> "Tdnupdn" 
       * (note that char-1 may have a Symmetrize prefix)
       *)
      rootName = exprChars[[ openItal[[a]]-1 ]];
      Do[
	 rootName = SequenceForm[ rootName, tensorArray[a][[2]][[b]] ]
      ,{b,Length[ tensorArray[a][[2]] ]}];
      (*
       * ADD INDICES
       *
       * now modify the string to form the internal representation
       *)
       newEntry = ToString[SequenceForm[rootName,
		Join[{"grtG`metricName"}, tensorArray[a][[1]] ] ]];
       (*
	* angle brackets used to keep sym indices safe
	*)
       newEntry = StringReplace[ newEntry , 
		   {"{" -> "[", "}"->"]","<"->"{",">"->"}" }];
       (*
	* replace the head with newEntry and then delete
	* everything between the {}'s, taking care to
	* preserve any []'s at the end to close symmetry calls
	*)
       exprChars = ReplacePart[ exprChars, newEntry, openItal[[a]]-1];
       delList = Join[ delList, Range[openItal[[a]], closeItal[[a]]-1 ] ];
       (*
	* the closing } may have had square brackets added to it to close off
	* symmetry calls. We want to remove only the "}" part.
	*)
       exprChars = ReplacePart[ exprChars, 
		     StringReplace[ exprChars[[ closeItal[[a]] ]], "}"->""],
		      closeItal[[a]] ];
   ,{a,Length[openItal]}];
 (* tidy up and make back into a string *)
 exprChars = Delete[ exprChars, Partition[delList,1] ];
 expr = StringJoin[ exprChars];
 (*
  * Test that ToExpression will succeed
  *)
  If[ !SyntaxQ[ expr],
    Print["ToTensorExpr: Mathematica cannot parse the resulting expression:"];
    Print[expr];
    Print["Please check the expression."];
    Abort[];
  ];
  If [grDebug, Print["ToTensorExpr: ", expr]];
 
 Return[ ToExpression[expr] ];

 ] (* end module *)

(*=====================================================
 * grtF`compressAlpha[ exprChar]
 *====================================================*)

grtF`LetterDigitQ[ str_] :=
Module[ {chars, bool, a},
  (*
   * utility to determine if a string is composed of only
   * letters and numbers
   *)
  chars = Characters[ str];
  bool = True;
  Do[ bool = bool && (LetterQ[ chars[[a]]] || DigitQ[ chars[[a]] ]),
   {a,Length[chars]}];

  Return[bool];

];

grtF`compressAlpha[ exprIn_] :=
 Module[ {exprChars, delList, b},

   (*
    * Compress alpha stuff into a single entry
    * e.g. a,b,) -> ab,)
    *)
   delList = {};
   exprChars = exprIn;
   Do[ (* for each char in the string *)
    If[ grtF`LetterDigitQ[ exprChars[[b]] ] && 
	grtF`LetterDigitQ[exprChars[[b+1]] ],
    (*
     * adjacent letters, so merge them
     * (keep a list of slots to delete at the end)
     *)
       exprChars = ReplacePart[exprChars, 
		 ToString[ SequenceForm[exprChars[[b]],exprChars[[b+1]] ]] ,
		 b+1];
       delList =  Append[delList, {b}];
     ]; (* end If *)
   ,{b,Length[exprChars]-1}];
   exprChars = Delete[ exprChars, delList];
   Return[exprChars];
]

 
 grtF`insertSymCalls[ inChars_, openItal_, closeItal_, start_, stop_] := 
 Module[ {a,
	symList,
	symString,
	objHead,
	objTail,
	exprChars},

  exprChars = inChars; (* copy so we can alter it *)
  symList = grtF`findSymIndices[ openItal, start, stop];
  If[ start === SymStart,
    symString = "Symmetrize["
    ,(*else*)
    symString = "AntiSymmetrize[" 
  ];
  Do[
    (*
     * for each entry in the SymList we need to add a
     * "Symmetrize[ {indices}," to the head
     * and a "]" to the tail. 
     *)
     objHead = openItal[[ symList[[a]][[1]] ]]-1;
     objTail = closeItal[[ symList[[a]][[2]] ]];
     (*
      * Replace list {}'s in Symmetrize with <>'s so that
      * {}'s continue to be the unique delimiters of tensors
      *) 
     exprChars = ReplacePart[ exprChars, 
		   StringReplace[
		     ToString[SequenceForm[symString, symList[[a]][[3]],",",
		     exprChars[[objHead]] ] ], {"{"->"<","}"->">"}],
		   objHead];
     exprChars = ReplacePart[ exprChars, exprChars[[objTail]]<>"]", objTail];
  ,{a,Length[symList]}];    

 Return[exprChars];

] (* end grtF`insertSymCalls *)



grtF`findSymIndices[ openItal_, start_, stop_] := 
Module[ {a,b,c,
	found,
	inBar,
	lastStop,
	startInteger,
	stopInteger,
	startType,
	symIndices,
	symNameList, 
	symTypeList,
	symStartList,
	symStopList,
	returnList},
  (*
   * begin by building a flat list with index types and names
   * (e.g. for "T{a (b} * xi{c)}" -> {1, a, SymStart, b, 2, c, SymStop} )
   *)
   returnList = {};
   symNameList = {};
   symTypeList = {};
   (*
    * form a long list of indices, with indices from tensors
    * seperated by integers indicating their entry in tensorArray[]
    *)
   Do[ symNameList = Join[symNameList,{a}, tensorArray[a][[1]] ];
       symTypeList = Join[symTypeList,{a}, tensorArray[a][[2]] ]
   ,{a,Length[openItal]}];
   (*
    * for the purposes of symmetry cdn=dn etc.
    *)
   symTypeList = ReplaceAll[ symTypeList , 
		{"cdn"->"dn","pdn"->"dn","cup"->"up","pup"->"up"}];
   (*
    * tag the positions of the start and stop symbols and put
    * them in lists
    *)
   symStartList = Flatten[ Position[ symNameList, start]];
   symStopList = Flatten[ Position[ symNameList, stop]];
  
  lastStop = 0;
  Do[
    (*
     * for each start, find the corresponding stop
     *)
    startType = symTypeList[[ symStartList[[ a]]+1 ]];
    b = lastStop+1;
    found = False;
    While[ b <= Length[symStopList] && !found, 
       (*
	* look through the stop list and find the nearest stop
	* which is preceeded by an index of the same type 
	* as the SymStart
	*)
       If[ startType =!= symTypeList[[ symStopList[[b]] - 1]],
	   b++; 
	,(*else*)
	 (*
	  * found the matching stop. Accumulate the indices
	  * of the same type and watch for bach bars
	  *)
	  found = True;
	  lastStop = b;
	  symIndices = {};
	  inBar = False;
	  For[ c = symStartList[[a]], c < symStopList[[b]], c++,
	    If[ symTypeList[[c]] === BachBar ,
	       If[ symTypeList[[c+1]] === startType && !inBar,
		  inBar = True
 	       ];
	       If[ inBar && symTypeList[[c-1]] === startType,
		  inBar = False
 	       ];
            ];
	    If[ startType === symTypeList[[c]] && !inBar,
	      symIndices = Append[ symIndices, symNameList[[c]] ]
	    ]; (* end if *)
          ]; (* end for *)
        ]; (* end If *)
    ]; (* end While *)
    (*
     * now find the integer references to the head and tail
     * and build the sublist to return
     *)
     If[ b > Length[symStopList], 
       Print["Unbalanced symmetry brackets."];
       Abort[];
     ];
     startInteger = symStartList[[a]];
     While[ !IntegerQ[ symNameList[[ startInteger]] ], startInteger-- ];
     stopInteger = symStopList[[b]];
     While[ !IntegerQ[ symNameList[[ stopInteger]] ], stopInteger-- ];
     returnList = Append[ returnList, 
			  {symNameList[[startInteger]],
			    symNameList[[stopInteger]], 
			    symIndices}]; 
  
  ,{a, Length[symStartList]}];

 Return[returnList];

]




 grtF`makeIndexList[ indexParm_] :=
 Module[ {b, 
	  upFlag, explicitFlag, CoDFlag, pDFlag, tetradFlag,
	  index, indexList, typeList,
	  delList,
	  indexChars},
(* Print[ "grtF`makeIndexList ",indexParm]; *)      
      indexChars = indexParm; (* copy across so we can modify it *)
       (* set some flags *)
       CoDFlag = False;
       pDFlag = False;
       tetradFlag = False;
       explicitFlag = False;
       upFlag = False;
       indexList = {};
       typeList = {};
       Do[ (* for each entry in indexChars *)
	 If[ LetterQ[ indexChars[[b]] ], (* then *)
	    (*
	     * found an index name. Determine it's type and add
	     * it the list  
	     *)
	     If[ upFlag, index = "up", index = "dn"];
	     (* is it a tetrad ?? *)
	     If[ tetradFlag, 
	       (*
		* check there is a trailing ")"
		* If not then the ( is part of a symmetrization
		*)
		If[ indexChars[[b+1]] =!= ")" ,
	           indexList = Append[ indexList, SymStart];
		   typeList = Append[ typeList, SymStart];
		   tetradFlag = False
		, (* else *)
		   index = "b"<>index
                ];
             ];
	     (* all upper case -> Spinor *)
             If[ UpperCaseQ[ indexChars[[b]] ], index = "s"<>index ];
             If[ CoDFlag, index = "c"<>index ];
	     If[ pDFlag, index = "p"<>index ];
	     (* add the index to the list *)
	     typeList = Append[ typeList, index];
	     (*
	      * now add the index to the global index lists
	      * These are used to identify listing and dummy indices
	      * and to ensure that dummies come in up/dn pairs.
	      *
	      * We regard cup and pup as being simply up
	      * Maintain Sets indexed by up,dn,bup,bdn,sup,sdn
	      *)
	      If[ !explicitFlag,
	         indexList = Append[ indexList, indexChars[[b]] ];
                 index = Replace[ ToExpression[index], 
			 {cup->up, pup->up, pdn->dn, cdn->dn}];
                 If[ grDebug, Print["  found index: ", index] ];
	         grtG`defineIndices[index] = Union[ grtG`defineIndices[index],
			       {ToExpression[ indexChars[[b]] ]}];
              ,(*else*)
		(*
		 * Explicit index. We need to wrap this with a
		 * function which converts the coordinate name to
		 * the appropriate number at run time.
		 *
		 * We also need to hold this function
		 *)
	         indexList = Append[ indexList, 
			StringJoin[ "Hold[ grtF`explicitIndex[ ",
				indexChars[[b]], "]]" ]];
              ];
	     (* reset *some* flags [note CoD and partial D's don't get reset!] *)
             explicitFlag = False;
             upFlag = False;
	  , (*else*)
	    (* some kind of special char *)
	    Switch[ indexChars[[b]],
	      ";", CoDFlag = True,
	      ",", pDFlag = True,
	      "(", tetradFlag = True,
	      ")", If[ tetradFlag, tetradFlag = False
		   ,(*else*)
		     (*
		      * indicates the end of a sym. 
		      *)
                      indexList = Append[ indexList, SymStop];
		      typeList = Append[ typeList, SymStop];
		   ],
              "[", (* for now ignore operators, and assume it's asym *)
		   indexList = Append[ indexList, AsymStart];
		   typeList = Append[ typeList, AsymStart],
	      "]", indexList = Append[ indexList, AsymStop];
		   typeList = Append[ typeList, AsymStop],
	      "|", indexList = Append[ indexList, BachBar];
		   typeList = Append[ typeList, BachBar],
	      "$", explicitFlag = True,
	      "^", upFlag = True

            ];
          ];

       ,{b,Length[indexChars]}];
(*  Print[{indexList,typeList}]; *)
  Return[{indexList,typeList}];

 ] (* end module *)


 (*===================================
 * SYMMETRIZATION ROUTINES
 *===================================*)

 Symmetrize[ indices_, expr_] :=
 Module[ {a,b, s,
	dummy,
	repList,
	newExpr,
	newTerm,
	perms, 
	signs},

  s = 0;
  perms = Permutations[ indices];
  signs = Map[ Signature, perms];

  newExpr = expr;
  Do[ (* for each element in a *)
       newExpr = ReplaceAll[ newExpr, indices[[b]] -> dummy[ indices[[b]] ] ];
  ,{b,Length[indices]}];
  Do[
      newTerm = newExpr;
      Do[ (* for each element in a *)
         newTerm = ReplaceAll[ newTerm, 
		      dummy[ indices[[b]] ] -> perms[[a]][[b]] ];
      ,{b,Length[indices]}];
      s = s + newTerm;
  ,{a,Length[signs]}];

  Return[s/Factorial[Length[indices]]];

 ]

 AntiSymmetrize[ indices_, expr_] :=
 Module[ {a, b, s,
	newExpr,
	newTerm,
	indexSig,
	perms, 
	signs},

  indexSig = Signature[indices];
  s = 0;
  perms = Permutations[ indices];
  signs = Map[ Signature, perms];

  newExpr = expr;
  Do[ (* for each element in a *)
       newExpr = ReplaceAll[ newExpr, indices[[b]] -> dummy[ indices[[b]] ] ];
  ,{b,Length[indices]}];
  Do[
      newTerm = newExpr;
      Do[ (* for each element in a *)
         newTerm = ReplaceAll[ newTerm, 
		      dummy[ indices[[b]] ] -> perms[[a]][[b]] ];
      ,{b,Length[indices]}];
      s += indexSig * signs[[a]] * newTerm; 
  ,{a,Length[signs]}];

  Return[s/Factorial[Length[indices]]];

] (* end Module *)


