(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

grt`grload[gnameIn_,gfile_] :=
  Module[ {i, j, tempDim,
	metricName,
	coordName,
        gname,
        basisupFlag,
	metricStyle},
	(* check the name against existing ones *)
	If[ MemberQ[grtG`metricSet, gnameIn] ,
	   gname = Unique[ToString[gnameIn]];
	   Message[grii::overload, gnameIn, gname];
        ,(*else*)
	   gname = gnameIn;
	];
	grtG`metricName = gname;
	grtG`record[gname] = {};
	If[ Get[gfile] === $Failed,
	   Write[ grt`stream,"Failed to read ",gfile];
	   Abort[];
	];

	(*
	 * copy metric elements into internal array
	 * and then unassign ones read in.
	 *
	 * Note: Ndim is in context grt`
	 *)
	metricStyle = 1; (* metric instead of basis by default *)
       Which[ 

	IntegerQ[ makeg`Ndim],
	     (*
	      * new style names
	      *)
	      grtG`Ndim[gname] = makeg`Ndim;
	      Clear[makeg`Ndim];
	      coordName = makeg`x;
	      metricName = makeg`metric;
	      metricStyle = 2, (* basis *)
          True,
	      Print["Bad metric file (Ndim not defined)."]; Abort[]
        ];
	(*
	 * several styles of input files. Handle based on
	 * makeg`style ( 1= metric, 2=basis, 3=np )
	 *)
	 Which[
	   makeg`style === 1, (* metric *)
	      grtG`record[gname] = Union[ grtG`record[gname], 
		     {g[dn,dn], x[up]}]; 
	      grtF`loadMetric[ metricName, coordName, gname], 

           makeg`style === 2 || makeg`style === 3, (* basis/NP *)
	      grtG`record[gname] = Union[ grtG`record[gname], 
		     {eta[bup,bup], x[up]}]; 
	      grtF`loadEta[ makeg`eta, coordName, gname];
              basisupFlag = False;
              Do[ 
                 Do[
                   (* test for an element in the
                    * array 
                    *)
	           If[ Head[ makeg`basisu[i,j] ] =!= makeg`basisu,
                       basisupFlag = True
		   ],	
                 {j,grtG`Ndim[gname]}], 
              {i,grtG`Ndim[gname]}];

              If[ basisupFlag, 
	         grtG`record[gname] = Union[ grtG`record[gname], 
		     {e[bdn,up]}]; 
                 grtF`loadBasis[ makeg`basisu, "up", gname, makeg`style];
              ,(*else*)
	         grtG`record[gname] = Union[ grtG`record[gname], 
		     {e[bdn,dn]}]; 
                 grtF`loadBasis[ makeg`basisd, "dn", gname, makeg`style];
              ];

        ];
	If[ ListQ[ makeg`constraint],
	   grtG`constraint[ gname] = makeg`constraint;
	   Clear[ makeg`constraint];
	   grtG`record[ gname] = Union[ grtG`record[gname], {constraint}];
        ];
	If[ StringQ[ makeg`info],
	   grtG`Info[ gname] = makeg`info;
	   Clear[ makeg`info];
	   grtG`record[ gname] = Union[ grtG`record[gname], {Info}];
        ];
	grtF`initMetric[gname, makeg`style];
        Clear[ makeg`Ndim, makeg`style, makeg`eta, makeg`metric,
		makeg`basisu, makeg`basisd]; 

] (* end grload *)

(**************************************
* grtF`loadBasis
*
* copy the metric from the named array into GRT internals
* (also used in the load option from makeg)
***************************************)

grtF`loadBasis[ basis_ , type_, gname_, style_] := 
Module[ {i,j, root, vname},
        (*
	 * build the root based on vector type (up/dn)
	 *)
        Which[ 
	  type === "up" && style === 3,
	     vname = {grtG`NPlup, grtG`NPnup, grtG`NPmup, grtG`NPmbarup},
          type === "up" && style === 2,
	     vname = {grtG`e1up, grtG`e2up, grtG`e3up, grtG`e4up},
	  type === "dn" && style === 3,
	     vname = {grtG`NPldn, grtG`NPndn, grtG`NPmdn, grtG`NPmbardn},
          type === "dn" && style === 2,
	     vname = {grtG`e1dn, grtG`e2dn, grtG`e3dn, grtG`e4dn}
        ];
	root = ToExpression[ ToString[ SequenceForm[ "grtG`ebdn", type]]];
	Do[
	  Do[
	     (* 
	      * assign the vector, but reference to the basis
	      * entry, so that when that changes, the vector does
	      *)
	      If[ Head[ basis[i,j]] =!= basis, 
	        vname[[i]][ gname, j] = basis[i,j];
		root[gname,i,j] = basis[i,j];
              ,(*else*)
	        vname[[i]][ gname, j] = 0;
		root[gname,i,j] = 0;
              ];
	  ,{ j , grtG`Ndim[gname]} ],
	{ i, grtG`Ndim[gname] } ];
	Return[];
]

(**************************************
* grtF`loadMetric
*
* copy the metric from the named array into GRT internals
* (also used in the load option from makeg)
***************************************)

grtF`loadMetric[ metricName_ , coordName_, gname_] := 
Module[ {i,j},
	(*
	 * If it's a metric then 
	 *)
	Do[
	  grtG`xup[gname,i] = coordName[i];
	  Do[
	     If[ Head[ metricName[i,j]] =!= metricName ,
		(* assign like this so only need to simplify [i,j] *)
		grtG`gdndn[gname,j,i] = grtG`gdndn[gname,i,j];
		grtG`gdndn[gname,i,j] = metricName[i,j];
		Clear
	     ,(*else*)
		grtG`gdndn[gname,j,i] = 0;
		grtG`gdndn[gname,i,j] = 0;
             ];
	  ,{ j , i, grtG`Ndim[gname], 1} ],
	{ i, grtG`Ndim[gname] } ];
 ]

(**************************************
* grtF`loadEta
*
* copy eta from the named array into GRT internals
* (also used in the load option from makeg)
***************************************)

grtF`loadEta[ metricName_ , coordName_, gname_] := 
Module[ {i,j},


	Do[
	  grtG`xup[gname,i] = coordName[i];
	  Do[
	     If[ Head[ metricName[i,j]] =!= metricName ,
		(* assign like this so only need to simplify [i,j] *)
		grtG`etabupbup[gname,j,i] = grtG`etabupbup[gname,i,j];
		grtG`etabupbup[gname,i,j] = metricName[i,j];
	     ,(*else*) 
		grtG`etabupbup[gname,j,i] = 0;
		grtG`etabupbup[gname,i,j] = 0;
	     ]; 
	 ,{ j , i, grtG`Ndim[gname]} ]
	,{i, grtG`Ndim[gname] } ];
 ]

(**************************************
* grtF`initMetric[ gname]
*
*  gname    name of the metric to initalize
*
* Common metric initialization code. This is
* separate because in the future will wish to
* allow metric to be specified by a formula.
*
***************************************)

grtF`initMetric[ gname_, gtype_] :=
 Module[ {i}, 
	(* initalize records for the new metric *)
	(*
	 * build array of coord names with leading spaces for use
	 * by grComponents
	 *)
	Do[
		grtG`coordStr[gname,i] = ToString[ grt`xup[gname,i] ];
	 {i, grtG`Ndim[gname] }];
	grtG`defaultMetric = gname;
	grtG`metricSet = Join[ grtG`metricSet, {gname}];
	Write[grt`stream, "Default metric = ",  gname];
	(*
	 * naughty, but explicitly calc ds (since if goes through simpHow
	 * it gets messy and this is best work around)
	 *)
	grtG`metricName = gname;
	grt`grdisplay[x[up], metric->gname]; 
	Which[
	   gtype === 1, (* metric *)
	     grtG`ds[gname] = grtF`calcds[ds,{}];
	     (* add to list of calculated objects *)
	     grtG`record[gname] = Join[ grtG`record[gname], {ds} ];
	     grt`grdisplay[ ds, metric->gname],
           
	   gtype === 2 || gtype === 3, (* basis *)
	     grt`grdisplay[ eta[bup,bup] ];
             grtG`EtaNotConstant[ gname] = 
                   grtF`checkEtaComponents[ grtG`Ndim[gname] ];
	     If[ MemberQ[ grtG`record[ gname], e[bdn,up] ],
	        grt`grdisplay[ e[bdn,up] ];
             ];
	     If[ MemberQ[ grtG`record[ gname], e[bdn,dn] ],
	        grt`grdisplay[ e[bdn,dn] ];
             ],
           gtype === 3, (* NP *)
	     Print["load NP not finished"]
        ];
	(*
	 * display info line and constraint if they exist
	 *)
	 If[ MemberQ[ grtG`record[gname], Info],
	    grt`grdisplay[ Info];
 	 ];
	 If[ MemberQ[ grtG`record[gname], constraint], 
	    grt`grdisplay[ constraint];
         ];
	 Return[];
 ]

(*
 * check to see if eta is constant. If it is then we can use 
 * a higher symmetry for rot[bdn,bdn,bdn] and a simpler R[bdn,bdn,bdn,bdn]
 * definition
 *)

grtF`checkEtaComponents[ n_] := 
Module[ {a,b,c},
  For[ a=1, a <= n, a++, 
    For[ b=1, b <= n, b++,
      For[ c=1, c <= n, c++,
        If[ D[ grtG`etabupbup[ grtG`metricName, a, b], 
                  grtG`xup[ grtG`metricName, c] ] =!= 0,
            Return[True];
        ];
      ];
    ];
  ]; 
  Return[ False];
];
