(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 




Clear["grt`*","grtC`*","grtF`*","grtG`*"];


grt`stream = $Output;



BeginPackage["grt`"];

(*

 * turn off Needs no context warnings, leave it turned off

 * so when we autoload NP or cmscalar we don't get messages

 * 

 *)

Off[Needs::nocont];



(* GRTensor code *)

Needs["grii`cmdsup`"]

Needs["grii`miscfn`"];

Needs["grii`globals`"];

Needs["grii`core`"];

Needs["grii`symmetry`"];     (* this makes use of functions in miscfn`"]; *)

Needs["grii`grcomp`"];

Needs["grii`calcsum`"];

Needs["grii`newindex`"];

Needs["grii`makeg`"];

Needs["grii`grload`"];



(* grdefine stuff *)

Needs["grii`define`"];

Needs["grii`define2`"];



(* object definition files *)

Needs["grii`autoLoad`"];

Needs["grii`objects`tensors1`"];

Needs["grii`objects`gcalc`"];

Needs["grii`objects`basis`"];

Needs["grii`objects`bdweyl`"];

Needs["grii`objects`Vectors`"];



(* help files *)

Needs["grii`grhelp`"];

Needs["grii`Options`"];

(*

 * OPTIONS defined in Options.m

 *

 * This is used to paste them into a form acceptable

 * to the ported Maple code

 *)



grtF`setOptions[] := (

	grt`OptionMetricPath = MetricPath /. Options[grii];

	grt`OptionDefaultSimp = DefaultSimplification /. Options[grii];

	grt`OptionDisplayLimit = DisplayLimit /. Options[grii];

	grt`OptionTimeStamp = TimeStamp /. Options[grii];

	);

(*

 * MESSAGES

 *)



grii::create = "Created a definition for `1`";

grii::autoload = "Loading objects from `1`";

grii::overload = "Metric `1` exists, loading as `2`";

grii::nodef = "No definition found for object `1`";



(**************************************

* grcalc[ ]

*

* Calculate the components of tensor or scalar objects.

*

***************************************)



grt`grcalc[ objects_,  opts___ ] :=

 Module[ {i, newArgs, toc},

	 toc = TimeUsed[];

	 If[ !VectorQ[ objects], newArgs = {objects} , newArgs = objects];


	 (* option check *)

	 grtG`metricName = metric /. {opts} /. metric->grtG`defaultMetric;

	 grtG`simpHow = DefaultSimplification /. {opts} /. Options[grii];

	 grt`OptionDisplayLimit = DisplayLimit /. {opts} /. Options[grii];

	 grt`OptionTimeStamp = TimeStamp /. {opts} /. Options[grii];

	 grt`OptionVerbose = Verbose /. {opts} /. Options[grii];


	 (* check that all required objects exist and have not been calc'd *)

	 newArgs = grtF`screenArgs[ newArgs, False, True];


	 (* set the global variables *)

	 grtG`fnCode = grtC`calc;

	 grtG`calc = True;

	 grtG`callComp = False;

	 grtG`simp = True;

	 If[ grt`OptionVerbose,

	     Write[ grt`stream, "For the ", grtG`metricName, " metric."]

	 ];

	 Do[

	    grtF`core[ newArgs[[i]] ],

	 { i, Length[ newArgs ] }];

	 If[ grt`OptionVerbose,

	    Write[ grt`stream, "CPU Time = ", TimeUsed[] - toc," sec."]

	 ];

	 Return[];

 ]



(**************************************

* grclear[ ]

*

* If there is only one name then it might be the

* keyword metric or results, otherwise it is a

* list of objects which are to be cleared.

*

**************************************)



grt`grclear[ objects_, opts___] :=

 Module[ { prompt, yorn, clearList, i},



	 (* extract the metric name if it is an option *)

	 grtG`metricName = metric /. {opts} /. grtG`DefaultOptions[grdisplay];



	 If[ !VectorQ[ objects], newArgs = {objects} , newArgs = objects];



	 (* are we being asked to clear metric or results ?? *)

	 If[ newArgs[[1]] === results || newArgs[[1]] === metric,

	   (*

	* prompt for a confirmation, unless an override was

	* specified in the opts field

	*)

	prompt = confirm /. {opts} /. {confirm->True};

	If[ prompt,

	   (* prompt for confirmation *)

	   yorn = InputString[ SequenceForm[ "Clear ",newArgs[[1]]," for the ",

		 grtG`metricName, " metric (y/n) ? "] ];

	   If[ yorn == "n" , Abort[] ];

	];

	(*

	 * Build a list of things to clear

	 *)

	If[ newArgs[[1]] === metric,

	  (*

	   * Clear the metric and all results

	   * (Dimension is needed to clear other objects, make

	   *  sure it goes at the end of the list! )

	   *)

	   clearList = Complement[ grtG`record[grtG`metricName], {dimension}];

	   clearList = Join[ clearList, {dimension}];

	   grtG`record[grtG`metricName] = .; (* unassign *)

	   grtG`metricSet = Complement[ grtG`metricSet, {grtG`metricName}];

	,(*else*)

	   (*

		* Clear everthing except the metric, dimension and xup

		*)

	   clearList = Complement[ grtG`record[grtG`metricName],

			grtG`recordInit ];

	   grtG`record[ grtG`metricName] = grtG`recordInit;

	];



	 ,(*else*)

	   (*

	* check that all required objects exist and have been calc'd

	* and remove them from the record of calculated objects

	*)

	   clearList = grtF`screenArgs[ newArgs, True, False];

	   grtG`record[grtG`metricName] =

		Complement[ grtG`record[grtG`metricName], newArgs ];

	 ];



	 (* set the global variables *)

	 grtF`setOptions[];

	 grtG`fnCode = grtC`clear;

	 grtG`simp = True;

	 grtG`calc = False;

	 grtG`callComp = False;

	 grtG`simpHow = Unset;



	 Write[ grt`stream, "For the ", grtG`metricName, " metric."];

	 Do[

	 grtF`core[ clearList[[i]] ],

	{ i, Length[ clearList ] }

	 ];

	 Return[];



 ]



(*----------------------------------------------------------

* grcomponent[ object, iList]

*    metric     metric name

*    object     generic object name

*    iList      index list

*               may be in coord number of name

*

* Return the value of the component of the object with

* indices as given (numerical or coord Name) in iList.

*----------------------------------------------------------*)



grcomponent[ object_, iList_:{}, opts___] :=

 Module[ { expr, coord, a, b, newList, i},



  (* extract the metric name if it is an option *)

  grtG`metricName = metric /. {opts} /. metric->grtG`defaultMetric;





  If[ !MemberQ[ grtG`record[ grtG`metricName], object],

	Write[grt`stream, object," is not defined for ", grtG`metricName];

	Abort[];

  ];



  (*

   * convert from coord name to number if required

   *)

  root = grtG`ObjDef[ object][ grtC`root];

  If[ iList =!= {},

     coord = Table[ grtG`xup[ grtG`metricName,i], {i,grtG`Ndim[grtG`metricName]}];

     If[ !IntegerQ[ iList[[1]] ],

       (*

	* coord name, so look it up

	*)

	newList = Flatten[ Map[ Flatten[Position[ coord, #]]&, iList] ];

     ,(*else*)

	 newList = iList;

     ];

  , 

     newList = {};

  ];

  If[ Length[ newList] =!= Length[ grtG`ObjDef[object][grtC`indexList]],

    Print["Wrong number of indices given"];

    Abort[];

  ];

  If[ grDebug, Print["grcomponent: newList ", newList] ];

  newList = newList //. List->Sequence;

  Return[ root[grtG`metricName, newList] ];



]



(**************************************

* grdisplay[ ]

*

* Display the components of tensor or scalar objects.

*

***************************************)



grt`grdisplay[ objects_,  opts___ ] :=

 Module[ {i, newArgs},



	 (* extract the metric name if it is an option *)

	 grtG`metricName = metric /. {opts} /. metric->grtG`defaultMetric;

	 grt`OptionDisplayLimit = DisplayLimit /. {opts} /. Options[grii];



	 (* check that all required objects exist and have been calc'd *)

	 If[ !VectorQ[ objects], newArgs = {objects} , newArgs = objects];

	 
	newArgs = grtF`screenArgs[ newArgs, True, False];



	 (* set the global variables *)

	 grtG`fnCode = grtC`display;

	 grtG`simp = False; (* this must be before simpDecode *)

	 grtG`calc = False;

	 grtG`callComp = True;

	 Write[ grt`stream, "For the ", grtG`metricName, " metric."];

	 Do[

	    grtF`core[ newArgs[[i]] ],

	 { i, Length[ newArgs ] }];

	 Return[];

 ]



(**************************************

* grlist[ object]

* 

* Return the essential components (up to symmetry)

* of the named object as a list.

**************************************)

grtF`lister[ x_] :=

Module[{},

    (*

     * add to listerGlobal as a side effect

     *)

    grtG`listerGlobal = Join[ grtG`listerGlobal, {x}]; 

    Return[x];

];



grt`grlist[ object_, opts___] :=

Module[ {},

  (*

   * quick and dirty.  

   *)

   grtG`listerGlobal = {};

   grmap[ grtF`lister, object, opts, Verbose->False ];

   Return[ grtG`listerGlobal];



];



(**************************************

* grmap[]

*

* Apply a user specified function to tensor

* components or scalar objects. The component

* takes on the value returned by this function.

*

* Bug: if metric-> used this gets passed to the user function.

*

***************************************)



(* utility routine for grmap to apply constraints *)

ApplyConstraints[ x_] := ReplaceAll[ x, grtG`constraint[grtG`metricName]];



(* this is the old one *)

grt`grmap[ funct_, objects_, opts___ ] :=

 Module[ {i, newArgs, tic},

	 tic = TimeUsed[];

	 (* make sure both parameters end up being lists *)

	 If[ !ListQ[ objects], newArgs = {objects} , newArgs = objects];

	

	 (* options *)

	 grtG`metricName = metric /. {opts} /. metric->grtG`defaultMetric;

	 grt`OptionTimeStamp = TimeStamp /. {opts} /. Options[grii];

	 grt`OptionVerbose = Verbose /. {opts} /. Options[grii];

	

	 (* check that all required objects exist and have been calc'd *)

	 newArgs = grtF`screenArgs[ newArgs, False, False];



	 (* set the global variables *)

	 grtG`fnCode = grtC`alter;

	 grtG`calc = False;

	 grtG`callComp = False;

	 grtG`simpHow = funct;

	 grtG`simp = True;



	 If[ grt`OptionVerbose, 

	   Write[ grt`stream, "For the ", grtG`metricName, " metric."];

	 ];

	 Do[

	   If[ grt`OptionVerbose, 

	      Print["Applying ", funct, " to ", newArgs[[i]] ];

	   ];

	   grtF`core[ newArgs[[i]] ],

	 { i, Length[ newArgs ] }];

	If[ grt`OptionVerbose,

          Write[ grt`stream, "CPU Time = ", TimeUsed[] - tic," sec."];

	];



 ] (* end grmap *)





(************************************************

* grmetric[ new]

*

* Change the default metric to new (if it exists)

*

************************************************)



grt`grmetric[ new_] :=

 Module[ {},

  If[ MemberQ[ grtG`metricSet, new],

	 grtG`defaultMetric = new;

  , (*else*)

	 Write[ grt`stream, new, " is unknown."]

  ];

  Write[ grt`stream, "Default metric is ", grtG`defaultMetric]

 ]



(************************************************

* grtensor[]

*

* Display banner and getting sarted message.

*

************************************************)



grt`grtensor[] :=

 Module[ {},

  grtF`setOptions[];

  Write[ grt`stream," GRTensorM Version 1.2 for Mathematica 3.x"];

  Write[ grt`stream," July 20, 1998\n"];

  Write[ grt`stream," Copyright 1996-98 by P. Musgrave, D. Pollney and K. Lake"];
  Write[ grt`stream," Update by Mustapha Ishak B."]; 

  Write[ grt`stream," Use qload[] or grload[] to load a metric."];

  Write[ grt`stream," ?grtensor for help.\n"];

  Write[ grt`stream," Metric directory = ", InputForm[grt`OptionMetricPath]]; 

 ]



(************************************************

* qload[ name]

*

* Load a metric from c:\makeg\name\name.g

*

************************************************)



grt`qload[ name_] := (

   (* set the global variables *)

   grtF`setOptions[];

   grt`grload[ name,ToString[ SequenceForm[grt`OptionMetricPath,name,".g"]]]

   )

grt`grtensor[]



(* DON'T PUT ANYTHING AFTER HERE *)



EndPackage[];

