(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 
(* these are used to denote menu choices and hence must be assigned *)
grtG`metricComp = 1;
grtG`lineElement = 2;
grtG`basis = 3;
grtG`np = 4;
grtG`maxTermLength  =  OptionTermSize;




makeg[ metricName_] :=

 Module[ {a,b,
	metric,
	basisu,
	basisd,
	gdim,
	gtype,
	constraint,
	info,
	happy, abort, response, coords, metrictype, bases},

  Print[ "" ];
  Print[ "Makeg 2.0; GRTensor metric/basis entry utility" ];
  Print[ "" ];
  Print[ "To quit makeg, type 'Exit' at any prompt." ];
  Print[ "" ];

  (*
   *  Check metric name 
   *)
  If[ MemberQ[  grtG`metricSet, metricName] , (* then *)
     Print["Metric name has already been used."];
     Abort[];
  ]; (* end If *)
  
  gtype = grtF`setgtype[];

  (*
   *  Set coordinates and dimension.
   *)
  coords = grtF`getCoords[];
  If[ gtype ===  grtG`np && Length[ coords] =!= 4 , (* then *)
	Print[ "A null tetrad must have four coordinates." ];
	Abort[];
  ]; (* end If *)
  gdim  = Length[coords];

  (*
   *  Get input based on type of metric.
   *)
  basisu = 0;
  basisd = 0;
  Which[
    gtype ===  grtG`metricComp ,  (* then *)
	metric = grtF`getg[ gdim, coords],

    gtype ===  grtG`lineElement , (* then *)
	metric = grtF`getds[ gdim, coords ],
    
    gtype ===  grtG`basis , (* then *)
	bases  = grtF`promptBasisV[ gdim, gtype, coords];
	basisu = bases[[1]];
	basisd = bases[[2]];
	metric = grtF`getBasisInnerProd[ gdim, coords],

    gtype ===  grtG`np , (* then *)
	bases    = grtF`promptBasisV[ gdim, gtype, coords];
	basisu = bases[[1]];
	basisd = bases[[2]];
	(* init eta for NP tetrad *)
	Do[ Do[ metric[a,b] = 0, {b,4}], {a,4}];
	metric[2,1] = 1;
	metric[1,2] = 1;
	metric[3,4] = -1;
	metric[4,3] = -1
  ]; (* end Which *)
  happy = False;
  abort = False;
  While[ !happy ,
	grtF`displayEntry[metric,basisu,basisd,
			gdim,coords,gtype, constraint, info];
	response = grtF`getdisplayResponse[gtype,basisu,basisd];
	
	Which[ response  ===  newmetric , (* then *)
		metric = grtF`getg[ gdim, coords],

	response  ===  newbasisuv , (* then *)
		basisd = grtF`getBasisV[ gdim, gtype, coords,  1 ],

	response  ===  newbasisdv , (* then *)
		basisu = grtF`getBasisV[ gdim, gtype, coords, -1 ],

	response  ===  newbasisip , (* then *)
		metric = grtF`getBasisInnerProd[ gdim, coords ],

	response  ===  correctMetric , (* then *)
		metric = grtF`correctMetric[ metric, coords ],

	response  ===  correctbasisuv , (* then *)
		basisu = grtF`correctBasisV[ gdim, gtype, basisu, coords ],

	response  ===  correctbasisdv , (* then *)
		basisd = grtF`correctBasisV[ gdim, gtype, basisd, coords ],

	response  ===  addconstraints , (* then *)
		constraint = grtF`addConstraints[],

	response  ===  addtext , (* then *)
		info = grtF`addText[],

	response  ===  saveEntry , (* then *)
		grtF`saveEntry[metric,basisu,basisd,
			gdim,coords,gtype,constraint, info, metricName];
		happy = True,

	True, (* else *)
		Print[ "grmake aborted at user's request." ];
		happy = True;
		abort = True;

	]; (* end Which *)

    ]; (* end While *)

    If[ !abort , (* then *)
	response = 
grtF`myInput["\nDo you wish to use this spacetime in the current session? (1 = yes)\n"];
	If[ response  ===  1 , (* then *)
		qload[metricName];
	]; (* end If *)
    ]; (* end If *)
    Print[ "grmake completed." ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  addconstraints. *)
(* ------------------------------------------------------------------- *)
grtF`addConstraints[] :=
Module[ {constraint, prompt},
  Clear[constraint];
    prompt = StringJoin[ "Enter constraints as a list of rules\n",
       "e.g. {D[ m[r,t], r ] ->  r^2*m[r,t],\n",
       "      D[ m[r,t], t ] -> t^2*m[r,t]^2 }\n"];
  While[ !ValueQ[ constraint],
    constraint = grtF`myInput[prompt];
    If[ !ListQ[ constraint] , (* then *)
	Print[ "Invalid input. Please enter a List of rules." ];
	Clear[constraint];
    , (* else *)
     (*  
      *  Screen the constraints. Are they equations? 
      *)
      If[ Complement[ Map[ Head, constraint], {Rule}] =!= {},
	 Print["Each element must be a rule. Please re-enter"];
	 Clear[constraint];
      ]; 
    ]; (* end If *)
  ]; (* end While *)

 Return[ constraint ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  addtext. *)
(* ------------------------------------------------------------------- *)
grtF`addText[] :=
 Module[ {info},
	info = grtF`myInput[ "Enter text information (as a string):\n" ];
 Return[ info];
]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  grF_checkbasisproblems *)
(* ------------------------------------------------------------------- *)
grtF`checkBasisProblems[ gdim_, ip_, up_, dn_ ] :=

Module[ {a, b, c, ipu, ipd, btest, problem},
  (*
   * since the user entered both up/dn basis vectors
   * if they made an error then there might be
   * an inconsistency with the inner product they
   * specified. Check things are ok
   *)
  ipu = Table[ ip[a,b], {a,gdim},{b,gdim}];
  ipd = Inverse[ipu ];

  problem = False;
  For[ a = 1, a <=  gdim && problem  ===  False , a++ ,
    For[ b = 1, b <=  gdim && problem  ===  False , b++ ,
	btest[a,b] = 0;
	For[ c = 1, c <=  gdim  , c++ ,
	  btest[a,b] = btest[a,b] + up[a,c]*dn[b,c];
	]; (* end For *)
	If[ btest[a,b] =!= ipd[a,b] , (* then *)
	  problem = True;
	]; (* end If *)
    ]; (* end For *)
  ]; (* end For *)

  Return[ problem];
];
(* ------------------------------------------------------------------- *)
(*  correctBasisV. *)
(* ------------------------------------------------------------------- *)
grtF`correctBasisV[ gdim_, gtype_, basis_, coord_ ] :=

Module[ {vinput, i,
	npvector, 
	newvector, 
	a, e1, v, val,
	prompt},

  If[ gtype  ===  grtG`np , (* then *)
	vinput = "(l,n,m,mbar)";
	prompt = StringJoin["Which vector to you wish to correct ?\n",
		"1) l\n2) n\n3) m\n4) mbar"];
	npvector = {l,n,m,mbar};
  , (* else *)
	vinput = "{1,2,3,4}";
	prompt = StringJoin["Which vector to you wish to correct ?\n",
		"1..",ToString[gdim] ];
  ]; (* end If *)
  newvector = {};
  While[ !IntegerQ[ newvector ] ,
    newvector = grtF`myInput[prompt];
    If[ newvector < 1 || newvector > gdim || !IntegerQ[newvector], (* then *)
	  Print[ "Invalid input. Please enter a value from the set",vinput ];
	  newvector = {};
    ] (* If *)
  ]; (* end While *)

  val = 0;
  While[ !ListQ[ val],
     val = grtF`myInput["\nPlease enter the new vector as a List.\n"];
     If[ !ListQ[ val] || Length[ val] =!= gdim , (* then *)
	Print[ "Invalid input. Entry must be a LIST of dimension",gdim];
	val = 0;
     ]; (* end If *)
  ]; (* end While *)
  For[ a = 1, a <=  gdim , a++ ,
     basis[newvector,a] = val[[a]];
  ]; (* end For *)

 Return[ basis];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  correctMetric. *)
(* ------------------------------------------------------------------- *)
grtF`correctMetric[ metric_, coord_ ] :=
Module[ {a, e1, e2, newelement, val},
  Clear[newelement];
  While[ !ValueQ[ newelement],
    newelement = 
       grtF`myInput["\nWhich element do you wish to correct?(e.g. {r,theta}) \n" ];
    Which[
     !ListQ[ newelement] || Length[ newelement] =!= 2 , (* then *)
	Print[ "Invalid input. Please enter two coordinate names as a list." ];
	Clear[newelement],

     !MemberQ[ coord, newelement[[1]] ] || !MemberQ[ coord, newelement[[2]] ] , (* then *)
	Print[ "Enties are not coordinates specified for this metric." ];
	Clear[newelement]
    ]; (* end Which *)
  ]; (* end While *)

  e1 = Flatten[ Position[ coord, newelement[[1]] ]][[1]]; 
  e2 = Flatten[ Position[ coord, newelement[[2]] ]][[1]]; 
  Print[ "Current value:" ];
  Print[ SequenceForm[ g, 
       Subscript[ newelement[[1]] ], Subscript[ newelement[[2]] ]," = ",metric[e1,e2] ]];
  metric[e1,e2] = grtF`myInput["New value:" ];
  metric[e2,e1] = metric[e1,e2];

 Return[ metric ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  displaybasisip. *)
(* ------------------------------------------------------------------- *)
grtF`displayBasisIp[ gdim_, ip_ ] :=
Module[ {a,b},
  Print["The inner product:\n"];
  Do[
     Do[
       Print[ SequenceForm[ eta, 
	  Subscript[ StringJoin[ "(", ToString[a],")" ]],
	  Subscript[ StringJoin[ "(", ToString[b],")" ]],
	  " = ", ip[a,b] ]]
       ,{b,a,gdim}]
    ,{a,gdim}];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  displayBasisV. *)
(* ------------------------------------------------------------------- *)
grtF`displayBasisV[ gdim_, gtype_, basis_, btype_ ,coord_] :=
Module[ {a, b, maxSize, size, vname, v, e},

  If[ btype  ===  1 , (* then *)
	Print[ "Basis vectors:" ];
  , (* else *)
	Print[ "Basis 1-forms:" ];
  ]; (* end If *)

  vname = {"l","n","m","mbar"};

  For[ a = 1, a <=  gdim , a++ ,
    If[ gtype === grtG`basis, 
       vhead = SequenceForm[ "e", 
		  Subscript[ StringJoin["(",ToString[a],")"]]]
    ,(*else*)
       vhead = vname[[a]]
    ];
    For[ b = 1, b <=  gdim , b++ ,
       If[ btype  ===  1 , (* then *)
	     Print[ vhead, Superscript[ coord[[b]] ],
		  " = ", basis[a,b] ];
	   , (* else *)
	     Print[ vhead, Subscript[ coord[[b]] ], 
		  " = ", basis[a,b] ];
       ]; (* end If *)
     ]; (* end For *)
  ]; (* end For *)

]; (* end module *)

(* ------------------------------------------------------------------- *)
(*  displayEntry. *)
(* ------------------------------------------------------------------- *)
grtF`displayEntry[ metric_,
		   basisu_,
		   basisd_,
		   gdim_,
		   coord_,
		   gtype_,
		   constraint_,
		   info_] :=
Module[ {alert},
  
  Print[ "The values you have entered are:" ];
  Print[ "Coordinates =", coord];
  Which[
    gtype  ===  grtG`metricComp || gtype  ===  grtG`lineElement , (* then *)
	grtF`displayg[ gdim, metric, coord ],

    gtype  ===  grtG`basis || gtype  ===  grtG`np , (* then *)
	If[ basisu =!=0 , (* then *)
	   grtF`displayBasisV[ gdim, gtype, basisu, 1 , coord]
	]; (* end If *)

	If[ basisd =!=0 , (* then *)
	   grtF`displayBasisV[ gdim, gtype, basisd, -1, coord];
	]; (* end If *)
	grtF`displayBasisIp[ gdim, metric ];
	If[ basisd =!= 0 && basisu =!= 0 , (* then *)
	   alert = grtF`checkBasisProblems[ gdim, metric, basisu, basisd ];
	   If[ alert, (* then *)
       	     Print[ "Warning: the basis vectors are not consistent with the inner product" ];
	   ]; (* end If *)
	] (* end If *)
  ]; (* end Which *)

  If[ StringQ[ info ] , (* then *)
	Print[ "Text information:", info ];
  ]; (* end If *)

  If[ ListQ[ constraint], 
	Print["The constraints are: ", constraint];
  ];

]; (* end Module *)


(* ------------------------------------------------------------------- *)
(*  displayg. *)
(* ------------------------------------------------------------------- *)
grtF`displayg[gdim_, metric_, coord_ ] :=
Module[ {a, b},
  Print[ "Metric:" ];
  maxSize = 0;

      For[ a = 1, a <=  gdim , a++ ,
	 For[ b = a, b <=  gdim , b++ ,
	    Print[ SequenceForm[ g, 
	       Subscript[coord[[a]]]," ",
	       Subscript[coord[[b]]]," = ",metric[a,b] ]];
	 ]; (* end For *)
      ]; (* end For *)

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getBasisInnerProd. *)
(* ------------------------------------------------------------------- *)
grtF`getBasisInnerProd[ gdim_, gtype_ ] :=
Module[ { a, b, diag, metric},
 Array[ metric,{ gdim, gdim} ];

 While[ !ValueQ[ diag],
   diag = grtF`myInput[ StringJoin[
	"Is the basis inner product\n",
	"1) Diagonal, or\n",
        "2) Symmetric?\n" ]];
   If[ diag < 1 || diag > 2 , (* then *)
	Print[ "Invalid input. Please enter 1 or 2." ];
	Clear[diag];
   ]; (* end If *)
 ]; (* end For *)

 If[ diag  ===  1 , (* then *)
   For[ a = 1,  a <=  gdim , a++ ,
      For[ b = a+1, b <=  gdim , b++ ,
	metric[a,b] = 0;
	metric[b,a] = 0;
      ]; (* end For *)
   ]; (* end For *)

   For[ a = 1, a <=  gdim , a++ ,
	metric[a,a] = grtF`getIpElement[ a, a ];
   ]; (* end For *)
 , (* else *)

   For[ a = 1, a <=  gdim , a++ ,
     For[ b = a, b <=  gdim , b++ ,
	metric[a,b] = grtF`getIpElement[ a, b];
	metric[b,a] = metric[a,b];
     ]; (* end For *)
   ]; (* end For *)
 ]; (* end If *)

 Return[ metric];
]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getBasisV. *)
(* ------------------------------------------------------------------- *)
grtF`getBasisV[ gdim_, gtype_, coord_, vtype_ ] :=
Module[ {i, j, vec, npname, basis, vstr},

  vname = { "l", "n", "m", "mbar" };
  If[ gtype  ===  grtG`basis , (* then *)   (*  general basis *)
      (* build a list of generic basis vector names *)
       vname = {};
       Do[ vname = Append[ vname, ToString[ SequenceForm[ "e[",a,"]" ]]]
	,{a,gdim}];
  ];

  If[ vtype  ===  -1 , (* then *)
     vstr = "covariant";
  , (* else *)
     vstr = "contravariant";
  ]; (* end If *)

  For[ i = 1, i <=  gdim , i++ ,
    (* for each vector in the basis *)
    Clear[vec];
    While[ !ListQ[ vec ] ,
         vec = grtF`myInput[ ToString[ SequenceForm[
	    "Enter the ",vstr,
	    " components of basis vector ",vname[[i]],
	    " as a List (e.g. {1,0,0,0})\n" ]]];
       If[ !ListQ[ vec] , (* then *)
	   Print[ "Invalid input. Please enter the vector as a List." ];
	   Clear[vec];
       ]; (* end If *)
    ]; (* end For *)
    For[ j = 1, j <=  gdim , j++ , basis[i,j] = vec[[j]] ]; (* end For *)
  ]; (* end For *)

 Return[ basis ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getcoords. *)
(* ------------------------------------------------------------------- *)
grtF`getCoords[] :=
Module[ { coords},
   
   While[ !ListQ[ coords] , 
      coords = grtF`myInput["\nEnter coordinates as a LIST (e.g. {r,theta,phi,t})\n"];
      If[ !ListQ[ coords] , (* then *)
	Print[ "Invalid input. Coordinates must be a LIST. Please try again." ];
	coords = 1;
      ]; (* end If *)
   ]; (* end While *)

 Return[ coords];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getdisplayResponse *)
(* ------------------------------------------------------------------- *)
grtF`getdisplayResponse[ gtype_, basisu_, basisd_] :=
Module[ {choice, choices, selection, prompt, vtype, ambiguity1, ambiguity2},

  Which[
    gtype  ===  grtG`metricComp || gtype  ===  grtG`lineElement , (* then *)
	prompt = StringJoin[
	 "\nYou may choose to\n1) Save the metric as it is,\n",
	 "2) Correct an element of the metric,\n",
	 "3) Re-enter the metric,\n",
	 "4) Add/change constraints,\n",
	 "5) Add a text description, or\n",
	 "6) Abandon this metric and return to Mma\n"];
	choices = { saveEntry, correctMetric, newmetric, 
		    addconstraints, addtext, abandon },

    gtype  ===  grtG`basis , (* then *)
	prompt = StringJoin[
	 "\nYou may choose to\n1) Save the metric as it is,\n",
	 "2) Re-enter a basis vector,\n",
	 "3) Re-enter the inner product,\n",
	 "4) Add/change constraints,\n",
	 "5) Add a text description, or\n",
	 "6) Abandon this metric and return to Mma\n" ];
	Which[
          basisu =!= 0 && basisd === 0 , (* then *)
		choices = { saveEntry, correctbasisuv, newbasisip, 
		     addconstraints, addtext, abandon },

          basisd =!=0 && basisu === 0 , (* then *)
		choices = { saveEntry, correctbasisdv, newbasisip, 
		    addconstraints, addtext, abandon },
          True, (* else *)
	        choices = { saveEntry, correctBasisV, newbasisip, 
		    addconstraints, addtext, abandon }
	], (* end Which  *)

    gtype  ===  grtG`np , (* then *)
	prompt = StringJoin[
	 "\nYou may choose to\n1) Save the metric as it is,\n",
	 "2) Re-enter a basis vector,\n",
	 "3) Add/change constraints,\n",
	 "4) Add a text description, or\n",
	 "5) Abandon this metric and return to Mma\n" ];
	choices = { saveEntry, correctBasisV, addconstraints, addtext, abandon,  nil }
    ]; (* end Which *)

   prompt = StringJoin["(the metric has been displayed on the worksheet)\n\n",
	     prompt];

   Clear[choice];
   While[ !ValueQ[ choice ] ,
	choice = grtF`myInput[prompt];
	If[ !MemberQ[ { 1, 2, 3, 4, 5, 6 }, choice] , (* then *)
		selection = nil;
	, (* else *)
		selection = choices[[choice]];
	]; (* end If *)
	If[ selection  ===  nil , (* then *)
		Print[ "Invalid input." ];
		Clear[choice];
	]; (* end If *)
   ]; (* end While *)

   If[ selection  ===  correctBasisV , (* then *)
	ambiguity1 = basisu === 0 && basisd === 0;
	ambiguity2 = !( basisu === 0 || basisd === 0 );
	Which[
          ambiguity1 || ambiguity2 , (* then *)
 	    While[ !ValueQ[ vtype] ,
	      vtype = grtF`myInput[ StringJoin[
		  "Would you like  to correct a\n1) covariant vector, or\n",
	          "2) contravariant vector?"] ];
	      If[ !MemberQ[ { 1, 2 }, vtype] , (* then *)
		Print[ "Invalid input. Please enter either 1 or 2."];
		Clear[vtype];
	      ]; (* end If *)
	    ], (* end While *)

	  basisu === 0 , (* then *)
		vtype  ===  1,

	  basisd === 0 , (* then *)
		vtype  ===  2
         ]; (* end which *)
	 If[ vtype  ===  1 , (* then *)
		selection = correctbasisdv
	 , (* else *)
	        rselection = correctbasisuv
	 ]; (* end If *)
   ]; (* end If *)

   Return[ selection ];

]; (* end Module *)

(*------------------------------------------------------------------- *)
(*  getds. *)
(*------------------------------------------------------------------- *)
grtF`getds[ gdim_, coord_ ] :=

Module[ {ds, metric, i, j, entry, happy, divby},

  happy = False;
  While[ !happy , 
    happy = True;
    (*
     * Clear the metric array
     *)
    Do[
      Do[
        metric[a,b] = 0,
      {a,gdim}],
    {b,gdim}];

    ds = grtF`myInput[StringJoin[
         "Enter the line element ds^2 using 'dcoord' to indicate differentials\n",
         "For example: r^2*(dtheta^2 + Sin[theta]^2*dphi^2)\n",
         "(Type Exit to quit makeg.)\n" ]];

   (*  
    *  now expand ds^2 and sift through picking out terms 
    *  and putting them in the metric array 
    *)
    ds = Expand[ds];

    If[ Length[ds]  ===  1 , (* then *)
    	Print["Line element has only a single term."];
        Abort[];
    ]; (* end If *)

   (*
    * keep an array metric, into which we accumlate metric 
    * coeeficents
    *)
    For[ i = 1, i <=  Length[ds] , i++ ,
     (*  
      * for each term in the expanded ds, find out which d<>coords
      * it has.
      *)
      termAtoms = Level[ ds[[i]], {-1}];
      entry = {};
      eliminateD = {};
      For[ j = 1, j <=  gdim , j++ ,
	(* 
	 * for each coordinate , 
	 * a bit wasteful to rebuild elimianteD each loop 
	 *)
	dCoord = ToExpression[ToString[SequenceForm["d",coord[[j]] ]]];
	eliminateD = Append[ eliminateD, dCoord -> 1];
      	If[ MemberQ[ termAtoms, dCoord], 
      		entry = Union[entry,{j}]
      	]; (* end If *)
      ]; (* end For *)
      divby = 2;

      Which[

      entry  ===  {} , (* then *)
      	Print["Error - a term is missing d[coord]"];
      	happy = False;
      	Break[],

      Length[entry]  ===  1 , (* then *)
       	metric[entry[[1]],entry[[1]]] += ReplaceAll[ ds[[i]], eliminateD],

      True	, (* else *)
       	metric[entry[[1]],entry[[2]]] += ReplaceAll[ ds[[i]], eliminateD]/2;
      	(*  copy across to other half of the array *)
       	metric[entry[[2]],entry[[1]]] += ReplaceAll[ ds[[i]], eliminateD]/2
      
      ]; (* end Which *)
    ]; (* end For *)
  ]; 

  Return[ metric ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getg. *)
(* ------------------------------------------------------------------- *)
grtF`getg[ gdim_, coord_ ] :=
Module[ { a, b, diag, metric},

  diag = 0;
  While[ !MemberQ[ {1,2}, diag ] ,
    diag = grtF`myInput["\nIs the metric\n1) Diagonal, or\n2) Symmetric?\n"];
  ]; (* end While *)

  If[ diag  ===  1 , (* then *)
     (* zero the off diagonal terms *)
     For[ a = 1, a <=  gdim , a++ ,
	For[ b = a+1, b <= gdim , b++ ,
      	  metric[a,b] = 0;
	  metric[b,a] = 0;
	]; (* end For *)
     ]; (* end For *)

     For[ a = 1, a <=  gdim , a++ ,
	metric[a,a] = grtF`getMetricElement[ a, a, coord[[a]], coord[[a]] ];
     ]; (* end For *)
   , (* else *)
     (* non- diagonal *)
     For[ a = 1, a <=  gdim , a++ ,
	For[ b = a, b <=  gdim , b++ ,
	  metric[a,b] = grtF`getMetricElement[ a, b, coord[[a]], coord[[b]] ];
	  metric[b,a] = metric[a,b];
	]; (* end For *)
     ]; (* end For *)
   ]; (* end If *)

  Return[ metric ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  getIpElement. *)
(* ------------------------------------------------------------------- *)
grtF`getIpElement[ a_, b_] :=

Module[ { val},

  val = grtF`myInput[ ToString[ SequenceForm[
	 "Enter eta [", a, ",", b, "]:\n" ]]];
  Return[ val];

]; (* end Module *)


(* ------------------------------------------------------------------- *)
(*  getMetricElement. *)
(* ------------------------------------------------------------------- *)

grtF`getMetricElement[ a_, b_, coord1_, coord2_] :=

Module[ { val},
  val = grtF`myInput[ ToString[ SequenceForm[
	    "Enter g[" , coord1 , "," coord2 ,"]:\n" ]]];
  Return[ val ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  promptBasisV. *)
(* ------------------------------------------------------------------- *)
grtF`promptBasisV[ gdim_, gtype_, coord_ ] :=
 Module[ {vtype, basisu, basisd},

   While[ !IntegerQ[ vtype ] ,
     vtype = grtF`myInput[ StringJoin[
         "\nWould you like to enter\n",
	 "1) covariant components,\n",
         "2) contravariant components, or\n",
         "3) both.\n" ]];
     If[ !MemberQ[  {1,2,3},vtype] , (* then *)
	Print[ "Invalid input. Please enter 1, 2, or 3." ];
	Clear[vtype];
     ]; (* end If *)
   ]; (* end While *)

   Which[ vtype  ===  1 , (* then *)
	basisd = grtF`getBasisV[gdim, gtype, coord, -1 ];
	basisu = 0,
   vtype  ===  2 , (* then *)
	basisd = 0;
	basisu = grtF`getBasisV[ gdim, gtype, coord,  1 ],
   True, (* else *)
   	basisd = grtF`getBasisV[ gdim, gtype, coord, -1 ];
	basisu = grtF`getBasisV[ gdim, gtype, coord,  1 ]
   ]; (* end If *)

 Return[ {basisu, basisd} ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  setgtype. *)
(* ------------------------------------------------------------------- *)
grtF`setgtype[] :=
Module[ { gtype},
  While[ !ValueQ[ gtype] , 
     gtype = grtF`myInput[
         StringJoin["\nDo you wish to enter a\n",
		    "  1) metric: g[dn,dn],\n",
		    "  2) line element [ds],\n",
                    "  3) non-holonomic basis e1..e4, or\n",
                    "  4) null tetrad [l,n,m,mbar]?\n" ]];
  If[ !MemberQ[ {1,2,3,4},gtype ] , (* then *)
	 Print[ "Invalid input. Please enter a digit from 1 to 4." ];
	 Clear[gtype];
    ]; (* end If *)
  ]; (* end While *)

  Return[ gtype ];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  grF_my_readstat *)
(* ------------------------------------------------------------------- *)
grtF`myInput[ userprompt_] :=

Module[ {val},
  While[ !ValueQ[ val] , 
	val = Input[ userprompt];
	If[ val  ===  Exit , (* then *)
            Print["Exit requested"];
	    Abort[];
	]; (* end If *)
  ]; (* end While *)
  Return[ val];

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  saveEntry. *)
(* ------------------------------------------------------------------- *)
grtF`saveEntry[ metric_,
		basisu_, 
		basisd_, 
		gdim_, 
		coord_, 
		gtype_,
		constraint_,
		info_,
		metricName_] :=

Module[ {metricfile, path, r, i, j, outStream, a, b},

        path = MetricPath /. Options[grii];
	metricfile = ToString[ SequenceForm[
              path, metricName, ".g" ]];
        (*
         * Check to see if the file exists
         *
        SetDirectory[OptionMetricPath];
        fileList := FileNames[];
        If[ MemberQ[ fileList, metricFile],
           reply = Input["\nThis metric file already exists. Overwrite? (y/n)\n"];
           metricFile = Input["Please enter a new file name."];
        ];
        *)

	outStream = OpenWrite[ metricfile];
	WriteString[ outStream, ToString[SequenceForm[
			"makeg`Ndim = ",gdim,";\n"]]];
	Which[

         gtype  ===  grtG`metricComp || gtype  ===  grtG`lineElement , (* then *)
                (* write out the coordinates and the upper diagonal *)
	        WriteString[ outStream, "makeg`style = 1;\n"];
	        Do[
		  WriteString[ outStream, ToString[SequenceForm[
			"makeg`x[ ",a,"] = ",coord[[a]],";\n"]]];
                  Do[
		     WriteString[ outStream, ToString[SequenceForm[
			"makeg`metric[ ",a,",",b,"] = ",
			InputForm[metric[a,b]],";\n"]]];
	          ,{b,a,gdim}]
                ,{a,gdim}],

         gtype  ===  grtG`basis || gtype  ===  grtG`np , (* then *)
		If[ gtype === grtG`basis, 
	           WriteString[ outStream, "makeg`style = 2;\n"]
                ,(*else*)
	           WriteString[ outStream, "makeg`style = 3;\n"]
		];
                (* write out coordinates and eta (as metric) *)
	        Do[
		  WriteString[ outStream, ToString[SequenceForm[
			"makeg`x[ ",a,"] = ",coord[[a]],";\n"]]];
                  Do[
		     WriteString[ outStream, ToString[SequenceForm[
			"makeg`eta[ ",a,",",b,"] = ",
			InputForm[ metric[a,b] ],";\n"]]];
	          ,{b,gdim}]
                ,{a,gdim}];


		If[ basisu =!= 0 , (* then *)
                  Do[
		    Do[
		     WriteString[ outStream, ToString[SequenceForm[
			"makeg`basisu[ ",a,",",b,"] = ",
			InputForm[ basisu[a,b] ],";\n"]]];
	             ,{b,gdim}]
                  ,{a,gdim}]
	        ];

		If[ basisd =!= 0 , (* then *)
                  Do[
		    Do[
		     WriteString[ outStream, ToString[SequenceForm[
			"makeg`basisd[ ",a,",",b,"] = ",
			InputForm[ basisd[a,b] ],";\n"]]];
	             ,{b,gdim}]
                  ,{a,gdim}]
	        ]

	]; (* end Which *)

	If[ ListQ[ constraint ] , 
		WriteString[ outStream,  ToString[SequenceForm[
			"makeg`constraint = ", InputForm[constraint],";\n" ]] ];
	]; 

	If[ StringQ[ info ] , 
		WriteString[ outStream, ToString[SequenceForm[
		  "makeg`info = \"",InputForm[info],"\";" ]]];
	]; 

	Close[outStream];
	Print[ "Information written to ",metricfile ];

]; (* end Module *)
