(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 


(*----------------------------------------------------------
* grtF`makeD
*
* Create a new object, which is just a covariant or
* partial derivative of a known object
*---------------------------------------------------------*)
grtF`makeD[ newObject_] := 

Module[{calcFn,
	head,
	n,
	newIndex,
	newPerm,
	object,
	oldPerm,
	rootObj,
	raise,
	rootStr,
	},
 If[ grDebug, Print["makeD entered: ", newObject] ];

 raise = 0;
 n = Length[ newObject];
 (*
  * divide and conquer. Strip off an index, force creation
  * of the diminished object and then define the required
  * type of derivative here
  *)
 newIndex = newObject[[ n ]];
 baseObj = Delete[ newObject, n];
 If[ Length[ baseObj] === 0, 
    baseObj = Head[baseObj];
    head = baseObj
 ,(*else*)
    head = Head[baseObj]
 ];

 grtF`checkIfDefined[ baseObj, False]; (* recursion may happen *)
 object = newObject;
 If[ newIndex === cup,
   newIndex = cdn;
   raise = cup;
   object = ReplacePart[ newObject, cdn, n ];
 ];
 If[ newIndex === pup,
   newIndex = pdn;
   raise = pup

;
   object = ReplacePart[ newObject, pdn, n ];
 ];

 (*
  * fill in the object definition record for this object
  *)
  grtG`ObjDef[ object][grtC`rootStr] = ToString[ head];
  grtG`ObjDef[ object][grtC`indexList] = {};
  grtG`ObjDef[ object][grtC`indexList] = ReplaceAll[ object, head->List];
  rootStr = ToString[ SequenceForm[ head,
		ReplaceAll[ object, head->Sequence]]];
  grtG`ObjDef[ object][grtC`header] = ToString[ object];
  grtG`ObjDef[ object][grtC`root] = ToExpression[ StringJoin[ "grtG`", rootStr]];

  depends = { baseObj}; (* start a dependency set, add as we go *)
  grtG`ObjDef[ object][grtC`calcFn] = grtF`calcsum0;

  If[ newIndex === cdn, 
    (*
     * build a function to calc this object
     *)
     depends = Union[ depends, { grt`Chr[dn,dn,up]}];
     grtG`ObjDef[ object][grtC`calcFnParms] = 
	     grtF`buildCoDfunc[ baseObj ];
  ,(*else*)
     (*
      * just take a partial, so feed an argument to
      * calcsum0
      *
      * build a string with the listing indices in place 
      * of the up/dn etc.
      *)
       partial = StringJoin[ "D[ ", 
		  ToString[ grtG`ObjDef[baseObj][grtC`root] ],
		  "[grtG`metricName" ]; 
       Do[
	  partial = StringJoin[ partial, ",grtG`a", ToString[a] ];
       ,{a, n-1}];
       partial = StringJoin[ partial, " ],grtG`xup[grtG`metricName, grtG`a",
		  ToString[ n], "]]" ];
       grtG`ObjDef[object][grtC`calcFnParms] =
	   ToHeldExpression[partial];
   ];
   grtG`ObjDef[ object][grtC`depends] = depends;

   (*
    * SYMMETRY
    *
    * find a symmetry for the new object
    *)
   oldPerm = grtG`symmetry[ grtG`ObjDef[ baseObj][grtC`symmetry] ];
   newPerm = {};
   (* 
    * for each entry in the permList, tack on an extra index
    *)
   If[ oldPerm === {{}},
     (*
      * no symmetry in old object, so that must be the case here
      *)
      grtG`ObjDef[ object][grtC`symmetry] = ToExpression[
			StringJoin[ "grtF`symnosym", ToString[ n]]];
   ,(*else*)
      (*
       * build a new perm list by tacking an index on to 
       * each entry in the old perm list
       *)
      Do[
         newPerm = Union[ newPerm, 
		  { {Join[oldPerm[[a]][[1]], {n}], oldPerm[[a]][[2]]} } ]
      ,{a,Length[oldPerm] } ];
      grtG`ObjDef[ object][grtC`symmetry] = grtF`findSymmetry[ newPerm, n];
   ]; 
   (*
    * if the original object had cup/pup fix that now
    *)
   If[ MemberQ[ {cup,pup}, raise],
      grtG`allObjects = Union[ grtG`allObjects, {object}];
      grtF`newIndices[ newObject];
  ];
  grtG`allObjects = Union[ grtG`allObjects, {newObject}];
];

(*----------------------------------------------------------
* grtF`buildCoDfunc
*
* - build a function to take a covariant derivative
*   of the specified object 
*
---------------------------------------------------------*)

grtF`buildCoDfunc[ object_] :=
(* object is the base object which we are adding a cdn to *)
Module[ { a,b,
	headString,
	lastIndex,
	listing,
	newList,
	partial,
	},
      (*
       * build an object representation with the
       * listing indices in place (but as strings
       * so we don't get values by accident)
       *
       * "grtG`a1", "grtG`a2" 
       *)
      listing = Table[ StringJoin[ ", grtG`a", ToString[i]], {i,Length[object]}];
      (* build the head portion of the object reference
       *)
       headString = StringJoin[ 
		      ToString[ grtG`ObjDef[object][grtC`root] ],
		         "[grtG`metricName" ];
       lastIndex = StringJoin[ ",grtG`a", ToString[ Length[object]+1] ];
      (*
       * Start with the partial term
       *
       * build a string with the listing indices in place 
       * of the up/dn etc.
       *)
       partial = StringJoin[ "D[ ", headString, 
		  ToString[ SequenceForm[ ReplaceAll[
		      listing, List->Sequence]]], 
                  "] ,grtG`xup[grtG`metricName",
		  lastIndex, "]] " ];
      (*
       * now for each up or dn index add the appropriate
       * Chr contraction (using Sum)
       *)
       object2 = ReplaceAll[ object, {cup->up, pup->up, cdn->dn, pdn->dn}];
       Do[ (* for each index *)
	    newList = ReplacePart[ listing, ",grtG`dummy", a];
	 If[ object2[[a]] === dn,
            partial = StringJoin[ partial, " - Sum[ ", 
		  headString,
		  ToString[ SequenceForm[ ReplaceAll[
		      newList, List->Sequence]]], 
                  "] * grtG`Chrdndnup[ grtG`metricName",
		  listing[[a]], lastIndex, ",grtG`dummy] ",
		  ", {grtG`dummy, grtG`Ndim[grtG`metricName]}] "];
	 ];
	 If[ object2[[a]] === up,
            partial = StringJoin[ partial, " + Sum[ ", 
		  headString,
		  ToString[ SequenceForm[ ReplaceAll[
		      newList, List->Sequence]]], 
                  "] * grtG`Chrdndnup[ grtG`metricName",
		  ",grtG`dummy", lastIndex, listing[[a]],
		  "], {grtG`dummy, grtG`Ndim[grtG`metricName]}] "];
	 ];
       ,{a,Length[object]}];
       If[ !SyntaxQ[partial],
	  Print["CoD bug, cannot parse:\n",partial]; Abort[] 
       ];
       Return[ ToHeldExpression[ partial] ];
];


(*----------------------------------------------------------
* grtF`newIndices
*
* Defines a new ObjDef record for an object which can be
* constructed by raising/lowering/converting indices from
* an existing object record.
*
* Should determine symmetry !! (think about this)
*----------------------------------------------------------*)

grtF`newIndices[object_] :=
 Module[ { a,b,cousins,root, newList, baseObj, dependSet,
      dist, nearest, nearObj,expr, exprProd, objSeq, numUp, numDn,
      numSym, mostSym, indexStr, newRoot,
      numbUp,
      index, candidate, ok, realList,
      numbDn},

  root = object[[0]];

  (* make a list of new objects indices, and the new root name (i.e. Rupdn) *)

  newRoot = SequenceForm[ "grtG`", Head[object] ];

  newList = {};
  Do[
    newRoot = SequenceForm[ newRoot, object[[a]] ];
    newList = Join[ newList, {object[[a]]}],
   {a, Length[ object]}
  ];

  cousins = grtF`findCousins[ object];

  If[ grDebug, Print[" New indices ", object, cousins ] ];

 (*
  * FIND NEAREST COUSIN
  *
  * now determine a "distance" from the present object to each cousin.
  * For each entry in indexList different from that in the new object
  * this is a distance of 1 unit.
  * Keep the list with the "least distance"
  *
  * at the same time search for the cousin with the most indices
  * up or dn, this will have the maximal symmetry
  *)

  nearest = 100; (* set a big initial distance *)
  numSym = 0;

  Do[
    (* for each entry in the cousins list *)
    dist = 0;
    numUp = 0;
    numDn = 0;
    numbUp = 0;
    numbDn = 0;
    mostSym = -1;
    Do[
      (* for each index *)
      If[ MemberQ[ {up,cup,pup}, cousins[[a]][[b]] ], numUp += 1];
      If[ MemberQ[ {dn,cdn,pdn}, cousins[[a]][[b]] ], numDn += 1];
      If[ cousins[[a]][[b]] === bdn , numbDn += 1];
      If[ cousins[[a]][[b]] === bup , numbUp += 1];
      (*
       * determine a "distance" of the cousin to the object.
       * Make tensor <-> tetrad conversion more expensive
       * than staying within the same formalism
       *)
      If[ cousins[[a]][[b]] =!= object[[b]], 
         If[ (MemberQ[ {bup,bdn}, cousins[[a]][[b]] ] &&
	     !MemberQ[ {bup,bdn}, object[[b]] ] )
             || (!MemberQ[ {bup,bdn}, cousins[[a]][[b]] ] &&
	          MemberQ[ {bup,bdn}, object[[b]] ])  ,
	     (* tensor <-> tetrad *)
	         dist += 2
             ,(*else*)
	         dist += 1
         ];
      ];
    ,{b, Length[ object]}];
    (* is this one the nearest ? *)
    If[ dist < nearest ,
       nearest = dist;
       nearObj = a;
    ];
    (* is this the most symmetric ? *)
    If[ numUp > numSym ,
       numSym = numUp;
       mostSym = a;
    ,(*else*) If[ numDn > numSym ,
       numSym = numDn;
       mostSym = a;
     ];
    ],
   {a, Length[cousins]}
  ];


  If[ dist === 0 ,
     Write[ grt`stream, "Asked to define an existing object", object];
     Abort[];
  ];
 If[ grDebug, Print["nearest = ",cousins[[nearObj]] ," mostSym = ",
	cousins[[mostSym]] ] ];
 (*
  * BUILD EXPRESSION
  *
  * now build an expression to calculate the new object.
  * One of the grtF`calcsumN routines will be used to iterate this expr.
  *
  *)
 (*
  * for the purposes of index conversion pup,cup -> up etc.
  *)
  realList = newList;
  newList = ReplaceAll[ newList, {pup->up,cup->up,pdn->dn,cdn->dn}];
  realCousins = cousins;
  cousins = ReplaceAll[ cousins, {pup->up,cup->up,pdn->dn,cdn->dn}];
  indexStr = "";
  dependSet = {};
  exprStr = "Hold[";
  b = 1; (* b keeps track of the number of the dummy index *)
  Do[
    (* seperate subsequent indices with a comma *)
    If[ a =!= 1, indexStr = SequenceForm[ indexStr,","] ];
    If[ newList[[a]] =!= cousins[[nearObj]][[a]] ,
     (*
      * find out what kind of term is needed
      * Note: can't assume up,dn pairs since could be up tdn() etc.
      * use Hold[] to prevent evaluation of the metric name. Want this
      * assigned during the calculation, not now!! This means we have to
      * build the expression as a string first.
      *)
      Which[
       newList[[a]] === up && cousins[[nearObj]][[a]] === dn,
	 exprStr = SequenceForm[ exprStr,
			"grtG`gupup[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet, { g[up,up] }],

       newList[[a]] === dn && cousins[[nearObj]][[a]] === up ,
	 exprStr = SequenceForm[ exprStr,
			"grtG`gdndn[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet,{ g[dn,dn] }],

       newList[[a]] === bup && cousins[[nearObj]][[a]] === up,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebupdn[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet, { e[bup,dn] }],

       newList[[a]] === bup && cousins[[nearObj]][[a]] === dn ,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebupup[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet,{ e[bup,up] }],

       newList[[a]] === bdn && cousins[[nearObj]][[a]] === up,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebdndn[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet, { e[bdn,dn] }],

       newList[[a]] === bdn && cousins[[nearObj]][[a]] === dn ,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebdnup[grtG`metricName,grtG`a",a,",grtG`s",b, "] "];
	 dependSet = Union[ dependSet,{ e[bdn,up] }],

       newList[[a]] === up && cousins[[nearObj]][[a]] === bup,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebdnup[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet, { e[bdn,up] }],

       newList[[a]] === up && cousins[[nearObj]][[a]] === bdn ,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebupup[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet,{ e[bup,up] }],

       newList[[a]] === dn && cousins[[nearObj]][[a]] === bup,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebdndn[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet, { e[bdn,dn] }],

       newList[[a]] === dn && cousins[[nearObj]][[a]] === bdn ,
	 exprStr = SequenceForm[ exprStr,
			"grtG`ebupdn[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet,{ e[bup,dn] }],
       
       newList[[a]] === bdn && cousins[[nearObj]][[a]] === bup,
	 exprStr = SequenceForm[ exprStr,
		"grtG`etabdnbdn[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet, { eta[bdn,bdn] }],

       newList[[a]] === bup && cousins[[nearObj]][[a]] === bdn ,
	 exprStr = SequenceForm[ exprStr,
		"grtG`etabupbup[grtG`metricName,grtG`s",b,",grtG`a",a, "] "];
	 dependSet = Union[ dependSet,{ eta[bup,bup] }],
       
       True,
	Write[ grt`stream, "cannot convert index difference"];

      ]; (* end Which[] *)

      (* add a summation index to the index list for objects *)
      indexStr = SequenceForm[ indexStr, "grtG`s", b ];
      b += 1; (* advance the dummy index *)

    ,(*else*)
      (* add a listing index to the base object *)
      expr[a] = 1;
      indexStr = SequenceForm[ indexStr, "grtG`a", a ];
    ],  (* end If[] *)
    {a, Length[ object]}
  ];

  exprStr = ToString[ SequenceForm[ exprStr," ",
			grtG`ObjDef[ realCousins[[nearObj]] ][grtC`root],
			"[grtG`metricName,", indexStr,"] ]" ]];

  grtG`ObjDef[object][grtC`calcFnParms] = ToExpression[ exprStr];


  (*
   * that's most of the hard work done, now update other fields of the
   * object def record
   *
   * build the root name (from the object concatenated with the
   * parameters)
   *)
  newRoot = StringReplace[ ToString[newRoot], "grt`" -> ""];
  grtG`ObjDef[object][grtC`root] = ToExpression[ newRoot];
  grtG`ObjDef[object][grtC`header] = ToString[object];
  grtG`ObjDef[object][grtC`rootStr] = grtG`ObjDef[ realCousins[[nearObj]] ][grtC`rootStr];
  grtG`ObjDef[object][grtC`indexList] = realList;
  grtG`ObjDef[object][grtC`calcFn] = ToExpression[ "grtF`calcsum" <>
		ToString[ b-1 ] ];
  grtG`ObjDef[object][grtC`depends] = Union[ {realCousins[[nearObj]]}, dependSet];

  grtG`ObjDef[object][grtC`symmetry] =
	 grtF`newSymmetry[ object, realCousins[[mostSym]] ];

  If[ grt`OptionVerbose,
    (*
     * not a message, since General::stop is a pain (but I
     * don't really want to turn it off in general
     *)
     Print[ "Created a definition for ", object];
  ];
  Return[];

]

(*----------------------------------------------------------
* grtF`newSymmetry[ newObject,oldObject]
*
* Determine the symmetry of a new object which is based on shifting
* indices of oldObject.
*
*----------------------------------------------------------*)


grtF`newSymmetry[newObject_, oldObject_] :=
 Module[ { a, b, oldPerms, newPerm, permList, okay, thisPerm},


 (* if number of indices < 2 then just copy sym from oldObject *)
 If[ Length[newObject] < 2 ,
   Return[ grtG`ObjDef[oldObject][grtC`symmetry] ];
 ];

 (* get the set of perms for the oldObject *)
 oldPerms = grtG`symmetry[ grtG`ObjDef[oldObject][grtC`symmetry] ];

 If[ grDebug, Print["Entered newSymmetry ", oldPerms] ];

 (* which index positions are different types in new object ? *)
 newPerm = {};

 (*
  * permList is of the form:
  *    { {{2,1,3,4},-1}, {{1,2,4,3},-1},{{3,4,1,2},1} };
  *
  * Loop over each entry in the old Perm list
  *)
 If[ oldPerms =!= {{}}, 
  Do[
   permList = oldPerms[[a]][[1]]; (* get a perm list entry *)
   okay = True;
   (*
    * For entry in a given Perm: i.e. for {2,1,3,4}
    *)
   Do[
       (*
	* if the index permuted to this position is a different type
	* from the index in this position then must reject this perm
	*)
        
       If[ newObject[[b]] =!= newObject[[ permList[[b]] ]] ,
	    okay = False
       ],
    { b, Length[ permList]}
   ];
   If[ okay ,
      newPerm = Union[ newPerm, { oldPerms[[a]] }];
   ],
  {a, Length[ oldPerms]}
  ];
 ];

 (* now find a symmetry function (RETURN this value) *)

 If[ grDebug, Print["Exiting newSymmetry"] ];

 Return[ grtF`findSymmetry[ newPerm, Length[newObject] ] ];

]

