(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"];

(* ----------------------------- *)
(*  Debever connection forms, sigma. *)
(* ----------------------------- *)
 ObjDef[grt`DCF[bup,bup]][grtC`header] = "Debever connection forms, sigma(a,b)";
 ObjDef[grt`DCF[bup,bup]][grtC`root] = grtG`DCFbupbup;
 ObjDef[grt`DCF[bup,bup]][grtC`rootStr] = "sigma";
 ObjDef[grt`DCF[bup,bup]][grtC`indexList] = {bup,bup};
 ObjDef[grt`DCF[bup,bup]][grtC`calcFn] = grtF`calcDCF;
 ObjDef[grt`DCF[bup,bup]][grtC`preCalcFn] = grtF`precalcDCF;
 ObjDef[grt`DCF[bup,bup]][grtC`calcFnParms] = {};
 ObjDef[grt`DCF[bup,bup]][grtC`symmetry] = grtF`symnosym2;
 ObjDef[grt`DCF[bup,bup]][grtC`depends] = { grt`Cnu,grt`Ctau,
     grt`Cgamma,grt`Clambda,grt`Crho,grt`Calpha,grt`Cmu,grt`Csigma,
     grt`Cbeta,grt`Cpi,grt`Ckappa,grt`Cepsilon };

grtF`precalcDCF[ object_] :=
Module[ {a},
	DCFbupbup[metricName, 1, 1] = -2*NPnu[metricName];
	DCFbupbup[metricName, 2, 1] = -2*NPtau[metricName];
	DCFbupbup[metricName, 3, 1] = -2*NPgamma[metricName];
	DCFbupbup[metricName, 1, 3] =  2*NPlambda[metricName];
	DCFbupbup[metricName, 2, 3] =  2*NPrho[metricName];
	DCFbupbup[metricName, 3, 3] =  2*NPalpha[metricName];
	DCFbupbup[metricName, 1, 4] =  2*NPmu[metricName];
	DCFbupbup[metricName, 2, 4] =  2*NPsigma[metricName];
	DCFbupbup[metricName, 3, 4] =  2*NPbeta[metricName];
	DCFbupbup[metricName, 1, 2] = -2*NPpi[metricName];
	DCFbupbup[metricName, 2, 2] = -2*NPkappa[metricName];
	DCFbupbup[metricName, 3, 2] = -2*NPepsilon[metricName];
	For[ a = 1, a <=  4 , a++ ,
		DCFbupbup[metricName,4,a] = 0;
	]; (* e2 For *)
	DCFbupbup[metricName,0,0] = True;
]; (* e2 Module *)

grtF`calcDCF[ object_,index_] :=
Module[ {s},
	s = DCFbupbup[metricName,a1,a2];
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Basis projection of Debever connection form 1. *)
(* ----------------------------- *)
 ObjDef[grt`DCFp[dn]][grtC`header] = "Debever connection form, p";
 ObjDef[grt`DCFp[dn]][grtC`root] = grtG`DCFpdn;
 ObjDef[grt`DCFp[dn]][grtC`rootStr] = "p ";
 ObjDef[grt`DCFp[dn]][grtC`indexList] = {dn};
 ObjDef[grt`DCFp[dn]][grtC`calcFn] = grtF`calcDCFpdn;
 ObjDef[grt`DCFp[dn]][grtC`calcFnParms] = {};
 ObjDef[grt`DCFp[dn]][grtC`symmetry] = grtF`symvector;
 ObjDef[grt`DCFp[dn]][grtC`depends] = { grt`e[bdn,dn], grt`DCF[bup,bup] };

grtF`calcDCFpdn[object_, index_] :=
Module[ {a, s},
	s = 0;
	For[ a = 1, a <=  Ndim[metricName] , a++ ,
		s = s + DCFbupbup[metricName,1,a]*ebdndn[metricName,a,a1];
	]; (* e2 For *)
  Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Basis projection of Debever connection form 2. *)
(* ----------------------------- *)
 ObjDef[grt`DCFq[dn]][grtC`header] = "Debever connection form, q";
 ObjDef[grt`DCFq[dn]][grtC`root] = grtG`DCFqdn;
 ObjDef[grt`DCFq[dn]][grtC`rootStr] = "q ";
 ObjDef[grt`DCFq[dn]][grtC`indexList] = {dn};
 ObjDef[grt`DCFq[dn]][grtC`calcFn] = grtF`calcDCFqdn;
 ObjDef[grt`DCFq[dn]][grtC`calcFnParms] = {};
 ObjDef[grt`DCFq[dn]][grtC`symmetry] = grtF`symvector;
 ObjDef[grt`DCFq[dn]][grtC`depends] = { grt`e[bdn,dn], grt`DCF[bup,bup] };

grtF`calcDCFqdn[ object_,index_] :=
Module[ {	a, s},
	s = 0;
	For[ a = 1, a <=  Ndim[metricName] , a++ ,
		s = s + DCFbupbup[metricName,2,a]*ebdndn[metricName,a,a1];
	]; (* e2 For *)
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Basis projection of Debever connection form 3. *)
(* ----------------------------- *)
 ObjDef[grt`DCFr[dn]][grtC`header] = "Debever connection form, r";
 ObjDef[grt`DCFr[dn]][grtC`root] = grtG`DCFrdn;
 ObjDef[grt`DCFr[dn]][grtC`rootStr] = "r ";
 ObjDef[grt`DCFr[dn]][grtC`indexList] = {dn};
 ObjDef[grt`DCFr[dn]][grtC`calcFn] = grtF`calcDCFrdn;
 ObjDef[grt`DCFr[dn]][grtC`calcFnParms] = {};
 ObjDef[grt`DCFr[dn]][grtC`symmetry] = grtF`symvector;
 ObjDef[grt`DCFr[dn]][grtC`depends] = { grt`e[bdn,dn], grt`DCF[bup,bup] };

grtF`calcDCFrdn[ object_,index_] := 
Module[ {a, s},
	s = 0;
	For[ a = 1, a <=  Ndim[metricName] , a++ ,
		s = s + DCFbupbup[metricName,3,a]*ebdndn[metricName,a,a1];
	]; (* e2 For *)
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Psi0 *)
(* ----------------------------- *)
 ObjDef[grt`CPsi0][grtC`header] = "Weyl Scalar, Psi0";
 ObjDef[grt`CPsi0][grtC`displayName] = grt`Psi0;
 ObjDef[grt`CPsi0][grtC`root] = grtG`Psi0;
 ObjDef[grt`CPsi0][grtC`rootStr] = "Psi0";
 ObjDef[grt`CPsi0][grtC`indexList] = {};
 ObjDef[grt`CPsi0][grtC`calcFn] = grtF`calcCPsi0;
 ObjDef[grt`CPsi0][grtC`calcFnParms] = {};
 ObjDef[grt`CPsi0][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPsi0][grtC`depends] = { grt`Cbeta, grt`Ckappa, 
	grt`Cepsilon, grt`Csigma, grt`DCFq[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPsi0[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*
			  DCFqdnpdn[metricName,i,j]*ebdndn[metricName,1,k]*
			  ebdndn[metricName,3,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPbeta[metricName]*NPkappa[metricName] - 
	    NPepsilon[metricName]*NPsigma[metricName])*2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Psi1 *)
(* ----------------------------- *)
 ObjDef[grt`CPsi1][grtC`header] = "Weyl Scalar, Psi1";
 ObjDef[grt`CPsi1][grtC`displayName] = grt`Psi1;
 ObjDef[grt`CPsi1][grtC`root] = grtG`Psi1;
 ObjDef[grt`CPsi1][grtC`rootStr] = "Psi1";
 ObjDef[grt`CPsi1][grtC`indexList] = {};
 ObjDef[grt`CPsi1][grtC`calcFn] = grtF`calcCPsi1;
 ObjDef[grt`CPsi1][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPsi1][grtC`depends] = { grt`Cpi, grt`Csigma, grt`Ckappa, grt`Cmu, 
	grt`DCFr[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPsi1[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) 
		For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*
			DCFrdnpdn[metricName,i,j]*ebdndn[metricName,1,k]*
			ebdndn[metricName,3,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 - NPpi[metricName]*NPsigma[metricName] + 
	    NPkappa[metricName]*NPmu[metricName];
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Psi2 *)
(* ----------------------------- *)
 ObjDef[grt`CPsi2][grtC`header] = "Weyl Scalar, Psi2";
 ObjDef[grt`CPsi2][grtC`displayName] = grt`Psi2;
 ObjDef[grt`CPsi2][grtC`root] = grtG`Psi2;
 ObjDef[grt`CPsi2][grtC`rootStr] = "Psi2";
 ObjDef[grt`CPsi2][grtC`indexList] = {};
 ObjDef[grt`CPsi2][grtC`calcFn] = grtF`calcCPsi2;
 ObjDef[grt`CPsi2][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPsi2][grtC`depends] = { grt`Cmu, grt`Crho, grt`Clambda, grt`Csigma, 
	grt`Cpi, grt`Ctau, grt`Ckappa, grt`Cnu, 
	grt`Cgamma, grt`Calpha, grt`DCFq[dn,pdn], grt`detb, 
	grt`LevCS[up,up,up,up], grt`DCFr[dn,pdn]  };
grtF`calcCPsi2[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*(
			DCFrdnpdn[metricName,i,j]*ebdndn[metricName,1,k]*ebdndn[metricName,2,l] -
			DCFrdnpdn[metricName,i,j]*ebdndn[metricName,3,k]*ebdndn[metricName,4,l] +
			DCFqdnpdn[metricName,i,j]*ebdndn[metricName,4,k]*ebdndn[metricName,2,l] );

	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/6 + (NPmu[metricName]*NPrho[metricName] - 
	    NPlambda[metricName]*NPsigma[metricName] - 
	    NPpi[metricName]*NPtau[metricName] + 
	    NPkappa[metricName]*NPnu[metricName] + 
	    2*NPgamma[metricName]*NPrho[metricName] - 
	    2*NPalpha[metricName]*NPtau[metricName] )/3;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Psi3 *)
(* ----------------------------- *)
 ObjDef[grt`CPsi3][grtC`header] = "Weyl Scalar, Psi3";
 ObjDef[grt`CPsi3][grtC`displayName] = grt`Psi3;
 ObjDef[grt`CPsi3][grtC`root] = grtG`Psi3;
 ObjDef[grt`CPsi3][grtC`rootStr] = "Psi3";
 ObjDef[grt`CPsi3][grtC`indexList] = {};
 ObjDef[grt`CPsi3][grtC`calcFn] = grtF`calcCPsi3;
 ObjDef[grt`CPsi3][grtC`calcFnParms] = {};
 ObjDef[grt`CPsi3][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPsi3][grtC`depends] = { grt`Clambda, grt`Ctau, grt`Cnu, grt`Crho, 
   grt`DCFr[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPsi3[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFrdnpdn[metricName,i,j]*ebdndn[metricName,4,k]*
			ebdndn[metricName,2,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 - NPlambda[metricName]*NPtau[metricName] + NPnu[metricName]*NPrho[metricName];
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Psi4 *)
(* ----------------------------- *)
 ObjDef[grt`CPsi4][grtC`header] = "Weyl Scalar, Psi4";
 ObjDef[grt`CPsi4][grtC`displayName] = grt`Psi4;
 ObjDef[grt`CPsi4][grtC`root] = grtG`Psi4;
 ObjDef[grt`CPsi4][grtC`rootStr] = "Psi4";
 ObjDef[grt`CPsi4][grtC`indexList] = {};
 ObjDef[grt`CPsi4][grtC`calcFn] = grtF`calcCPsi4;
 ObjDef[grt`CPsi4][grtC`calcFnParms] = {};
 ObjDef[grt`CPsi4][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPsi4][grtC`depends] = { grt`Calpha, grt`Cnu, grt`Cgamma, grt`Clambda, 
	grt`DCFp[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPsi4[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFpdnpdn[metricName,i,j]*ebdndn[metricName,4,k]*
			ebdndn[metricName,2,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPalpha[metricName]*NPnu[metricName] - NPgamma[metricName]*NPlambda[metricName])*2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi00 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi00][grtC`header] = "Ricci Scalar, Phi00";
 ObjDef[grt`CPhi00][grtC`displayName] = grt`Phi00;
 ObjDef[grt`CPhi00][grtC`root] = grtG`NPPhi00;
 ObjDef[grt`CPhi00][grtC`rootStr] = "Phi00";
 ObjDef[grt`CPhi00][grtC`indexList] = {};
 ObjDef[grt`CPhi00][grtC`calcFn] = grtF`calcCPhi00;
 ObjDef[grt`CPhi00][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi00][grtC`depends] = { grt`Calpha, grt`Ckappa, grt`Cepsilon, 
	grt`Crho, grt`DCFq[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi00[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) 
	      For[ k = 1, k <=  Ndim[metricName] , k++ ,
	        If[ k=!=i && k=!=j , (* then *) 
		  For[ l = 1, l <=  Ndim[metricName] , l++ ,
	            If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*
			  DCFqdnpdn[metricName,i,j]*ebdndn[metricName,4,k]*
			ebdndn[metricName,1,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPalpha[metricName]*NPkappa[metricName] - 
		NPepsilon[metricName]*NPrho[metricName])*2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi01 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi01][grtC`header] = "Ricci Scalar, Phi01";
 ObjDef[grt`CPhi01][grtC`displayName] = grt`Phi01;
 ObjDef[grt`CPhi01][grtC`root] = grtG`NPPhi01;
 ObjDef[grt`CPhi01][grtC`rootStr] = "Phi01";
 ObjDef[grt`CPhi01][grtC`indexList] = {};
 ObjDef[grt`CPhi01][grtC`calcFn] = grtF`calcCPhi01;
 ObjDef[grt`CPhi01][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi01][grtC`depends] = { grt`Ckappa, grt`Cepsilon, grt`Cgamma, 
	grt`Ctau, grt`DCFq[dn,pdn], grt`CPsi1, grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi01[object_, index_] := 
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFqdnpdn[metricName,i,j]*ebdndn[metricName,3,k]*
			ebdndn[metricName,4,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = -detb[metricName]*s/2 + (NPgamma[metricName]*NPkappa[metricName] - 
	     NPepsilon[metricName]*NPtau[metricName])*2 - Psi1[metricName];
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi02 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi02][grtC`header] = "Ricci Scalar, Phi02";
 ObjDef[grt`CPhi02][grtC`displayName] = grt`Phi02;
 ObjDef[grt`CPhi02][grtC`root] = grtG`NPPhi02;
 ObjDef[grt`CPhi02][grtC`rootStr] = "Phi02";
 ObjDef[grt`CPhi02][grtC`indexList] = {};
 ObjDef[grt`CPhi02][grtC`calcFn] = grtF`calcCPhi02;
 ObjDef[grt`CPhi02][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi02][grtC`depends] = { grt`Cgamma, grt`Csigma, grt`Cbeta, 
	grt`Ctau, grt`DCFq[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi02[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFqdnpdn[metricName,i,j]*ebdndn[metricName,2,k]*
			ebdndn[metricName,3,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPgamma[metricName]*NPsigma[metricName] - NPbeta[metricName]*NPtau[metricName])*2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi11 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi11][grtC`header] = "Ricci Scalar, Phi11";
 ObjDef[grt`CPhi11][grtC`displayName] = grt`Phi11;
 ObjDef[grt`CPhi11][grtC`root] = grtG`NPPhi11;
 ObjDef[grt`CPhi11][grtC`rootStr] = "Phi11";
 ObjDef[grt`CPhi11][grtC`indexList] = {};
 ObjDef[grt`CPhi11][grtC`calcFn] = grtF`calcCPhi11;
 ObjDef[grt`CPhi11][grtC`calcFnParms] = {};
 ObjDef[grt`CPhi11][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi11][grtC`depends] = { grt`Cpi, grt`Ctau, grt`Ckappa, grt`Cnu, 
	grt`Crho, grt`Cmu, grt`Clambda, grt`Csigma, grt`DCFr[dn,pdn], 
	grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi11[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) 
		For[ l = 1, l <=  Ndim[metricName] , l++ ,
	         If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*
			  DCFrdnpdn[metricName,i,j]*
			( ebdndn[metricName,3,k]*ebdndn[metricName,4,l] + 
			  ebdndn[metricName,1,k]*ebdndn[metricName,2,l] );
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = -detb[metricName]*s/4 - ( NPpi[metricName]*NPtau[metricName] - 
	    NPkappa[metricName]*NPnu[metricName] + 
	    NPrho[metricName]*NPmu[metricName] - 
	    NPlambda[metricName]*NPsigma[metricName] )/2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi12 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi12][grtC`header] = "Ricci Scalar, Phi12";
 ObjDef[grt`CPhi12][grtC`displayName] = grt`Phi12;
 ObjDef[grt`CPhi12][grtC`root] = grtG`NPPhi12;
 ObjDef[grt`CPhi12][grtC`rootStr] = "Phi12";
 ObjDef[grt`CPhi12][grtC`indexList] = {};
 ObjDef[grt`CPhi12][grtC`calcFn] = grtF`calcCPhi12;
 ObjDef[grt`CPhi12][grtC`calcFnParms] = {};
 ObjDef[grt`CPhi12][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi12][grtC`depends] = { grt`Cnu, grt`Csigma, grt`Cmu, 
	grt`Ctau, grt`DCFr[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi12[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFrdnpdn[metricName,i,j]*ebdndn[metricName,2,k]*
			ebdndn[metricName,3,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPnu[metricName]*NPsigma[metricName] - NPmu[metricName]*NPtau[metricName]); (*  *2 on last term *)
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: Phi22 *)
(* ----------------------------- *)
 ObjDef[grt`CPhi22][grtC`header] = "Ricci Scalar, Phi22";
 ObjDef[grt`CPhi22][grtC`displayName] = grt`Phi22;
 ObjDef[grt`CPhi22][grtC`root] = grtG`NPPhi22;
 ObjDef[grt`CPhi22][grtC`rootStr] = "Phi22";
 ObjDef[grt`CPhi22][grtC`indexList] = {};
 ObjDef[grt`CPhi22][grtC`calcFn] = grtF`calcCPhi22;
 ObjDef[grt`CPhi22][grtC`calcFnParms] = {};
 ObjDef[grt`CPhi22][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CPhi22][grtC`depends] = { grt`Cnu, grt`Cbeta, grt`Cmu, 
	grt`Cgamma, grt`DCFp[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };
grtF`calcCPhi22[object_, index_] := 
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*DCFpdnpdn[metricName,i,j]*ebdndn[metricName,2,k]*
			ebdndn[metricName,3,l];
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = detb[metricName]*s/2 + (NPnu[metricName]*NPbeta[metricName] - NPmu[metricName]*NPgamma[metricName])*2;
Return[s];
]; (* e2 Module *)

(* ----------------------------- *)
(*  Curvature coefficient: CLambda *)
(* ----------------------------- *)

 ObjDef[grt`CLambda][grtC`header] = "Lambda = Ricci Scalar/24";
 ObjDef[grt`CLambda][grtC`displayName] = grt`Lambda;
 ObjDef[grt`CLambda][grtC`root] = grtG`NPLambda;
 ObjDef[grt`CLambda][grtC`rootStr] = "Lambda";
 ObjDef[grt`CLambda][grtC`indexList] = {};
 ObjDef[grt`CLambda][grtC`calcFn] = grtF`calcDRicci;
 ObjDef[grt`CLambda][grtC`calcFnParms] = {};
 ObjDef[grt`CLambda][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`CLambda][grtC`depends] = { grt`Cpi, grt`Ctau, grt`Ckappa, 
	grt`Cnu, grt`Crho, grt`Cmu, grt`Clambda, grt`Csigma, 
	grt`Cgamma, grt`Calpha, grt`DCFr[dn,pdn], 
	grt`DCFq[dn,pdn], grt`detb, grt`LevCS[up,up,up,up] };

grtF`calcDRicci[object_, index_] :=
Module[ { i,j,k,l,s},
	s = 0;
	For[ i = 1, i <=  Ndim[metricName] , i++ ,
	  For[ j = 1, j <=  Ndim[metricName] , j++ ,
	    If[ j=!=i , (* then *) For[ k = 1, k <=  Ndim[metricName] , k++ ,
	      If[ k=!=i && k=!=j , (* then *) For[ l = 1, l <=  Ndim[metricName] , l++ ,
	        If[ l=!=i && l=!=j && l=!=k , (* then *)
	          s = s + LevCSupupupup[metricName,i,j,k,l]*(
			DCFrdnpdn[metricName,i,j]*ebdndn[metricName,3,k]*ebdndn[metricName,4,l] -
			DCFrdnpdn[metricName,i,j]*ebdndn[metricName,1,k]*ebdndn[metricName,2,l] +
		      2*DCFqdnpdn[metricName,i,j]*ebdndn[metricName,4,k]*ebdndn[metricName,2,l] );
	        ]; (* e2 If *)
	      ]; ]; (* e2 If *)
	    ]; ]; (* e2 If *)
	  ]; (* e2 For *)
	]; (* e2 For *)
	s = 2*detb[metricName]*s + (NPpi[metricName]*NPtau[metricName] - NPkappa[metricName]*NPnu[metricName] - NPrho[metricName]*NPmu[metricName] + NPlambda[metricName]*NPsigma[metricName] + (NPgamma[metricName]*NPrho[metricName]-NPalpha[metricName]*NPtau[metricName])*4 )*4;
Return[s/24];
]; (* e2 Module *)

End[];

