(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 
Begin["grtG`"];

(* ---------------------------- *)
(*  Determinant of the basis matrix. *)
(* ---------------------------- *)
 ObjDef[grt`detb][grtC`header] = "Determinant of the basis";
 ObjDef[grt`detb][grtC`root] = grtG`detb;
 ObjDef[grt`detb][grtC`rootStr] = "b ";
 ObjDef[grt`detb][grtC`indexList] = {};
 ObjDef[grt`detb][grtC`preCalcFn] = grtF`calcdetb;
 ObjDef[grt`detb][grtC`symmetry] = grtF`symscalar;
 ObjDef[grt`detb][grtC`depends] = { };

grtF`calcdetb[ object_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ {a, b, A, fn},
	Which[ 
	  grtF`checkIfAssigned[ grt`e[bdn,up]] , 
		fn = grtG`ebdnup,
	  grtF`checkIfAssigned[ grt`e[bup,dn]] , 
		fn = grtG`ebupdn,
	  grtF`checkIfAssigned[ grt`e[bdn,dn]] , 
		grtF`core[ grt`e[bup,dn] ];
		fn = grtG`ebupdn,
	  grtF`checkIfAssigned[ grt`e[bup,up]] , (* then *)
		grtF`core[ grt`e[bdn,up] ];
		fn = grtG`ebdnup,
	  True, 
	       Print[ "detb: A basis must be defined to calculate this object." ];
	       Abort[];
	]; (* end Which*)
	A = Table[ fn[metricName, a, b], 
	    {a,Ndim[metricName]},{b,Ndim[metricName]}]; 
	grtG`detb[grtG`metricName] = Det[ A ];
	If[ fn  ===  grtG`ebupdn , (* then *)
		grtG`detb[grtG`metricName] = 1/grtG`detb[grtG`metricName];
	]; (* end If *)

]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  Jacobi-McLenaghan algorithm for lambda[bdn,bdn,bdn]. *)
(* ------------------------------------------------------------------- *)
(* (  ) *)
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`header] = "pre-Rotation Coefficients";
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`root] = grtG`lambdabdnbdnbdn;
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`rootStr] = "lambda";
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn};
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcClambdabdnbdnbdn;
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`calcFnParms] = {};
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`symmetry] = grtF`sym3a13;
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcClambdabdnbdnbdn;
 ObjDef[grt`Clambda[bdn,bdn,bdn]][grtC`depends] = { grt`e[bup,dn], 
	grt`e[bdn,dn,pdn], grt`LevCS[up,up,up,up], grt`detb };
(* gr[grtC`displayName] := lambda[bdn,bdn,bdn]: *)
(* gr[grtC`useWhen] := grtF`when_Clambdabdnbdnbdn: *)

grtF`whenClambdabdnbdnbdn[ ] :=
Module[ {},
   If[ grtF`checkIfAssigned[ grt`e[bdn,dn]], (* then *)
	Return[True];
   , (* else *)
	Return[False];
   ]; (* end If *)
]; (* end Module *)

grtF`calcClambdabdnbdnbdn[ object_, list_ ] :=
Module[ {a, b, c, d, e, f, s},
  s = 0;
  For[ a = 1, a <=  Ndim[metricName] , a++ ,
    For[ b = 1, b <=  Ndim[metricName] , b++ ,
      If[ b=!=a , (* then *) For[ c = 1, c <=  Ndim[metricName] , c++ ,
        If[ c=!=a && c=!=b , (* then *) For[ d = 1, d <=  Ndim[metricName] , d++ ,
          If[ d=!=a && d=!=b && d=!=c , (* then *) For[ e = 1, e <=  Ndim[metricName] , e++ ,
            If[ e=!=a1 && e=!=a3 , (* then *) For[ f = 1, f <=  Ndim[metricName] , f++ ,
              If[ f=!=e && f=!=a1 && f=!=a3 , (* then *) 
                s = s + grtG`ebdndnpdn[grtG`metricName,a2,a,b]*
		  grtG`LevCSupupupup[grtG`metricName,a,b,c,d]*
		  grtG`LevCSupupupup[grtG`metricName,a1,a3,e,f]*
		  grtG`ebupdn[grtG`metricName,e,c]*grtG`ebupdn[grtG`metricName,f,d];
              ]; (* end If *)
            ]; ]; (* end If *)
          ]; ]; (* end If *)
        ]; ]; (* end If *)
      ]; ]; (* end If *)
    ]; (* end For *)
  ]; (* end For *)
 Return[s*grtG`detb[grtG`metricName]/2];
]; (* end Module *)

(* ------------------------------------------------------------------- *)
(*  Alternate definition for rotation coefficients in terms of Clambda. *)
(* ------------------------------------------------------------------- *)
 ObjDef[grt`Crot[bdn,bdn,bdn]][grtC`displayName] = rot[bdn,bdn,bdn];
 ObjDef[grt`Crot[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcCrotbdnbdnbdn;
 ObjDef[grt`Crot[bdn,bdn,bdn]][grtC`useWhen] = grtF`whenCrotbdnbdnbdn;
 ObjDef[grt`Crot[bdn,bdn,bdn]][grtC`depends] = { grt`Clambda[bdn,bdn,bdn] };

grtF`whenCrotbdnbdnbdn[ ] :=
Module[ {},
    If[ grtF`checkIfAssigned[ grt`Clambda[bdn,bdn,bdn]]  ===  True , (* then *)
	Return[True];
    , (* else *)
	Return[False];
    ]; (* end If *)
]; (* end Module *)

grtF`calcCrotbdnbdnbdn[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { s},
	s = (grtG`Clambdabdnbdnbdn[grtG`metricName,a1,a2,a3]
            + grtG`Clambdabdnbdnbdn[grtG`metricName,a3,a1,a2]
            - grtG`Clambdabdnbdnbdn[grtG`metricName,a2,a3,a1])/2;
Return[s];
]; (* end Module *)

(* ---------------------------- *)
(*  LevCS *)
(* ---------------------------- *)
 ObjDef[grt`LevCS[up,up,up,up]][grtC`header] = "Levi-Civita Symbol";
 ObjDef[grt`LevCS[up,up,up,up]][grtC`root] = grtG`LevCSupupupup;
 ObjDef[grt`LevCS[up,up,up,up]][grtC`rootStr] = "e ";
 ObjDef[grt`LevCS[up,up,up,up]][grtC`indexList] = {up,up,up,up};
 ObjDef[grt`LevCS[up,up,up,up]][grtC`preCalcFn] = grtF`preCalcLevCS;
 ObjDef[grt`LevCS[up,up,up,up]][grtC`symmetry] = grtF`symLevC4;
 ObjDef[grt`LevCS[up,up,up,up]][grtC`depends] = {};


grtF`preCalcLevCS[ object_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
(* globals  grtG`LevCSupupupup_; *)
Module[ { a, master, pList, i, j, k, l},
	If[ Ndim[metricName] =!= 4 , (* then *)
	   Print[ "LevC[dn,dn,dn,dn] may only be calculated for n =  4"];
	   Abort[];
	]; (* end If *)
	For[ i = 1, i <=  4 , i++ ,
	  For[ j = 1, j <=  4 , j++ ,
	    For[ k = 1, k <=  4 , k++ ,
	      For[ l = 1, l <=  4 , l++ ,
		 grtG`LevCSupupupup[grtG`metricName,i,j,k,l] = 0;
	      ]; (* end For *)
	    ]; (* end For *)
	  ]; (* end For *)
	]; (* end For *)
	(*  Brute force *)
(*  *)
(*  built by make_lev tool *)
(*  *)
grtG`LevCSupupupup[grtG`metricName,1,2,4,3] = -1;
grtG`LevCSupupupup[grtG`metricName,1,3,2,4] = -1;
grtG`LevCSupupupup[grtG`metricName,1,3,4,2] =  1;
grtG`LevCSupupupup[grtG`metricName,1,4,2,3] =  1;
grtG`LevCSupupupup[grtG`metricName,1,4,3,2] = -1;
grtG`LevCSupupupup[grtG`metricName,2,1,3,4] = -1;
grtG`LevCSupupupup[grtG`metricName,2,1,4,3] =  1;
grtG`LevCSupupupup[grtG`metricName,2,3,1,4] =  1;
grtG`LevCSupupupup[grtG`metricName,2,3,4,1] = -1;
grtG`LevCSupupupup[grtG`metricName,2,4,1,3] = -1;
grtG`LevCSupupupup[grtG`metricName,2,4,3,1] =  1;
grtG`LevCSupupupup[grtG`metricName,3,1,2,4] =  1;
grtG`LevCSupupupup[grtG`metricName,3,1,4,2] = -1;
grtG`LevCSupupupup[grtG`metricName,3,2,1,4] = -1;
grtG`LevCSupupupup[grtG`metricName,3,2,4,1] =  1;
grtG`LevCSupupupup[grtG`metricName,3,4,1,2] =  1;
grtG`LevCSupupupup[grtG`metricName,3,4,2,1] = -1;
grtG`LevCSupupupup[grtG`metricName,4,1,2,3] = -1;
grtG`LevCSupupupup[grtG`metricName,4,1,3,2] =  1;
grtG`LevCSupupupup[grtG`metricName,4,2,1,3] =  1;
grtG`LevCSupupupup[grtG`metricName,4,2,3,1] = -1;
grtG`LevCSupupupup[grtG`metricName,4,3,1,2] = -1;
grtG`LevCSupupupup[grtG`metricName,4,3,2,1] =  1;

grtG`LevCSupupupup[grtG`metricName,1,2,3,4] = 1;

]; (* end Module *)

End[];

