(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"];

(* ----------------------------- *)
(*  Petrov type using a general tetrad. *)
(* ----------------------------- *)
ObjDef[grt`Petrov][grtC`header] = "Petrov Type";
ObjDef[grt`Petrov][grtC`root] = grtG`Petrov;
ObjDef[grt`Petrov][grtC`rootStr] = "Petrov Type ";
ObjDef[grt`Petrov][grtC`indexList] = {};
ObjDef[grt`Petrov][grtC`calcFn] = grtF`calcPetrov;
ObjDef[grt`Petrov][grtC`symmetry] = grtF`symscalar;
ObjDef[grt`Petrov][grtC`depends] = { grt`Psi0, grt`Psi1, grt`Psi2, 
	grt`Psi3, grt`Psi4 };

grtF`calcPetrov[ object_, index_] :=
Module[ { a, s},

  s = grtF`grpetrov[  Psi0[metricName],
                 Psi1[metricName],
                 Psi2[metricName],
                 Psi3[metricName],
                 Psi4[metricName] ];
  Return[s];
]; (* end Module *)

(* =================================================== *)
(*  Calculation routine *)
(* =================================================== *)

grtF`grpetrov[ Psi0_, Psi1_, Psi2_, Psi3_, Psi4_ ] := 
Module[ {i, mask, Psi, ptype, f1, f2, f3, t1, t2, t3, t4, 
        A, D, E, F, G, H, I, J, Q, S, U, V, W, Z, str, str1, str2, str3, str4},

  grtG`reportSize = 0;
  grtG`reportVsize = 0;

  (* code porting leads to uglyness like this *)
  Psi[0] = Psi0;
  Psi[1] = Psi1;
  Psi[2] = Psi2;
  Psi[3] = Psi3;
  Psi[4] = Psi4;

  mask = 0;
  For[ i = 0, i <=  4 , i++ , 
    str = StringJoin["Weyl scalar Psi", ToString[i]];
    grtF`report[ "", str, Psi[i] ];
    If[ Psi[i] =!= 0 , (* then *) 
      mask = mask + 2^(4-i);
    ]; (* end If *)
  ]; (* end For *)
      
  If[ grDebug, Print["mask = ",mask] ];

  Which[ 
  mask  ===  0 , 
    ptype = "O",
  mask  ===  1 , 
    ptype = "N (or simpler)",
  mask  ===  2 , 
    ptype = "III (or simpler)",
  mask  ===  3 , 
    ptype = "III (or simpler)",
  mask  ===  4 , 
    ptype = "D (or simpler)",
  mask  ===  5 , 
    ptype = "II (or simpler)",
  mask  ===  6 , 
    ptype = "II (or simpler)",
  mask  ===  7 , 
    str = "2*Psi[3]^2-3*Psi[2]*Psi[4]";
    I = grtF`pSimplify[ 2*Psi[3]^2 - 3*Psi[2]*Psi[4] ];
    If[ I  ===  0 , (* then *)
      ptype = "D (or simpler)"; 
    , (* else *)
      ptype = "II (or simpler)";
    ]; (* end If *)
    grtF`report[ "", str, I ],
  mask  ===  8 , (* then *)
    ptype = "III (or simpler)",
  mask  ===  9 , (* then *)
    ptype = "I (or simpler)",
  mask  ===  10 , (* then *)
    ptype = "I (or simpler)",
  mask  ===  11 , (* then *)
    str = "27*Psi[1]*Psi[4]^2 + 64*Psi[3]^3";
    I = grtF`pSimplify[ 27*Psi[1]*Psi[4]^2 + 64*Psi[3]^3 ];
    If[  I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *) 
      ptype = "I (or simpler)";
    ]; (* end If *)
    grtF`report[ "", str, I ],
  mask  ===  12 , (* then *)
    ptype = "II (or simpler)",
  mask  ===  13 , (* then *)
    str = "Psi[1]^2*Psi[4] + 2*Psi[2]^3";
    I = grtF`pSimplify[ Psi[1]^2*Psi[4] + 2*Psi[2]^3 ];
    If[ I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *)
      ptype = "I (or simpler)";
    ]; (* end If *)
    grtF`report[ "", str, I ],
  mask  ===  14 , (* then *)
    str = "9*Psi[2]^2-16*Psi[1]*Psi[3]";
    I = grtF`pSimplify[ 9*Psi[2]^2-16*Psi[1]*Psi[3] ];
    If[ I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *) 
      ptype = "I (or simpler)";
    ]; (* end If *)
    grtF`report[ "", str, I ],
  mask  ===  15 , (* then *)
    str1 = "3*Psi[2]^2 - 4*Psi[1]*Psi[3]";
    str2 = "2*Psi[2]*Psi[3] - 3*Psi[1]*Psi[4]";
    I  = grtF`pSimplify[ 3*Psi[2]^2 - 4*Psi[1]*Psi[3] ];
    f1 = grtF`pSimplify[ 2*Psi[2]*Psi[3] - 3*Psi[1]*Psi[4] ];
    grtF`report[ "", str1, I ];
    grtF`report[ "", str2, f1 ];
    If[ I  ===  0 , (* then *)
      If[ f1  ===  0 , (* then *) 
        ptype = "III (or simpler)";
      , (* else *)
        ptype = "I (or simpler)";
      ]; (* end If *)
    , (* else *)
      If[ f1  ===  0 , (* then *)
        ptype = "I (or simpler)";
      , (* else *)
        str3 = "9*Psi[2]*Psi[4] - 8*Psi[3]^2";
        f2 = grtF`pSimplify[ 9*Psi[2]*Psi[4] - 8*Psi[3]^2 ];
        grtF`report[ "", str3, f2 ];
        If[ f2  ===  0 , (* then *)
          ptype = "I (or simpler)";
        , (* else *)
          str4 = "3*f1^2 + 2*I*f2";
          f3 = grtF`pSimplify[ 3*f1^2 + 2*I*f2 ];
          grtF`report[ "", str4, f3 ];
          If[ f3  ===  0 , (* then *)
            ptype = "II (or simpler)";
          , (* else *)
            ptype = "I (or simpler)";
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ], (* end If *)
  mask  ===  16 , 
    ptype = "N (or simpler)",
  mask  ===  17 , 
    ptype = "I (or simpler)",
  mask  ===  18 ,  
    ptype = "I (or simpler)",
  mask  ===  19 , 
    str = "Psi[0]*Psi[4]^3 - 27*Psi[3]^4";
    I = grtF`pSimplify[ Psi[0]*Psi[4]^3 - 27*Psi[3]^4 ];
    grtF`report[ "", str, I ];
    If[ I  ===  0 , 
      ptype = "II (or simpler)";
    , (* else *) 
      ptype = "I (or simpler)";
    ], (* end If *)
  mask  ===  20 , 
    ptype = "II (or simpler)",
  mask  ===  21 , 
    str = "9*Psi[2]^2 - Psi[0]*Psi[4]";
    I = grtF`pSimplify[ 9*Psi[2]^2 - Psi[0]*Psi[4] ];
    grtF`report[ "", str, I ];
    If[  I  ===  0 , 
      ptype = "D (or simpler)";
    , (* else *)
      ptype = "I (or simpler)";
    ], (* end If *)
  mask  ===  22 , 
    str = "Psi[0]*Psi[4]^2 + 2*Psi[2]^3";
    I = grtF`pSimplify[ Psi[0]*Psi[4]^2 + 2*Psi[2]^3 ];
    grtF`report [ "", str, I ];
    If[ I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *)
      ptype = "I (or simpler)";
    ], (* end If *)
  mask  ===  23 , 
    str1 = "Psi[0]*Psi[4] + 3*Psi[2]^2";
    str2 = "4*Psi[2]*Psi[4] - 3*Psi[3]^2";
    I = grtF`pSimplify[ Psi[0]*Psi[4] + 3*Psi[2]^2 ];
    J = grtF`pSimplify[ 4*Psi[2]*Psi[4] - 3*Psi[3]^2 ];
    grtF`report[ " I", str1, I ];
    grtF`report[ " J", str2, J ];
    If[ I  ===  0 , (* then *)
      If[ J  ===  0 , (* then *)
        ptype = "III (or simpler)";
      , (* else *)
        ptype = "I (or simpler)";
      ]; (* end If *)
    , (* else *) 
      If[ J  ===  0 , (* then *) 
        ptype = "I (or simpler)";
      , (* else *) 
        str3 = "Psi[0]*J - 2*Psi[2]*I";
	f3 = grtF`pSimplify[ Psi[0]*J - 2*Psi[2]*I ];
        grtF`report[ " f3", str3, f3 ];
        If[ f3  ===  0 , (* then *)
          ptype = "I (or simpler)";
        , (* else *)
          str4 = "Psi[4]*I^2 - 3*J*f3";
          D = grtF`pSimplify[ Psi[4]*I^2 - 3*J*f3 ];
          grtF`report[ " D", str4, D ];
          If[ D  ===  0 , (* then *)
             ptype = "II (or simpler)";
          , (* else *)
             ptype = "I (or simpler)";
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ], (* end If *)
  mask  ===  24 , 
    ptype = "III (or simpler)",
  mask  ===  25 , 
    str = "Psi[4]*Psi[0]^3 - 27*Psi[1]^4";
    I = grtF`pSimplify[ Psi[4]*Psi[0]^3 - 27*Psi[1]^4 ];
    grtF`report[ "", str, I ];
    If[  I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *) 
      ptype = "I (or simpler)";
    ], (* end If *)
  mask  ===  26 , 
    str = "27*Psi[3]*Psi[0]^2 + 64*Psi[1]^3";
    I = grtF`pSimplify[ 27*Psi[3]*Psi[0]^2 + 64*Psi[1]^3 ];
    grtF`report[ "", str, I ];
    If[ I  ===  0 , (* then *)
      ptype = "II (or simpler)";
    , (* else *) 
      ptype = "I (or simpler)";
    ], (* end If *)
  mask  ===  27 , (* then *)
    t1 = grtF`pSimplify[ Psi[0]*Psi[4] ];
    t2 = grtF`pSimplify[ Psi[1]*Psi[3] ];
    t3 = grtF`pSimplify[ Psi[0]*Psi[3]^2 ];
    t4 = grtF`pSimplify[ Psi[4]*Psi[1]^2 ];
    str = "Psi[0]*Psi[3]^2 -  Psi[4]*Psi[1]^2";
    V  = grtF`pSimplify[ t3 - t4 ];
    grtF`report[ " V", str, V ];
    If[ V  ===  0 , (* then *)
      str = "Psi[0]*Psi[4] + 2*Psi[1]*Psi[3]";
      U = grtF`pSimplify[ t1 + 2*t2 ];
      grtF`report[ " U", str, U ];
      If[ U  ===  0 , (* then *)
        ptype = "D (or simpler)";
      , (* else *)
        str = "Psi[0]*Psi[4] - 16*Psi[1]*Psi[3]";
        W = grtF`pSimplify[ t1 - 16*t2 ];
        grtF`report[ " W", str, W ];
        If[ W  ===  0 , (* then *)
          ptype = "II (or simpler)";
        , (* else *)
          ptype = "I (or simpler)";
        ]; (* end If *)
      ]; (* end If *)
    , (* else *) 
      str = "Psi[0]*Psi[4] - 4*Psi[1]*Psi[3]";
      I = grtF`pSimplify[ t1 - 4*t2 ];
      grtF`report[ " I", str, I ];
      str = "-Psi[0]*Psi[3]^2 - Psi[4]*Psi[1]^2";
      J = grtF`pSimplify[ -t3 - t4 ];
      grtF`report[ " J", str, J ];
      If[ I  ===  0 , (* then *)
        If[ J  ===  0 , (* then *)
          ptype = "III (or simpler)";
        , (* else *)
          ptype = "I (or simpler)";
        ]; (* end If *)
      , (* else *)
        If[ J  ===  0 , (* then *)
          ptype = "I (or simpler)";
        , (* else *)
          str = "I^3 - 27*J^2";
          D = grtF`pSimplify[ I^3 - 27*J^2 ];
          grtF`report[ " D", str, D ];
          If[ D  ===  0 , (* then *) 
            ptype = "II (or simpler)";
          , (* else *)
            ptype = "I (or simpler)";
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ], (* end If *)
  mask  ===  28 , (* then *)
    str = "2*Psi[1]^2 - 3*Psi[2]*Psi[0]";
    I = grtF`pSimplify[ 2*Psi[1]^2 - 3*Psi[2]*Psi[0]];
    grtF`report[ "", str, I ];
    If[ I  ===  0 , (* then *)
      ptype = "D (or simpler)";
    , (* else *)
      ptype = "II (or simpler)";
    ], (* end If *)
  mask  ===  29 , (* then *)
    str = "Psi[4]*Psi[0] + 3*Psi[2]^2";
    I = grtF`pSimplify[ Psi[4]*Psi[0] + 3*Psi[2]^2 ];
    grtF`report[ " I", str, I ];
    str = "4*Psi[2]*Psi[0] - 3*Psi[1]^2";
    J = grtF`pSimplify[ 4*Psi[2]*Psi[0] - 3*Psi[1]^2 ];
    grtF`report[ " J", str, J ];
    If[ I  ===  0 , (* then *)
      If[ J  ===  0 , (* then *)
        ptype = "III (or simpler)";
      , (* else *)
        ptype = "I (or simpler)";
      ]; (* end If *)
    , (* else *) 
      If[ J  ===  0 , (* then *) 
        ptype = "I (or simpler)";
      , (* else *) 
        str = "Psi[4]*J - 2*Psi[2]*I";
	f3 = grtF`pSimplify[ Psi[4]*J - 2*Psi[2]*I ];
        grtF`report[ " f3", str, f3 ];
        If[ f3  ===  0 , (* then *)
          ptype = "I (or simpler)";
        , (* else *)
          str = "Psi[0]*I^2 - 3*J*f3";
          D = grtF`pSimplify[ Psi[0]*I^2 - 3*J*f3 ];
          grtF`report[ " D", str, D ];
          If[ D  ===  0 , (* then *)
             ptype = "II (or simpler)";
          , (* else *)
             ptype = "I (or simpler)";
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ], (* end If *)
  mask  ===  30 , (* then *)
    str = "3*Psi[2]^2 - 4*Psi[1]*Psi[3]";
    I  = grtF`pSimplify[ 3*Psi[2]^2 - 4*Psi[1]*Psi[3] ];
    grtF`report[ " I", str, I ];
    str = "2*Psi[2]*Psi[1] - 3*Psi[3]*Psi[0]";
    f1 = grtF`pSimplify[ 2*Psi[2]*Psi[1] - 3*Psi[3]*Psi[0] ];
    grtF`report[ " f1", str, f1 ];
    If[ I  ===  0 , (* then *)
      If[ f1  ===  0 , (* then *) 
        ptype = "III (or simpler)";
      , (* else *)
        ptype = "I (or simpler)";
      ]; (* end If *)
    , (* else *)
      If[ f1  ===  0 , (* then *)
        ptype = "I (or simpler)";
      , (* else *)
        str = "9*Psi[2]*Psi[0] - 8*Psi[1]^2";
        f2 = grtF`pSimplify[ 9*Psi[2]*Psi[0] - 8*Psi[1]^2 ];
        grtF`report[ " f2", str, f2 ];
        If[ f2  ===  0 , (* then *)
          ptype = "I (or simpler)";
        , (* else *)
          str = "3*f1^2 + 2*I*f2";
          J = grtF`pSimplify[ 3*f1^2 + 2*I*f2 ];
          grtF`report[ "", str, J ];
          If[ J  ===  0 , (* then *)
            ptype = "II (or simpler)";
          , (* else *)
            ptype = "I (or simpler)";
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ], (* end If *)
  True, (* else *) 
    str1 = "Psi[0]*Psi[2] - Psi[1]^2";
    str2 = "Psi[0]*Psi[4] - Psi[2]^2";
    str3 = "Psi[0]*Psi[3] - Psi[1]*Psi[2]";
    H = grtF`pSimplify[ Psi[0]*Psi[2] - Psi[1]^2 ];
    E = grtF`pSimplify[ Psi[0]*Psi[4] - Psi[2]^2 ];
    F = grtF`pSimplify[ Psi[0]*Psi[3] - Psi[1]*Psi[2] ];
    grtF`report[ " H", str1, H ];
    grtF`report[ " E", str2, E ];
    grtF`report[ " F", str3, F ];
    If[ H  ===  0 , (* then *)
      If[ F  ===  0 , (* then *)
        If[ E  ===  0 , (* then *)
          ptype = "N (or simpler)";
        , (* else *)
          ptype = "I (or simpler)";
        ]; (* end If *)
      , (* else *)
        If[ E  ===  0 , (* then *)
          str = "37*Psi[2]^2+27*Psi[1]*Psi[3]";
          Q = grtF`pSimplify[ 37*Psi[2]^2+27*Psi[1]*Psi[3] ];
          grtF`report[ " Q", str, Q ];
          If[ Q  ===  0 , (* then *)
            ptype = "II (or simpler)";
          , (* else *) 
            ptype = "I (or simpler)";
          ]; (* end If *)
        , (* else *)
          str = "E - 4*(Psi[1]*Psi[3] - Psi[2]^2)";
          A = grtF`pSimplify[ Psi[1]*Psi[3] - Psi[2]^2 ];
          I = grtF`pSimplify[ E - 4*A ];
          grtF`report[ " I", str, I ];
          If[ I  ===  0 , (* then *) 
            ptype = "I (or simpler)";
          , (* else *)
            str = "Psi[2]*(Psi[1]*Psi[3] - Psi[2]^2) - Psi[3]*F";
            J = grtF`pSimplify[ Psi[2]*A - Psi[3]*F ];
            grtF`report[ " J", str, J ];
            str = "I^3 - 27*J^2";
            D = grtF`pSimplify[ I^3 - 27*J^2 ];
            grtF`report[ " D", str, D ];
            If[ D  ===  0 , (* then *)
              ptype = "II (or simpler)";
            , (* else *) 
              ptype = "I (or simpler)";
            ]; (* end If *)
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    , (* else *) 
      str = "Psi[1]*Psi[3] - Psi[2]^2";
      A = grtF`pSimplify[ Psi[1]*Psi[3] - Psi[2]^2 ];
      grtF`report[ " A", str, A ];
      str = "E - 4*A";
      I = grtF`pSimplify[ E - 4*A ];
      grtF`report[ " I", str, I ];
      If[ I  ===  0 , (* then *)
        str = "Psi[2]*A - Psi[3]*F + Psi[4]*H";
        J = grtF`pSimplify[ Psi[2]*A - Psi[3]*F + Psi[4]*H ];
        grtF`report[ " J", str, J ];
        If[ J  ===  0 , (* then *)
          ptype = "III (or simpler)";
        , (* else *) 
          ptype = "I (or simpler)";
        ]; (* end If *)
      , (* else *) 
        str = "Psi[0]*F - 2*Psi[1]*H";
        G = grtF`pSimplify[ Psi[0]*F - 2*Psi[1]*H ];
        grtF`report[ " G", str, G ];
        If[ G  ===  0 , (* then *)
          str = "Psi[0]^2*I - 12*H^2";
          Z = grtF`pSimplify[ Psi[0]^2*I - 12*H^2 ];
          grtF`report[ " Z", str, Z ];
          If[ Z  ===  0 , (* then *) 
            ptype = "D (or simpler)";
          , (* else *)
            str = "Psi[0]^2*I - 3*H^2";
            S = grtF`pSimplify[ Psi[0]^2*I - 3*H^2 ];
            grtF`report[ " S", str, S ];
            If[ S  ===  0 , (* then *)
              ptype = "II (or simpler)";
            , (* else *)
              ptype = "I (or simpler)";
            ]; (* end If *)
          ]; (* end If *)
        , (* else *)
          str = "Psi[2]*A - Psi[3]*F + Psi[4]*H";
          J = grtF`pSimplify[ Psi[2]*A - Psi[3]*F + Psi[4]*H ];
          grtF`report[ " J", str, J ];
          If[ J  ===  0 , (* then *)
            ptype = "I (or simpler)";
          , (* else *) 
            str = "I^3 - 27*J^2";
            D = grtF`pSimplify[ I^3 - 27*J^2 ];
            grtF`report[ " D", str, D ];
            If[ D  ===  0 , (* then *) 
              ptype = "II (or simpler)";
            , (* else *)
              ptype = "I (or simpler)";
            ]; (* end If *)
          ]; (* end If *)
        ]; (* end If *)
      ]; (* end If *)
    ]; (* end If *)

  ]; (* end Which *)

  Return[ ptype ];

]; (* end Module *)

(* ----------------------------------------- *)
(*  Simplification to be carried out through calculation. *)
(* ----------------------------------------- *)

grtF`pSimplify[ expr_] := simpHow[ expr ];

(* ----------------------------------------- *)
(*  Output reports. *)
(* ----------------------------------------- *)

grt`PetrovReport[ ] :=
Module[ {a, ptype},

  Which[ 
   ValueQ[ Petrov[`metricName] ] ,  
      ptype = Petrov[`metricName],
   ValueQ[ NPPetrov[metricName] ] ,  
      ptype = NPPetrov[metricName],
   True, (* else *)
      Print[ "The Petrov type has not been calculated." ];
      Abort[];
  ]; 

  If[ !ValueQ[ grtG`reportSize ] || grtG`reportSize  ===  0 , (* then *)
    Print[ "No report has been generated." ];
  , (* else *)
    Print[ "The conclusion Petrov type  ===  ",ptype ];
    Print[ "For the  ",metricName," metric" ];
    Print[ "was based on the following results:" ];

    For[ a = 1, a <=  grtG`reportSize , a++ ,
      Print[ grtG`reportLine[a] ];
    ]; (* end For *)

    Print[ "---> Therefore the metric is Petrov ",ptype,"." ];

    Print[ "-------------------------------------------------------" ];

    Print[ "The quantities that could not be evaluated to zero are; " ];
    For[ a = 1, a <=  grtG`reportVsize , a++ ,
      Print[ grtG`reportVline[a] ];
    ]; (* end For *)
  ]; (* end If *)
]; (* end Module *)

(* ------------------------------------------ *)
(*  *)
(*  Generate an `= 0` or `is not zero` line  *)
(*  for a report. *)
(*  *)
(* ------------------------------------------- *)

grtF`report[  vname_, str_, value_ ] :=

Module[ {},

If[ value  ===  0 , (* then *)
  grtG`reportSize = grtG`reportSize + 1;
  If[ vname =!= "" , (* then *)
    grtG`reportLine[grtG`reportSize] = StringJoin[
	ToString[vname],":", str," =  0"];
  , (* else *)
    grtG`reportLine[grtG`reportSize] = StringJoin[ str," = 0"];
  ]; (* end If *)
, (* else *)
  grtG`reportSize = grtG`reportSize + 1;
  grtG`reportVsize = grtG`reportVsize + 1;
  If[ vname =!= "" , (* then *)
    grtG`reportLine[grtG`reportSize] = StringJoin[ 
	ToString[vname],": ",str," could not be evaluated to zero."];
  , (* else *) 
    grtG`reportLine[grtG`reportSize] = StringJoin[str,
		" could not be evaluated to zero."];
  ]; (* end If *)
  If[ vname =!= "" , (* then *)
    grtG`reportVline[grtG`reportVsize] = vname  ==  value;
  , (* else *) 
    grtG`reportVline[grtG`reportVsize] = str  ==  value;
  ]; (* end If *)
]; (* end If *)

]; (* end Module *)

End[];

