(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"];

(* add all the objects to the object list *)

grtG`allObjects = Union[ grtG`allObjects,
   {grt`e[bdn,up], grtG`e[bdn,dn], grt`e[bup,up],
    grt`eta[bup,bup], grt`eta[bdn,bdn],
    grt`lambda[bdn,bdn,bdn], grt`rot[bdn,bdn,bdn],
    grt`str[bdn,bdn,bdn], grt`R[bdn,bdn,bdn,bdn],
    grt`R[bdn,bdn], grt`G[bdn,bdn], grt`S[bdn,bdn],
    grt`Weyl[bdn,bdn,bdn,bdn] } ];
    
(* ----------------------------------------------------------- *)
(*  multipleDefs for the basis library: *)
(* ----------------------------------------------------------- *)

multipleDef[grt`Phi00] = {grt`RPhi00, grt`Phi00};
multipleDef[grt`Phi01] = {grt`RPhi01, grt`Phi01};
multipleDef[grt`Phi02] = {grt`RPhi02, grt`Phi02};
multipleDef[grt`Phi10] = {grt`RPhi10, grt`Phi10};
multipleDef[grt`Phi11] = {grt`RPhi11, grt`Phi11};
multipleDef[grt`Phi12] = {grt`RPhi12, grt`Phi12};
multipleDef[grt`Phi20] = {grt`RPhi20, grt`Phi20};
multipleDef[grt`Phi21] = {grt`RPhi21, grt`Phi21};
multipleDef[grt`Phi22] = {grt`RPhi22, grt`Phi22};
multipleDef[grt`Lambda] = {grt`RLambda, grt`Lambda};
multipleDef[grt`Psi0] = {grt`WPsi0, grt`Psi0};
multipleDef[grt`Psi1] = {grt`WPsi1, grt`Psi1};
multipleDef[grt`Psi2] = {grt`WPsi2, grt`Psi2};
multipleDef[grt`Psi3] = {grt`WPsi3, grt`Psi3};
multipleDef[grt`Psi4] = {grt`WPsi4, grt`Psi4};

(* support for non-constant eta *)
multipleDef[grt`rot[bdn,bdn,bdn]] = { grt`genrot[bdn,bdn,bdn], 
				grt`rot[bdn,bdn,bdn] };
multipleDef[grt`R[bdn,bdn,bdn,bdn]] = { grt`genR[ bdn,bdn,bdn,bdn],
                                        grt`R[bdn,bdn,bdn,bdn] };
(*
 * useWhen function for special definitions for non-constant eta.
 * (For constant eta we can take advantage of special calc fns and 
 * symmetries.)
 *
 * Eta is checked in grload and grtG`EtaNotConstant is assigned there
 *
 * Used by genrot[bdn,bdn,bdn] and genR[bdn,bdn,bdn,bdn]
 *)
grtF`useWhenEtaNotConstant[] := 
Module[ {}, 
    If[ ValueQ[ EtaNotConstant[ metricName] ],
       Return[ EtaNotConstant[ metricName] ];
    ,(*else*)
       Return[ False];
    ];
];

(* =========================================================== *)
(*  Fundamental basis objects: 				       *)
(* =========================================================== *)

(* ----------------------------- *)
(*  grt`e[bdn,up] - standard definition. *)
(* ----------------------------- *)

 ObjDef[grt`e[bdn,up]][grtC`header] = "Basis vectors";
 ObjDef[grt`e[bdn,up]][grtC`root] = ebdnup;
 ObjDef[grt`e[bdn,up]][grtC`rootStr] = "e ";
 ObjDef[grt`e[bdn,up]][grtC`indexList] = {bdn,up};
 ObjDef[grt`e[bdn,up]][grtC`calcFn] = grtF`ebdnup;
 ObjDef[grt`e[bdn,up]][grtC`preCalcFn] = grtF`precalcebdnup;
 ObjDef[grt`e[bdn,up]][grtC`symmetry] = grtF`symnosym2;
 ObjDef[grt`e[bdn,up]][grtC`depends] = { grt`g[up,up] };

grtF`precalcebdnup[  object_, index_ ] := Module[{},
    If[ !ebdndn[metricName,0,0]  ===  True , (* then *)
	Print["Unable to calculate grt`e[bdn,up] (No basis vectors have been specified)"];
	Abort[];
    ]; (* end If *)
];

grtF`ebdnup[ object_, index_] := 
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a, s},
  s = 0;
  For[ a = 1, a <=  Ndim[metricName] , a++ ,
	s = s + gupup[metricName,a2, a]*
		ebdndn[metricName,a1,a];
  ]; (* end For *)

  Return[s];

]; (* end Module *)

(* ----------------------------- *)
(*  grt`e[bdn,dn] *)
(* ----------------------------- *)
 ObjDef[grt`e[bdn,dn]][grtC`header] = "Basis 1-forms";
 ObjDef[grt`e[bdn,dn]][grtC`root] = ebdndn;
 ObjDef[grt`e[bdn,dn]][grtC`rootStr] = "e ";
 ObjDef[grt`e[bdn,dn]][grtC`indexList] = {bdn,dn};
 ObjDef[grt`e[bdn,dn]][grtC`calcFn] = grtF`calcebdndn;
 ObjDef[grt`e[bdn,dn]][grtC`preCalcFn] = grtF`precalcebdndn;
 ObjDef[grt`e[bdn,dn]][grtC`symmetry] = grtF`symnosym2;
 ObjDef[grt`e[bdn,dn]][grtC`depends] = { grt`g[dn,dn] };

grtF`precalcebdndn[  object_, index_ ] := Module[{}, 
    If[ !ebdnup[metricName,0,0]  ===  True , (* then *)
	Print["Unable to calculate grt`e[bdn,up] (No basis vectors have been specified)"];
	Abort[];
    ]; (* end If *)
];

grtF`calcebdndn[ object_, index_] := 
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a,s},
  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
    s = s + gdndn[metricName,a,a2]*
	    ebdnup[metricName,a1,a];
  ]; (* end For *)
  Return[s];
]; (* end Module *)

(* ----------------------------- *)
(*  grt`e[bup,up] *)
(* ----------------------------- *)
ObjDef[grt`e[bup,up]][grtC`header] = "Basis vectors";
ObjDef[grt`e[bup,up]][grtC`root] = ebupup;
ObjDef[grt`e[bup,up]][grtC`rootStr] = "e ";
ObjDef[grt`e[bup,up]][grtC`indexList] = {bup,up};
ObjDef[grt`e[bup,up]][grtC`calcFn] = grtF`calcebupup;
ObjDef[grt`e[bup,up]][grtC`symmetry] = grtF`symnosym2;
ObjDef[grt`e[bup,up]][grtC`depends] = {grt`e[bdn,up],grt`eta[bup,bup]};

grtF`calcebupup[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a,s},
  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
     s = s + etabupbup[metricName,a,a1]*
	     ebdnup[metricName,a,a2];
  ]; (* end For *)

  Return[s];

]; (* end Module *)

(* ----------------------------- *)
(*  grt`e[bup,dn] *)
(* ----------------------------- *)
 ObjDef[grt`e[bup,dn]][grtC`header] = "Basis 1-forms";
 ObjDef[grt`e[bup,dn]][grtC`root] = ebupdn;
 ObjDef[grt`e[bup,dn]][grtC`rootStr] = "e ";
 ObjDef[grt`e[bup,dn]][grtC`indexList] = {bup,dn};
 ObjDef[grt`e[bup,dn]][grtC`calcFn] = grtF`calcebupdn;
 ObjDef[grt`e[bup,dn]][grtC`symmetry] = grtF`symnosym2;
 ObjDef[grt`e[bup,dn]][grtC`depends] = {grt`e[bdn,dn],grt`eta[bup,bup]};

grtF`calcebupdn[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a,s},

  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
      s = s + etabupbup[metricName,a,a1]*
	      ebdndn[metricName,a,a2];
  ]; (* end For *)
  Return[s];
]; (* end Module *)

(* ----------------------------- *)
(*  grt`eta[bup,bup] *)
(* ----------------------------- *)
 ObjDef[grt`eta[bup,bup]][grtC`header] = "Basis inner product";
 ObjDef[grt`eta[bup,bup]][grtC`root] = etabupbup;
 ObjDef[grt`eta[bup,bup]][grtC`rootStr] = "eta";
 ObjDef[grt`eta[bup,bup]][grtC`indexList] = {bup,bup};
 ObjDef[grt`eta[bup,bup]][grtC`calcFn] = grtF`loadeta;
 ObjDef[grt`eta[bup,bup]][grtC`symmetry] = grtF`symsym2;
 ObjDef[grt`eta[bup,bup]][grtC`depends] = {};

grtF`loadeta[ object_, index_] :=
Module[{},
   Print["eta[bup,bup] must be created with makeg"];
   Abort[];
];

(* ----------------------------- *)
(*  grt`eta[bdn,bdn] *)
(* ----------------------------- *)
 ObjDef[grt`eta[bdn,bdn]][grtC`header] = "Basis inner product";
 ObjDef[grt`eta[bdn,bdn]][grtC`root] = etabdnbdn;
 ObjDef[grt`eta[bdn,bdn]][grtC`rootStr] = "eta";
 ObjDef[grt`eta[bdn,bdn]][grtC`indexList] = {bdn,bdn};
 ObjDef[grt`eta[bdn,bdn]][grtC`preCalcFn] = grtF`calcetabdnbdn;
 ObjDef[grt`eta[bdn,bdn]][grtC`symmetry] = grtF`symsym2;
 ObjDef[grt`eta[bdn,bdn]][grtC`depends] = {eta[bup,bup]};

grtF`calcetabdnbdn[ object_] := 
Module[ {},

  grtF`invMetric[ etabdnbdn, etabupbup];

]; (* end Module *)

(* ----------------------------- *)
(*  grt`lambda[bdn,bdn,bdn] *)
(* ----------------------------- *)
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`header] = "pre-Rotation Coefficients";
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`root] = lambdabdnbdnbdn;
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`rootStr] = "lambda";
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn};
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`calcFn] = grtF`calclambdabdnbdnbdn;
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`symmetry] = grtF`sym3a13;
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`depends] = {grt`e[bdn,dn],grt`e[bdn,up]};
 ObjDef[grt`lambda[bdn,bdn,bdn]][grtC`attributes] = {usediffconstraint};

grtF`calclambdabdnbdnbdn[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a,b,s},
  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
    For[ b = 1, b <=  Ndim[metricName], b++ ,
       s = s + D[ebdndn[metricName,a2,a], xup[metricName, b] ] * 
               ( ebdnup[metricName,a1,a] * 
		 ebdnup[metricName,a3,b]
               - ebdnup[metricName,a1,b] * 
		 ebdnup[metricName,a3,a] );
    ]; (* end For *)
  ]; (* end For *)
  Return[s];
]; (* end Module *)

(* ----------------------------- *)
(*  grt`rot[bdn,bdn,bdn] *)
(* ----------------------------- *)
(* (  ) *)
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`header] = "Rotation Coefficients";
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`root] = rotbdnbdnbdn;
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`rootStr] = "gamma";
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn};
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcrotbdnbdnbdn;
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`symmetry] = grtF`sym3a12;
 ObjDef[grt`rot[bdn,bdn,bdn]][grtC`depends] = {grt`lambda[bdn,bdn,bdn]};

grtF`calcrotbdnbdnbdn[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { s},
	s = (lambdabdnbdnbdn[metricName,a1,a2,a3]
            + lambdabdnbdnbdn[metricName,a3,a1,a2]
            - lambdabdnbdnbdn[metricName,a2,a3,a1])/2;
	Return[s];
]; (* end Module *)

(*-----------------------------*)
(* genrot(bdn,bdn,bdn)         *)
(*-----------------------------*)
 ObjDef[grt`genrot[bdn,bdn,bdn]][grtC`displayName] = rot[bdn,bdn,bdn];
 ObjDef[grt`genrot[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcgenrotbdnbdnbdn;
 ObjDef[grt`genrot[bdn,bdn,bdn]][grtC`useWhen] = grtF`useWhenEtaNotConstant;
 ObjDef[grt`genrot[bdn,bdn,bdn]][grtC`depends] = {grt`e[bdn,up], 
                                                  grt`e[bdn,dn,cdn]};
 ObjDef[grt`genrot[bdn,bdn,bdn]][grtC`symmetry] = grtF`symnosym3;


grtF`calcgenrotbdnbdnbdn[object_, index_] := 
Module[ {s,a,b}, 
	s = 0;
	For[ a=1, a <= Ndim[metricName], a++, 
	  For[ b=1, b <= Ndim[metricName], b++, 
	    s = s + ebdnup[metricName,a1,a]*
	            ebdndncdn[metricName,a2,a,b]*
	            ebdnup[metricName,a3,b];
	  ];
	];
	Return[s];
];

(* ----------------------------- *)
(*  grt`str[bdn,bdn,bdn] *)
(* ----------------------------- *)
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`header] = "Structure Constants";
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`root] = strbdnbdnbdn;
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`rootStr] = "C";
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn};
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`calcFn] = grtF`calcstrbdnbdnbdn;
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`symmetry] = grtF`sym3a23;
 ObjDef[grt`str[bdn,bdn,bdn]][grtC`depends] = {grt`rot[bdn,bdn,bdn]};

grtF`calcstrbdnbdnbdn[ object_, index_] := 
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { s},
  s = rotbdnbdnbdn[metricName,a1,a3,a2] - 
      rotbdnbdnbdn[metricName,a2,a3,a1];
  Return[s];
]; (* end Module *)

(* ----------------------------- *)
(*  grt`R[bdn,bdn,bdn,bdn] [Alternate form of the Riemann tensor] *)
(* ----------------------------- *)
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`header] = "Covariant Riemann";
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`root] = Rbdnbdnbdnbdn;
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`rootStr] = "R";
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn,bdn};
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`calcFn] = grtF`calcRbdnbdnbdnbdn;
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`symmetry] = grtF`symRiem;
 ObjDef[grt`R[bdn,bdn,bdn,bdn]][grtC`depends] = {grt`rot[bdn,bdn,bdn],
		grt`rot[bdn,bup,bdn]};

grtF`calcRbdnbdnbdnbdn[ object_, index_] := 
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a, s},

  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
    s = s - D[ rotbdnbdnbdn[metricName,a1,a2,a3], 
	       xup[metricName,a] ]*
       ebdnup[metricName,a4,a] +
       D[ rotbdnbdnbdn[metricName,a1,a2,a4], 
	   xup[metricName,a] ]*
       ebdnup[metricName,a3,a] +
       rotbdnbdnbdn[metricName,a2,a1,a]*
     ( rotbdnbupbdn[metricName,a3,a,a4] - 
       rotbdnbupbdn[metricName,a4,a,a3] )
     + rotbdnbdnbdn[metricName,a,a1,a3]*
       rotbdnbupbdn[metricName,a2,a,a4]
     - rotbdnbdnbdn[metricName,a,a1,a4]*
       rotbdnbupbdn[metricName,a2,a,a3];
  ]; (* end For *)

  Return[s];
]; (* end Module *)

(*-----------------------------*)
(* genR(bdn,bdn,bdn,bdn)       *)
(*-----------------------------*)
 ObjDef[grt`genR[bdn,bdn,bdn,bdn]][grtC`displayName] = R[bdn,bdn,bdn,bdn];
 ObjDef[grt`genR[bdn,bdn,bdn,bdn]][grtC`calcFn] = grtF`calcgenRbdnbdnbdn;
 ObjDef[grt`genR[bdn,bdn,bdn,bdn]][grtC`useWhen] = grtF`useWhenEtaNotConstant;
 ObjDef[grt`genR[bdn,bdn,bdn,bdn]][grtC`depends] = {grt`eta[bdn,bdn],
                                                    grt`e[bdn,up],
                                                    grt`rot[bdn,bdn,bdn], 
                                                    grt`rot[bdn,bup,bdn],
                                                    grt`rot[bdn,bup,bup,pdn]};


grtF`calcgenRbdnbdnbdn[object_, index_] := 
Module[ {s,a,b,c, gname},
        gname = metricName;
	s = 0;
	For[ a=1, a <= Ndim[metricName], a++, 
	  For[ b=1, b <= Ndim[metricName], b++, 
	    For[ c=1, c <= Ndim[metricName], c++,  
	          s = s + rotbdnbupbuppdn[gname,a1,a,b,c]*
                          ebdnup[gname,a3,c]*
                          etabdnbdn[gname,a,a2]*
                          etabdnbdn[gname,b,a4]
                      -   rotbdnbupbuppdn[gname,a1,a,b,c]*
                          ebdnup[gname,a4,c]*
                          etabdnbdn[gname,a,a2]*
                          etabdnbdn[gname,b,a3];
	      ]; (* For c *)
	    ]; (* For b *)
	  ]; (* For a *)
	For[ a=1, a <= Ndim[metricName], a++, 
	  s = s + rotbdnbdnbdn[gname,a2,a1,a]* 
                ( rotbdnbupbdn[gname,a3,a,a4] - rotbdnbupbdn[gname,a4,a,a3] )
	         + rotbdnbdnbdn[gname,a,a1,a3]* rotbdnbupbdn[gname,a2,a,a4]
	         - rotbdnbdnbdn[gname,a,a1,a4]* rotbdnbupbdn[gname,a2,a,a3];
	]; (* For *)
	Return[s];
];


(* ----------------------------- *)
(*  grt`R[bdn,bdn] [Alternate form of the Ricci tensor] *)
(* ----------------------------- *)
 ObjDef[grt`R[bdn,bdn]][grtC`header] = "Covariant Ricci";
 ObjDef[grt`R[bdn,bdn]][grtC`root] = Rbdnbdn;
 ObjDef[grt`R[bdn,bdn]][grtC`rootStr] = "R";
 ObjDef[grt`R[bdn,bdn]][grtC`indexList] = {bdn,bdn};
 ObjDef[grt`R[bdn,bdn]][grtC`calcFn] = grtF`calcRbdnbdn;
 ObjDef[grt`R[bdn,bdn]][grtC`symmetry] = grtF`symsym2;
 ObjDef[grt`R[bdn,bdn]][grtC`depends] = {grt`eta[bup,bup], 
                                         grt`R[bdn,bdn,bdn,bdn]};

grtF`calcRbdnbdn[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a, b, s},

  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
    For[ b = 1, b <=  Ndim[metricName], b++ ,
      s = s + etabupbup[metricName,a,b]*
	      Rbdnbdnbdnbdn[metricName,a,a1,b,a2];
    ]; (* end For *)
  ]; (* end For *)

  Return[s];
]; (* end Module *)

(* ----------------------------- *)
(*  bRicciscalar [Alternate form of the Ricci scalar] *)
(* ----------------------------- *)
 ObjDef[grt`bRicciscalar][grtC`header] = "Ricci Scalar (basis)";
 ObjDef[grt`bRicciscalar][grtC`root] = R;
 ObjDef[grt`bRicciscalar][grtC`rootStr] = "R";
 ObjDef[grt`bRicciscalar][grtC`indexList] = {};
 ObjDef[grt`bRicciscalar][grtC`symmetry] = grtF`symscalar;

 multipleDef[Ricciscalar] = {grt`bRicciscalar, Ricciscalar};
 ObjDef[grt`bRicciscalar][grtC`displayName] = Ricciscalar;
 ObjDef[grt`bRicciscalar][grtC`calcFn] = grtF`calcbRsc;
 ObjDef[grt`bRicciscalar][grtC`depends] = {grt`R[bdn,bdn]};
 ObjDef[grt`bRicciscalar][grtC`useWhen] = grtF`whenbRsc;

grtF`whenbRsc[] := 
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
   Return[ grtF`checkIfAssigned [ grt`R[bdn,bdn] ] ];

grtF`calcbRsc[ object_, index_] :=
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { a, b, s},

  s = 0;
  For[ a = 1, a <=  Ndim[metricName], a++ ,
    For[ b = 1, b <=  Ndim[metricName], b++ ,
      s = s + etabupbup[metricName,a,b]*
	      Rbdnbdn[metricName,a,b];
    ]; (* end For *)
  ]; (* end For *)

  Return[s];
]; (* end Module *)


(* ---------------------------- *)
(*  grt`Weyl[bdn,bdn,bdn,bdn] [Alternate form of the Weyl tensor] *)
(* ---------------------------- *)
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`header] = "Covariant Weyl";
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`root] = Weylbdnbdnbdnbdn;
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`rootStr] = "C";
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`indexList] = {bdn,bdn,bdn,bdn};
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`calcFn] = grtF`calcCbdnbdnbdnbdn;
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`symmetry] = grtF`symRiem;
 ObjDef[grt`Weyl[bdn,bdn,bdn,bdn]][grtC`depends] = {grt`eta[bdn,bdn], 
	grt`R[bdn,bdn,bdn,bdn], grt`R[bdn,bdn], grt`bRicciscalar};

grtF`calcCbdnbdnbdnbdn[ objectName_,index_] :=  
(*  `Copyright 1994 by Peter Musgrave, Denis Pollney and Kayll Lake`;*)
Module[ { s},
 s = Rbdnbdnbdnbdn[metricName,a1,a2,a3,a4]
     -1/(Ndim[metricName]-2)*
	 (etabdnbdn[metricName,a1,a3]*Rbdnbdn[metricName,a4,a2] -
	  etabdnbdn[metricName,a1,a4]*Rbdnbdn[metricName,a3,a2] -
	  etabdnbdn[metricName,a2,a3]*Rbdnbdn[metricName,a4,a1] +
	  etabdnbdn[metricName,a2,a4]*Rbdnbdn[metricName,a3,a1]) +
      1/(Ndim[metricName] - 1)/(Ndim[metricName]- 2)*
	    scalarR[metricName]*
    (etabdnbdn[metricName,a1,a3]*etabdnbdn[metricName,a4,a2] -
     etabdnbdn[metricName,a1,a4]*etabdnbdn[metricName,a3,a2] );

 Return[s];

]; (* end Module *)

(* ---------------------------- *)
(*  grt`G[bdn,bdn] [Alternate form of the Einstein tensor] *)
(* ---------------------------- *)
 ObjDef[grt`G[bdn,bdn]][grtC`header] = "Covariant Einstein";
 ObjDef[grt`G[bdn,bdn]][grtC`root] = Gbdnbdn;
 ObjDef[grt`G[bdn,bdn]][grtC`rootStr] = "G";
 ObjDef[grt`G[bdn,bdn]][grtC`indexList] = {bdn,bdn};
 ObjDef[grt`G[bdn,bdn]][grtC`calcFn] = grtF`calcsum0;
 ObjDef[grt`G[bdn,bdn]][grtC`calcFnParms] = Hold[Rbdnbdn[metricName,a1,a2]
	- 1/2*(etabdnbdn[metricName,a1,a2] * 
	  scalarR[metricName])];
 ObjDef[grt`G[bdn,bdn]][grtC`symmetry] = grtF`symsym2;
 ObjDef[grt`G[bdn,bdn]][grtC`depends] = {grt`eta[bdn,bdn], grt`R[bdn,bdn], 
		grt`bRicciscalar};

(* ---------------------------- *)
(*  grt`S[bdn,bdn] [Alternate form of the trace free Ricci tensor] *)
(* ---------------------------- *)
 ObjDef[grt`S[bdn,bdn]][grtC`header] = "Covariant trace-free Ricci";
 ObjDef[grt`S[bdn,bdn]][grtC`root] = Sbdnbdn;
 ObjDef[grt`S[bdn,bdn]][grtC`rootStr] = "S";
 ObjDef[grt`S[bdn,bdn]][grtC`indexList] = {bdn,bdn};
 ObjDef[grt`S[bdn,bdn]][grtC`calcFn] = grtF`calcsum0;
 ObjDef[grt`S[bdn,bdn]][grtC`calcFnParms] =
	Hold[Rbdnbdn[metricName,a1,a2] -
	(1/Ndim[metricName]) * etabdnbdn[metricName,a1,a2] * 
	scalarR[metricName] ];
 ObjDef[grt`S[bdn,bdn]][grtC`symmetry] = grtF`symsym2;
 ObjDef[grt`S[bdn,bdn]][grtC`depends] = { grt`eta[bdn,bdn] ,grt`R[bdn,bdn], 
		grt`bRicciscalar };


End[];


