(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 


Begin["grtG`"]; (* set scope to global vars *)




ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`header] = "Dual Weyl (first two indices dual)";
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`root] = Weylstarbdnbdnbdnbdn;
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`indexList] = {bdn,bdn,bdn,bdn};
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`calcFn] = grtF`calcbdual;
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`calcFnParms] = {LevCdndndndn, Weylbdnbdnbupbup};
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`symmetry] = grtF`symRiem;
ObjDef[ grt`Weylstar[bdn,bdn,bdn,bdn] ][grtC`depends] = {grt`LevC[dn,dn,dn,dn], grt`Weyl[bdn,bdn,bup,bup]};

grtF`calcbdual[ object_, iList_] :=

 (*
  * Loop structure assumes one of the root objects
  * is perm like (i.e. if p=q then result is zero, and
  * can sum over bupper diagonal only since defintion
  * of Weylstar has a 1/2 in it.)
  *)
 Module[ {p,q, s, root1, root2},
  root1 = grtG`ObjDef[object][grtC`calcFnParms][[1]];
  root2 = grtG`ObjDef[object][grtC`calcFnParms][[2]];
  s = 0;
  Do[
    Do[
      s = s + grtG`root1[metricName,p,q,a1,a2] * grtG`root2[metricName,a3,a4,p,q],
     {q, p+1, grtG`Ndim[ grtG`metricName], 1}
    ],
   { p, grtG`Ndim[grtG`metricName]}
  ];
  Return[s];
]

(*----------------------------
# Weylstar(bdn,bdn,bup,bup)
#----------------------------
#        cd              ijcd  
# MCstar     = 1/2 e    C      
#       ab          ijab
#
*)

ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`header] = "Dual Weyl (dual w.r.t first two indices)";
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`root] = Weylstarbdnbdnbupbup;
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`indexList] = {bdn,bdn,bup,bup};
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`calcFn] = grtF`calcdualbup;
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`calcFnParms] = {LevCdndndndn, Weylbupbupbupbup};
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`symmetry] = grtF`symMRiem;
ObjDef[ grt`Weylstar[bdn,bdn,bup,bup] ][grtC`depends] = {grt`LevC[dn,dn,dn,dn], grt`Weyl[bup,bup,bup,bup]};

(*----------------------------
# Cstar(bup,bup,bup,bup)
#----------------------------
#       abcd        ijab   cd
#     C *    = 1/2 e    C
#                        ij
*)

ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`header] = "Dual Weyl (dual w.r.t first two indices)";
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`root] = Weylstarbupbupbupbup;
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`indexList] = {bup,bup,bup,bup};
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`calcFn] = grtF`calcdualbup;
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`calcFnParms] = {LevCupupupup, Weylbdnbdnbupbup};
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`symmetry] = grtF`symRiem;
ObjDef[ grt`Weylstar[bup,bup,bup,bup] ][grtC`depends] = { grt`Weyl[bdn,bdn,bup,bup], grt`LevC[up,up,up,up] };

grtF`calcdualbup[object_, iList_] :=
 Module[ {p,q, s, root1, root2},
   root1 = ObjDef[object][grtC`calcFnParms][[1]];
   root2 = ObjDef[object][grtC`calcFnParms][[2]];
   s = 0;
   Do[
    Do[
      s = s + grtG`root1[metricName,p,q,a1,a2] * grtG`root2[metricName,p,q,a3,a4],
     {q, p, grtG`Ndim[grtG`metricName]}
    ],
    {p, grtG`Ndim[grtG`metricName]-1}
   ];

  Return[s];
]

End[]; (* end the global scope *)
