(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"]; (* set scope to global vars *)

(*---------------------------
# M1I
# - this is calculated when M1R is calculated
#----------------------------
#                    ef   dc ba                 ab  cd
# Im(M1)= 1/8  e    C    S  S     = 1/16 C*    S  S
#               acef   db                 acdb
*)

ObjDef[ grt`M1I ][grtC`header] = "CM invariant Im(M1)";
ObjDef[ grt`M1I ][grtC`root] = grtG`M1I;
ObjDef[ grt`M1I ][grtC`rootStr] = "M1I ";
ObjDef[ grt`M1I ][grtC`indexList] = {};
ObjDef[ grt`M1I ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M1I ][grtC`calcFn] = grtF`calcM1I;
ObjDef[ grt`M1I ][grtC`depends] = {grt`S[up,up], grt`Weylstar[dn,dn,dn,dn]};


grtF`calcM1I[object_, iList_] :=
 Module[ {a, b, c, d, s},
  s = 0;
  Do[
    Do[
      If[  grtG`Supup[metricName,a,b] =!= 0 ,
	Do[
	  Do[
	    If[ grtG`Supup[metricName,c,d] =!= 0 ,
	      s =  s + grtG`Weylstardndndndn[metricName,a,c,d,b]
		    * grtG`Supup[metricName,a,b] * grtG`Supup[metricName,c,d],
	    ],
	  {d, Ndim[metricName]}],
	{c, Ndim[metricName]}];
	If[ grtG`fnCode === grtC`CALCALTER  ,
		s = grtG`simpHow[ s];
	]
      ],
    { b, Ndim[metricName]}],
  {a, Ndim[metricName]}];

  Return[s/8];
]

(*---------------------------
# M1R
#----------------------------
#                     cd ab
#  Re(M1) = 1/8 C    S  S
#                acdb
*)

ObjDef[ grt`M1R ][grtC`header] = "CM invariant Re(M1)";
ObjDef[ grt`M1R ][grtC`root] = grtG`M1R;
ObjDef[ grt`M1R ][grtC`rootStr] = "M1R ";
ObjDef[ grt`M1R ][grtC`indexList] = {};
ObjDef[ grt`M1R ][grtC`calcFn] = grtF`calcM1R;
ObjDef[ grt`M1R ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M1R ][grtC`depends] = {grt`S[up,up], grt`Weyl[dn,dn,dn,dn]};

grtF`calcM1R[object_, iList_] :=
 Module[ {a, b, c, d, s2},
  s2 = 0;
  Do[
    Do[
      If[  grtG`Supup[metricName,a,b] =!= 0 ,
	Do[
	  Do[
	    If[ grtG`Supup[metricName,c,d] =!= 0 ,
	      s2 = s2 + grtG`Weyldndndndn[metricName,a,c,d,b] *
		    grtG`Supup[metricName,a,b] * grtG`Supup[metricName,c,d];
	    ],
	  {d, Ndim[metricName]}],
	{c, Ndim[metricName]}];
	If[ grtG`fnCode === grtC`CALCALTER  ,
		s2 = grtG`simpHow[ s2];
	]
      ],
    { b, Ndim[metricName]}],
  {a, Ndim[metricName]}];

  Return[s2/8];
]

(*---------------------------
# M2R
#----------------------------
#                cd           aefb            aekl ij     fb
# Re(M2) = 1/16 S  S  ( C    C    - 1/4 e    e    C    C    )
#                   ef   acdb            acij        db kl
#
#                cd           aefb                     aefb
# Re(M2) = 1/16 S  S  ( C    C    - 1/4 Weylstar    UWeylstar    )
#                   ef   acdb                acdb
#
# M3 is the same formula, except with a plus sign so store that
# result Do[ free
#
#                     ij   aefd bc                      aefd bc
# Im(M2) = 1/16 e    C    C    S  S    = 1/16 Weylstar    C    S  S
#                abij   cd         ef              abcd         ef
*)

ObjDef[ grt`M2R ][grtC`header] = "CM invariant Re(M2)";
ObjDef[ grt`M2R ][grtC`root] = grtG`M2R;
ObjDef[ grt`M2R ][grtC`rootStr] = "M2R ";
ObjDef[ grt`M2R ][grtC`indexList] = {};
ObjDef[ grt`M2R ][grtC`calcFn] = grtF`calcM2M3;
ObjDef[ grt`M2R ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M2R ][grtC`depends] = {grt`S[up,up], grt`S[dn,dn],
		    grt`Weyl[dn,dn,dn,dn], grt`Weyl[up,up,up,up],
		    grt`Weylstar[dn,dn,dn,dn], grt`Weylstar[up,up,up,up]};

grtF`calcM2M3[object_, iList_] :=
 Module[{a, b, c, d, e, f, s1, s2, m2i},
  s1 = 0;
  s2 = 0;
  m2i = 0;
  Do[
    Do[
     If[ grtG`Supup[metricName,c,d] =!= 0 ,
      Do[
	Do[
	  If[ grtG`Sdndn[metricName,e,f] =!= 0 ,
	   Do[
	     Do[
	      s1 = s1 + grtG`Supup[metricName,c,d] * grtG`Sdndn[metricName,e,f]
		    * grtG`Weyldndndndn[metricName,a,c,d,b]* grtG`Weylupupupup[metricName,a,e,f,b];
	      m2i = m2i + grtG`Supup[metricName,c,d] * grtG`Sdndn[metricName,e,f]
		    * grtG`Weylstardndndndn[metricName,a,c,d,b]* grtG`Weylupupupup[metricName,a,e,f,b];
	      s2 = s2 + grtG`Supup[metricName,c,d] * grtG`Sdndn[metricName,e,f]
		    * grtG`Weylstardndndndn[metricName,a,c,d,b]
		    * grtG`Weylstarupupupup[metricName,a,e,f,b],
	     {b, Ndim[metricName]}],
	   {a, Ndim[metricName]}];
	   If[ grtG`fnCode = grtC`CALCALTER  ,
	       s1 = grtG`simpHow[s1];
	       s2 = grtG`simpHow[s2];
	      m2i = grtG`simpHow[m2i];
	   ];
	  ],
	{f, Ndim[metricName]}],
      {e, Ndim[metricName]}]
     ],
    {d, Ndim[metricName]}],
  {c, Ndim[metricName]}];

  grtG`M3[metricName] = (s1+s2)/16;
  grtG`M2I[metricName] = m2i/8;
  Return[(s1-s2)/16];
]

(*---------------------------
# M2I
#----------------------------*)

ObjDef[ grt`M2I ][grtC`header] = "CM invariant Im(M2)";
ObjDef[ grt`M2I ][grtC`root] = grtG`M2I;
ObjDef[ grt`M2I ][grtC`rootStr] = "M2I ";
ObjDef[ grt`M2I ][grtC`indexList] = {};
ObjDef[ grt`M2I ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M2I ][grtC`depends] = {grt`M2R};

(*---------------------------
# M3
#----------------------------*)

ObjDef[ grt`M3 ][grtC`header] = "CM invariant M3";
ObjDef[ grt`M3 ][grtC`root] = grtG`M3;
ObjDef[ grt`M3 ][grtC`rootStr] = "M3 ";
ObjDef[ grt`M3 ][grtC`indexList] = {};
ObjDef[ grt`M3 ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M3 ][grtC`depends] = { grt`M2R};


(*---------------------------
# M4
#----------------------------
#             a  cd         befg    d  e       kcfg
# M4 = -1/32(S  S  S  C    C     + S  S  C    C
#              g    ef acdb          c  f keda
#
#                  cd         ijfg
#            -1/2 S  S  C    C     )
#                     cf ijda
*)

ObjDef[ grt`M4 ][grtC`header] = "CM invariant M4";
ObjDef[ grt`M4 ][grtC`root] = grtG`M4;
ObjDef[ grt`M4 ][grtC`rootStr] = "M4 ";
ObjDef[ grt`M4 ][grtC`indexList] = {};
ObjDef[ grt`M4 ][grtC`calcFn] = grtF`calcM4;
ObjDef[ grt`M4 ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M4 ][grtC`depends] = {grt`S[up,up], grt`S[up,dn],
		    grt`Weyl[dn,dn,dn,dn], grt`Weyl[dn,dn,up,up],
		    grt`Weylstar[dn,dn,dn,dn], grt`Weylstar[dn,dn,up,up] };


grtF`calcM4[object_, iList_] :=
 Module[ {a,b,c,d,e,f,g, s},
  s = 0;
  Do[
    Do[
      If[ grtG`Supdn[metricName,c,d] =!= 0 ,
	Do[
	  Do[
	   If[ grtG`Supup[metricName,e,f] =!= 0 ,
	    Do[
	      Do[
		If[ grtG`Supup[metricName,g,a] =!= 0 ,
		  Do[
		    s = s + grtG`Supup[metricName,g,a] *
			     grtG`Supup[metricName,e,f] *
			     grtG`Supdn[metricName,c,d] *
			  (grtG`Weyldndnupup[metricName,a,c,d,b] *
			   grtG`Weyldndndndn[metricName,b,e,f,g] +
			   grtG`Weylstardndnupup[metricName,a,c,d,b] *
			   grtG`Weylstardndndndn[metricName,b,e,f,g]),
		  {b, Ndim[metricName]}]
		],
	      {g, Ndim[metricName]}],
	    {a, Ndim[metricName]}];
	    If[ grtG`fnCode === grtC`calcalter ,
	       s = grtG`simpHow[ s];
	    ];
	   ],
	  {f, Ndim[metricName]}],
	{e, Ndim[metricName]}];
      ],
    {d, Ndim[metricName]}],
  {c, Ndim[metricName]}];
  Return[-s/32];
]


(*---------------------------
# M5R
# (calc functions defined in cm_m5.m)
#----------------------------*)

ObjDef[ grt`M5R ][grtC`header] = "CM invariant Re(M5)";
ObjDef[ grt`M5R ][grtC`root] = grtG`M5R;
ObjDef[ grt`M5R ][grtC`rootStr] = "M5R ";
ObjDef[ grt`M5R ][grtC`indexList] = {};
ObjDef[ grt`M5R ][grtC`calcFn] = grtF`calcM5;
ObjDef[ grt`M5R ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M5R ][grtC`depends] = {grt`S[up,up], grt`Weyl[dn,dn,dn,dn],
			grt`Weyl[up,up,up,up],
			grt`Weylstar[dn,dn,dn,dn], grt`Weylstar[up,up,up,up]};

(*---------------------------
# M5I
#----------------------------*)

ObjDef[ grt`M5I ][grtC`header] = "CM invariant Im(M5)";
ObjDef[ grt`M5I ][grtC`root] = grtG`M5I;
ObjDef[ grt`M5I ][grtC`rootStr] = "M5I ";
ObjDef[ grt`M5I ][grtC`indexList] = {};
ObjDef[ grt`M5I ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`M5I ][grtC`depends] = {grt`M5R};

(*---------------------------
# W1R
#----------------------------*)

ObjDef[ grt`W1R ][grtC`header] = "CM invariant Re(W1)";
ObjDef[ grt`W1R ][grtC`root] = grtG`W1R;
ObjDef[ grt`W1R ][grtC`rootStr] = "W1R ";
ObjDef[ grt`W1R ][grtC`indexList] = {};
ObjDef[ grt`W1R ][grtC`calcFn] = grtF`calcsum0;
ObjDef[ grt`W1R ][grtC`calcFnParms] = Hold[ grtG`WeylSq[grtG`metricName]/8];
ObjDef[ grt`W1R ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`W1R ][grtC`symPerm] = {};
ObjDef[ grt`W1R ][grtC`depends] = {grt`WeylSq};

(*---------------------------
# W1I
#----------------------------*)

ObjDef[ grt`W1I ][grtC`header] = "CM invariant Im(W1)";
ObjDef[ grt`W1I ][grtC`root] = grtG`W1I;
ObjDef[ grt`W1I ][grtC`rootStr] = "W1I ";
ObjDef[ grt`W1I ][grtC`indexList] = {};
ObjDef[ grt`W1I ][grtC`calcFn] = grtF`calcRiemSq;
ObjDef[ grt`W1I ][grtC`calcFnParms] = {Weylstardndnupup, Weyldndnupup, 8};
ObjDef[ grt`W1I ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`W1I ][grtC`depends] = {grt`Weylstar[dn,dn,up,up], grt`Weyl[dn,dn,up,up]};

(*---------------------------
# W2I
#----------------------------*)

ObjDef[ grt`W2I ][grtC`header] = "CM invariant Im(W2)";
ObjDef[ grt`W2I ][grtC`root] = grtG`W2I;
ObjDef[ grt`W2I ][grtC`rootStr] = "W2I ";
ObjDef[ grt`W2I ][grtC`indexList] = {};
ObjDef[ grt`W2I ][grtC`calcFn] = grtF`calcW2;
ObjDef[ grt`W2I ][grtC`calcFnParms] = { Weylstardndnupup, Weyldndnupup,-4};
ObjDef[ grt`W2I ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`W2I ][grtC`depends] = {grt`Weyl[dn,dn,up,up],grt`Weylstar[dn,dn,up,up]};

(*---------------------------
# W2R
#----------------------------
#                   ab   cd   ef                   abkl ij   cdef
# Re(W2)= -1/64 *( C    C    C    - 3/4 e    e    C    C    C    )
#                     cd   ef   ab       abij efkl        cd
#
#                 (3)
#       = (-1/16)    C   (simpli]edn)
#
# get a factor of 1/8 from structure of the Do loops
#
*)

ObjDef[ grt`W2R ][grtC`header] = "CM invariant Re(W2)";
ObjDef[ grt`W2R ][grtC`root] = grtG`W2R;
ObjDef[ grt`W2R ][grtC`rootStr] = "W2R ";
ObjDef[ grt`W2R ][grtC`indexList] = {};
ObjDef[ grt`W2R ][grtC`calcFn] = grtF`calcW2;
ObjDef[ grt`W2R ][grtC`calcFnParms] = {Weyldndnupup,Weyldndnupup,-2};
ObjDef[ grt`W2R ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`W2R ][grtC`depends] = {grt`Weyl[dn,dn,up,up]};

grtF`calcW2[object_, iList_] :=
 Module[ {a,b,c,d,e,f, root1, root2, s},
 root1 = grtG`ObjDef[object][grtC`calcFnParms][[1]];
 root2 = grtG`ObjDef[object][grtC`calcFnParms][[2]];
 s = 0;
 Do[
   Do[
     Do[
       Do[
	If[ grtG`root1[metricName,a,b,c,d] =!= 0 ,
	 Do[
	   Do[
	     s = s + root1[metricName,a,b,c,d] * root2[metricName,c,d,e,f]
		    * root2[metricName,e,f,a,b],
	   {f, e+1, Ndim[metricName], 1}],
	 {e, Ndim[metricName]-1}];
	 ],
       { d, c+1, Ndim[metricName], 1}],
     {c, Ndim[metricName]-1}],
   {b, a+1, Ndim[metricName], 1}],
 {a, Ndim[metricName]-1}];
 Return[ s/grtG`ObjDef[object][grtC`calcFnParms][[3]] ];
]

(*---------------------------
# R1
# need to divide result by 4 !!
#----------------------------*)

ObjDef[ grt`R1 ][grtC`header] = "CM invariant R1";
ObjDef[ grt`R1 ][grtC`root] = grtG`R1;
ObjDef[ grt`R1 ][grtC`rootStr] = "R1 ";
ObjDef[ grt`R1 ][grtC`indexList] = {};
ObjDef[ grt`R1 ][grtC`calcFn] = grtF`calcsum2;
ObjDef[ grt`R1 ][grtC`calcFnParms] = Hold[grtG`Supdn[metricName,s1,s2] * grtG`Supdn[metricName,s2,s1]/4];
ObjDef[ grt`R1 ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`R1 ][grtC`depends] = {grt`S[up,dn]};

(*---------------------------
# R2
#----------------------------*)

ObjDef[ grt`R2 ][grtC`header] = "CM invariant R2";
ObjDef[ grt`R2 ][grtC`root] = grtG`R2;
ObjDef[ grt`R2 ][grtC`rootStr] = "R2 ";
ObjDef[ grt`R2 ][grtC`indexList] = {};
ObjDef[ grt`R2 ][grtC`calcFn] = grtF`calcR2;
ObjDef[ grt`R2 ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`R2 ][grtC`symPerm] = {};
ObjDef[ grt`R2 ][grtC`depends] = {grt`S[up,dn]};

grtF`calcR2[ object_, iList_] :=
 Module[ {i, j, k, s, s1},
  s = 0;
  Do[
    Do[
      If[ grtG`Supdn[metricName,i,j] =!= 0 ,
	s1 = 0;
	Do[
	  s = s + grtG`Supdn[metricName,i,j] * grtG`Supdn[metricName,j,k] *
		 grtG`Supdn[metricName,k,i];
	  s1 = s1 + grtG`Supdn[metricName,i,k] * grtG`Supdn[metricName,k,j],
	{k, Ndim[metricName]}];
	grtG`SA[metricName,i,j] = s1;
      ,(*else*)
	grtG`SA[metricName,i,j] = 0;
      ],
    {j, Ndim[metricName]}],
  {i, Ndim[metricName]}];
  Return[-s/8];
]

(*---------------------------
# R3
#----------------------------*)

ObjDef[ grt`R3 ][grtC`header] = "CM invariant R3";
ObjDef[ grt`R3 ][grtC`root] = grtG`R3;
ObjDef[ grt`R3 ][grtC`rootStr] = "R3 ";
ObjDef[ grt`R3 ][grtC`indexList] = {};
ObjDef[ grt`R3 ][grtC`calcFn] = grtF`calcsum2;
ObjDef[ grt`R3 ][grtC`calcFnParms] =
  Hold[ grtG`SA[metricName,s1,s2] * grtG`SA[metricName,s2,s1]/16];
ObjDef[ grt`R3 ][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`R3 ][grtC`depends] = {grt`S[up,dn], grt`R2};


End[]; (* end the global scope *)
