(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 


Begin["grtG`"]; (* set scope to global vars *)

(*----------------------------
# LevC[dn,dn,dn]
#----------------------------*)

ObjDef[ grt`LevC[dn,dn,dn] ][grtC`header] = "Levi-Civita Tensor";
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`root] = grtG`LevCdndndn;
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`rootStr] = "e ";
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`indexList] = {dn,dn,dn};
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`preCalcFn] = grtF`preCalcLevC3;
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`calcFn] = grtF`calcNull;
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`symmetry] = grtF`symLevC3;
ObjDef[ grt`LevC[dn,dn,dn] ][grtC`depends] = {grt`detg};

(*----------------------------
# LevC[up,up,up]
# (depends on LevC[dn,dn,dn] which actually calc's
#  LevC(up,up,up)
#----------------------------*)

ObjDef[ grt`LevC[up,up,up] ][grtC`header] = "Levi-Civita Tensor";
ObjDef[ grt`LevC[up,up,up] ][grtC`root] = grtG`LevCupupup;
ObjDef[ grt`LevC[up,up,up] ][grtC`rootStr] = "e ";
ObjDef[ grt`LevC[up,up,up] ][grtC`indexList] = {up,up,up};
ObjDef[ grt`LevC[up,up,up] ][grtC`symmetry] = grtF`symLevC3;
ObjDef[ grt`LevC[up,up,up] ][grtC`depends] = {grt`LevC[dn,dn,dn]};


(*
 * Note that LevC has to depend on the metric since it involves
 * the determinant of the metric
 *)


grtF`preCalcLevC3[ object_] :=
 Module[ {a, master, pList, i, j, k, g},
	If[ Ndim[ metricName] =!= 3,
	   Write[ grt`stream,"LevC[dn,dn,dn] may only be calculated for n=3"];
	   Abort[];
	];
	g = PowerExpand[ Sqrt[ -detg[metricName]]];
	Do[
	  Do[
	    Do[
		 grtG`LevCupupupup[metricName,i,j,k] =
		    - Signature[{i,j,k}] / g;  (* tensor *)
		 grtG`LevCdndndndn[metricName,i,j,k] =
		    Signature[{i,j,k}] * g,  (* tensor *)
	    {k,4}]
	  {j,4}],
	 {i,4}]

]

(*----------------------------
# LevC[dn,dn,dn,dn]
# only for n=4 !!
#----------------------------*)

ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`header] = "Levi-Civita Tensor";
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`root] = grtG`LevCdndndndn;
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`rootStr] = "e ";
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`indexList] = {dn,dn,dn,dn};
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`preCalcFn] = grtF`preCalcLevC4;
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`calcFn] = grtF`calcNull;
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`symmetry] = grtF`symLevC4;
ObjDef[ grt`LevC[dn,dn,dn,dn] ][grtC`depends] = {grt`detg};

(*----------------------------
# LevC[up,up,up,up]
# only for n=4 !!
# (depends on LevC[dn,dn,dn,dn] which actually calc's
#  LevC[up,up,up,up] )
#----------------------------*)

ObjDef[ grt`LevC[up,up,up,up] ][grtC`header] = "Levi-Civita Tensor";
ObjDef[ grt`LevC[up,up,up,up] ][grtC`root] = grtG`LevCupupupup;
ObjDef[ grt`LevC[up,up,up,up] ][grtC`rootStr] = "e ";
ObjDef[ grt`LevC[up,up,up,up] ][grtC`indexList] = {up,up,up,up};
ObjDef[ grt`LevC[up,up,up,up] ][grtC`symmetry] = grtF`symLevC4;
ObjDef[ grt`LevC[up,up,up,up] ][grtC`depends] = {grt`LevC[dn,dn,dn,dn]};


(*
 * Note that LevC has to depend on the metric since it involves
 * the determinant of the metric
 *)

grtF`preCalcLevC4[ object_] :=
 Module[ {a, master, pList, i, j, k, l, g},
	If[ Ndim[metricName] =!= 4,
	   Write[ grt`stream,"LevC[dn,dn,dn,dn] may only be calculated for n=4"];
	   Abort[];
	];
	g = PowerExpand[ Sqrt[ -detg[metricName]]];
	Do[
	  Do[
	    Do[
	      Do[
		 grtG`LevCupupupup[metricName,i,j,k,l] =
		    - Signature[{i,j,k,l}] / g;  (* tensor *)
		 grtG`LevCdndndndn[metricName,i,j,k,l] =
		    Signature[{i,j,k,l}] * g,  (* tensor *)
	       {l,4}],
	     {k,4}],
	   {j,4}],
	 {i,4}]


]


(*----------------------------
# Weylstar(dn,dn,dn,dn)
#----------------------------
#                          ij
# Weylstar     = 1/2 e    C      =  C *
#         abcd        ijab   cd       ab cd
#
# (note this is the same as;
#                          ij
#             1/2 e     C
#                  ijab  cd
#
#  if the identities in the manual are usedn)
#
#-----------------------------*)

ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`header] = "Dual Weyl (first two indices dual)";
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`root] = grtG`Weylstardndndndn;
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`indexList] = {dn,dn,dn,dn};
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`calcFn] = grtF`calcdual;
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`calcFnParms] = {LevCdndndndn, Weyldndnupup};
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`symmetry] = grtF`symRiem;
ObjDef[ grt`Weylstar[dn,dn,dn,dn] ][grtC`depends] = {grt`LevC[dn,dn,dn,dn], grt`Weyl[dn,dn,up,up]};

grtF`calcdual[ object_, iList_] :=

 (*
  * Loop structure assumes one of the root objects
  * is perm like (i.e. if p=q then result is zero, and
  * can sum over upper diagonal only since defintion
  * of Weylstar has a 1/2 in it.)
  *)
 Module[ {p,q, s, root1, root2},
  root1 = grtG`ObjDef[object][grtC`calcFnParms][[1]];
  root2 = grtG`ObjDef[object][grtC`calcFnParms][[2]];
  s = 0;
  Do[
    Do[
      s = s + grtG`root1[metricName,p,q,a1,a2] * grtG`root2[metricName,a3,a4,p,q],
     {q, p+1, grtG`Ndim[ grtG`metricName], 1}
    ],
   { p, grtG`Ndim[grtG`metricName]}
  ];
  Return[s];
]

(*----------------------------
# Weylstar(dn,dn,up,up)
#----------------------------
#        cd              ijcd  
# MCstar     = 1/2 e    C      
#       ab          ijab
#
*)

ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`header] = "Dual Weyl (dual w.r.t first two indices)";
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`root] = grtG`Weylstardndnupup;
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`indexList] = {dn,dn,up,up};
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`calcFn] = grtF`calcdualup;
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`calcFnParms] = {LevCdndndndn, Weylupupupup};
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`symmetry] = grtF`symMRiem;
ObjDef[ grt`Weylstar[dn,dn,up,up] ][grtC`depends] = {grt`LevC[dn,dn,dn,dn], grt`Weyl[up,up,up,up]};

(*----------------------------
# Cstar(up,up,up,up)
#----------------------------
#       abcd        ijab   cd
#     C *    = 1/2 e    C
#                        ij
*)

ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`header] = "Dual Weyl (dual w.r.t first two indices)";
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`root] = grtG`Weylstarupupupup;
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`rootStr] = "C* ";
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`indexList] = {up,up,up,up};
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`calcFn] = grtF`calcdualup;
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`calcFnParms] = {LevCupupupup, Weyldndnupup};
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`symmetry] = grtF`symRiem;
ObjDef[ grt`Weylstar[up,up,up,up] ][grtC`depends] = { grt`Weyl[dn,dn,up,up], grt`LevC[up,up,up,up] };

grtF`calcdualup[object_, iList_] :=
 Module[ {p,q, s, root1, root2},
   root1 = ObjDef[object][grtC`calcFnParms][[1]];
   root2 = ObjDef[object][grtC`calcFnParms][[2]];
   s = 0;
   Do[
    Do[
      s = s + grtG`root1[metricName,p,q,a1,a2] * grtG`root2[metricName,p,q,a3,a4],
     {q, p, grtG`Ndim[grtG`metricName]}
    ],
    {p, grtG`Ndim[grtG`metricName]-1}
   ];

  Return[s];
]

End[]; (* end the global scope *)
