(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"]; (* set scope to global vars *)

(*----------------------------
 * Chr[dn,dn,dn]
 *----------------------------
 *)


grtF`addObject[ grt`Chr[dn,dn,dn] ];
ObjDef[ grt`Chr[dn,dn,dn]][grtC`header] = "Christoffel symbol of the first kind (symmetric in first two indices)";
ObjDef[ grt`Chr[dn,dn,dn]][grtC`root] = Chrdndndn;
ObjDef[ grt`Chr[dn,dn,dn]][grtC`rootStr] = "Chr ";
ObjDef[ grt`Chr[dn,dn,dn]][grtC`indexList] = {dn,dn,dn};
ObjDef[ grt`Chr[dn,dn,dn]][grtC`calcFn] = grtF`calcsum0;
ObjDef[ grt`Chr[dn,dn,dn]][grtC`calcFnParms] =  Hold[ 1/2*gdndnpdn[metricName,a1,a3,a2]
  + 1/2*gdndnpdn[metricName,a2,a3,a1] - 1/2*gdndnpdn[metricName,a1,a2,a3] ];
ObjDef[ grt`Chr[dn,dn,dn]][grtC`symmetry] = grtF`symChr1;
ObjDef[ grt`Chr[dn,dn,dn]][grtC`depends] = {  grt`g[dn,dn,pdn]};


(*----------------------------
# Chr[dn,dn,up]
#----------------------------*)

grtF`addObject[ grt`Chr[dn,dn,up] ];
ObjDef[ grt`Chr[dn,dn,up]][grtC`header] ="Christoffel symbol of the second kind (symmetric in first two indices)";
ObjDef[ grt`Chr[dn,dn,up]][grtC`root] = Chrdndnup;
ObjDef[ grt`Chr[dn,dn,up]][grtC`rootStr] ="Chr ";
ObjDef[ grt`Chr[dn,dn,up]][grtC`indexList] = {dn,dn,up};
ObjDef[ grt`Chr[dn,dn,up]][grtC`calcFn] = grtF`calcsum1;
ObjDef[ grt`Chr[dn,dn,up]][grtC`calcFnParms] = (* raise last index *)
  Hold[ gupup[metricName,a3,s1] * Chrdndndn[metricName, a1,a2,s1] ];
ObjDef[ grt`Chr[dn,dn,up]][grtC`symmetry] =grtF`symChr1;
ObjDef[ grt`Chr[dn,dn,up]][grtC`depends] ={ grt`g[up,up],  grt`Chr[dn,dn,dn]};

(*----------------------------
 * constraint 
 *----------------------------*)

grtF`addObject[ grt`constraint ];
ObjDef[ grt`constraint][grtC`header] ="Constraints ";
ObjDef[ grt`constraint][grtC`root] = constraint; 
ObjDef[ grt`constraint][grtC`rootStr] = "constraints ";
ObjDef[ grt`constraint][grtC`indexList] = {};
ObjDef[ grt`constraint][grtC`calcFn] = grtF`calcNull; (* this is illegal this will cause an error *)
ObjDef[ grt`constraint][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`constraint][grtC`depends] = {};

(*----------------------------
# detmetric  detg
#----------------------------*)

grtF`addObject[ grt`detg];
ObjDef[ grt`detg ][grtC`header] ="Determinant of the metric tensor";
ObjDef[ grt`detg ][grtC`root] = detg;
ObjDef[ grt`detg ][grtC`rootStr] ="g ";
ObjDef[ grt`detg ][grtC`indexList] = {};
ObjDef[ grt`detg ][grtC`preCalcFn] = grtF`preCalcDetg;
ObjDef[ grt`detg ][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`detg ][grtC`depends] ={};

grtF`preCalcDetg[ object_] :=
Module[ {},
  (*
   * we may have been asked for detg, but if a basis
   * was loaded g(dn,dn) may not have been calc'd yet
   *)
   Which[ 
     grtF`checkIfAssigned[ grt`g[dn,dn] ],
        (*
         * normal tensor stuff
         *)
         grtF`calcDetg[ gdndn ],
     grtF`checkIfAssigned[ grt`eta[bup,bup] ], 
        (*
         * there is basis, so use it
         *)
         If[ !grtF`checkIfAssigned[ grt`g[up,up] ], 
	    grtF`core[ grt`g[up,up] ]
         ];
	 grtF`calcDetg[ gupup ],
     True,
	 Print["Load a metric or basis first."];
	 Abort[]
     ]
]

(*----------------------------
 * dimension
 *----------------------------*)

grtF`addObject[ grt`dimension ];
ObjDef[ grt`dimension][grtC`header] ="Dimension";
ObjDef[ grt`dimension][grtC`root] = Ndim; (* if change this must also change in grtF`initMetric() *)
ObjDef[ grt`dimension][grtC`rootStr] = "dimension ";
ObjDef[ grt`dimension][grtC`indexList] = {};
ObjDef[ grt`dimension][grtC`calcFn] = grtF`calcNull; (* this is illegal this will cause an error *)
ObjDef[ grt`dimension][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`dimension][grtC`depends] ={};


(*----------------------------
 * ds
 *----------------------------*)

grtF`addObject[ grt`ds ];
ObjDef[ grt`ds][grtC`header] ="Line element";
ObjDef[ grt`ds][grtC`root] = ds; (* if change this must also change in grtF`initMetric() *)
ObjDef[ grt`ds][grtC`rootStr] = SequenceForm[ "ds ", Superscript[2]];
ObjDef[ grt`ds][grtC`indexList] = {};
ObjDef[ grt`ds][grtC`calcFn] =grtF`calcds; (* this is illegal this will cause an error *)
ObjDef[ grt`ds][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`ds][grtC`depends] ={  grt`g[dn,dn]};

grtF`calcds[object_, iList_] :=
  Module[ { dsvalue, i, j},
     dsvalue = 0;
     For[ i=1, i <= grtG`Ndim[grtG`metricName], i++,
	    dsvalue = dsvalue + gdndn[grtG`metricName, i,i] *
		ToExpression[StringInsert["d",ToString[xup[grtG`metricName,i]],2]]^2;
	    For[ j = i+1, j <= grtG`Ndim[grtG`metricName], j++,
	       dsvalue = dsvalue + 2 * gdndn[grtG`metricName,i,j] *
		ToExpression[StringInsert["d",ToString[xup[grtG`metricName,i]],2]] *
		ToExpression[StringInsert["d",ToString[xup[grtG`metricName,j]],2]];
	    ];
     ];
    Return[ dsvalue];
 ]

(*----------------------------
 * g[dn,dn]
 *----------------------------*)

grtF`addObject[ grt`g[dn,dn] ];
ObjDef[ grt`g[dn,dn]][grtC`header] ="Covariant metric tensor";
ObjDef[ grt`g[dn,dn]][grtC`root] = gdndn;
ObjDef[ grt`g[dn,dn]][grtC`rootStr] = "g ";
ObjDef[ grt`g[dn,dn]][grtC`indexList] = {dn,dn};
ObjDef[ grt`g[dn,dn]][grtC`preCalcFn] = grtF`preCalcgdndn; 
ObjDef[ grt`g[dn,dn]][grtC`calcFnParms] = {};
ObjDef[ grt`g[dn,dn]][grtC`symmetry] = grtF`symsym2;
ObjDef[ grt`g[dn,dn]][grtC`depends] = {};

grtF`preCalcgdndn[ object_] := 
Module[ {i,j,a,b},

  Which[

    grtF`checkIfAssigned[ grt`e[bdn,dn] ] ||
    grtF`checkIfAssigned[ grt`e[bup,dn] ],
      If[ !grtF`checkIfAssigned[ grt`e[bup,dn] ],
	 grtF`core[ grt`e[bup,dn] ]
      ];
      If[ !grtF`checkIfAssigned[ grt`e[bdn,dn] ],
	 grtF`core[ grt`e[bdn,dn] ]
      ];
      Do[
	gdndn[metricName,a,a] =
	      Sum[ ebdndn[metricName,i,a] * ebupdn[metricName,i,a],
		{i,Ndim[metricName]}];
	Do[
	  gdndn[metricName,b,a] = gdndn[metricName,a,b];
	  gdndn[metricName,a,b] = 
	      Sum[ ebdndn[metricName,i,a] * ebupdn[metricName,i,b],
		{i,Ndim[metricName]}]
          ,{b,a+1,Ndim[metricName]}]
      ,{a,Ndim[metricName]}],
    grtF`checkIfAssigned[ grt`e[bdn,up] ] ||
    grtF`checkIfAssigned[ grt`e[bup,up] ] ||
    grtF`checkIfAssigned[ grt`g[up,up] ],
      If[ !grtF`checkIfAssigned[g[up,up]],
	  grtF`core[ grt`g[up,up] ]
      ];
      If[ !grtF`checkIfAssigned[ detg],
	  grtF`core[ grt`detg]
      ];
      grtF`invMetric[ gdndn, gupup],
    True,
      Print["Please load a metric or basis"];
      Abort[]
    ]

]

(*----------------------------
# g[up,up]
#----------------------------*)

grtF`addObject[ grt`g[up,up] ];
ObjDef[ grt`g[up,up] ][grtC`header] ="Contravariant metric tensor";
ObjDef[ grt`g[up,up] ][grtC`root] = gupup;
ObjDef[ grt`g[up,up] ][grtC`rootStr] ="g ";
ObjDef[ grt`g[up,up] ][grtC`indexList] = {up,up};
ObjDef[ grt`g[up,up] ][grtC`preCalcFn] =grtF`preCalcguu;
ObjDef[ grt`g[up,up] ][grtC`symmetry] =grtF`symsym2;
ObjDef[ grt`g[up,up] ][grtC`depends] ={};


grtF`preCalcguu[ object_] := 
Module[ {i,j,a,b},
  Which[
    (* if we have a tetrad with a tensor index up then build
     * g[up,up] by using gupup = eta b b
     *)
    grtF`checkIfAssigned[ grt`e[bdn,up] ] ||
    grtF`checkIfAssigned[ grt`e[bup,up] ],
      If[ !grtF`checkIfAssigned[ grt`e[bup,up] ],
	 grtF`core[ grt`e[bup,up] ]
      ];
      If[ !grtF`checkIfAssigned[ grt`e[bdn,up] ],
	 grtF`core[ grt`e[bdn,up] ]
      ];
      Do[
	gupup[metricName,a,a] =
	      Sum[ ebdnup[metricName,i,a] * ebupup[metricName,i,a],
		{i,Ndim[metricName]}];
	Do[
	  gupup[metricName,b,a] = gupup[metricName,a,b];
	  gupup[metricName,a,b] = 
	      Sum[ ebdnup[metricName,i,a] * ebupup[metricName,i,b],
		{i,Ndim[metricName]}]
          ,{b,a+1,Ndim[metricName]}]
      ,{a,Ndim[metricName]}],
    (*
     * if there's a tetrad with index down, calc gdndn and then invert 
     *)
    grtF`checkIfAssigned[ grt`e[bdn,dn] ] ||
    grtF`checkIfAssigned[ grt`e[bup,dn] ] ||
    grtF`checkIfAssigned[ grt`g[dn,dn] ],
      If[ !grtF`checkIfAssigned[g[dn,dn]],
	  grtF`core[ grt`g[dn,dn] ]
      ];
      grtF`invMetric[ gupup, gdndn],
    True,
      Print["Please load a metric or basis"];
      Abort[]
    ];

]

(*----------------------------------
* g[dn,dn,pdn] (partial w.r.t last)
*---------------------------------*)

grtF`addObject[ grt`g[dn,dn,pdn] ];
ObjDef[ grt`g[dn,dn,dn]][grtC`header] ="First derivitive of the metric (diff w.r.t last index)";
ObjDef[ grt`g[dn,dn,pdn]][grtC`root] = gdndnpdn;
ObjDef[ grt`g[dn,dn,pdn]][grtC`rootStr] ="g ";
ObjDef[ grt`g[dn,dn,pdn]][grtC`indexList] = {dn,dn,pdn};
ObjDef[ grt`g[dn,dn,pdn]][grtC`calcFn] = grtF`calcsum0;
ObjDef[ grt`g[dn,dn,pdn]][grtC`calcFnParms] = Hold[ 
		D[gdndn[ metricName, a1, a2], xup[ metricName, a3 ]] ];
(* Hold[] on that D[] ! *)
ObjDef[ grt`g[dn,dn,pdn]][grtC`symmetry] =grtF`symChr1;
ObjDef[ grt`g[dn,dn,pdn]][grtC`depends] ={grt`g[dn,dn]};



(*----------------------------
# g[dn,dn,dn,dn] partial w.r.t last two
# - is this required anymore ?
#----------------------------*)

grtF`addObject[ grt`g[dn,dn,pdn,pdn] ];
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`header] ="Second derivitive of the metric (diff w.r.t last two indices)";
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`root] = gdndnpdnpdn;
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`rootStr] ="g ";
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`indexList] = {dn,dn,pdn,pdn};
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`calcFn] = grtF`calcsum0;
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`calcFnParms] =
    Hold[ D[ gdndnpdn[metricName, a1,a2,a3 ], xup[metricName, a4 ] ]];
(* Hold[] on that D[] ! *)
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`symmetry] =grtF`symd2;
ObjDef[ grt`g[dn,dn,pdn,pdn] ][grtC`depends] ={ grt`g[dn,dn,dn]};

(*----------------------------
# G[dn,dn]
#----------------------------*)

grtF`addObject[ grt`G[dn,dn] ];
ObjDef[ grt`G[dn,dn] ][grtC`header] ="Covariant Einstein";
ObjDef[ grt`G[dn,dn] ][grtC`root] = Gdndn;
ObjDef[ grt`G[dn,dn] ][grtC`rootStr] ="G ";
ObjDef[ grt`G[dn,dn] ][grtC`indexList] = {dn,dn};
ObjDef[ grt`G[dn,dn] ][grtC`calcFn] =grtF`calcsum0;
ObjDef[ grt`G[dn,dn] ][grtC`calcFnParms] =
  Hold[  Rdndn[metricName,a1,a2] - 1/2* gdndn[metricName,a1,a2] *
	 scalarR[metricName] ];
ObjDef[ grt`G[dn,dn] ][grtC`symmetry] =grtF`symsym2;
ObjDef[ grt`G[dn,dn] ][grtC`depends] ={ grt`g[dn,dn],  grt`Ricciscalar,  grt`R[dn,dn]};


(*----------------------------
 * Info 
 *----------------------------*)

grtF`addObject[ grt`Info ];
ObjDef[ grt`Info][grtC`header] ="Text Information";
ObjDef[ grt`Info][grtC`root] = Info; 
ObjDef[ grt`Info][grtC`rootStr] = "Text ";
ObjDef[ grt`Info][grtC`indexList] = {};
ObjDef[ grt`Info][grtC`calcFn] = grtF`enterComponent; 
ObjDef[ grt`Info][grtC`symmetry] = grtF`symscalar;
ObjDef[ grt`Info][grtC`depends] = {};


(*----------------------------
# R[dn,dn]  (DIRECT)
# - often faster to calcuate this than to contract Riemann (but not always)
#----------------------------*)

grtF`addObject[ grt`R[dn,dn] ];
ObjDef[ grt`R[dn,dn] ][grtC`header] ="Covariant Ricci";
ObjDef[ grt`R[dn,dn] ][grtC`root] = Rdndn;
ObjDef[ grt`R[dn,dn] ][grtC`rootStr] ="R ";
ObjDef[ grt`R[dn,dn] ][grtC`indexList] = {dn,dn};
ObjDef[ grt`R[dn,dn] ][grtC`calcFn] =grtF`calcRicciD;
ObjDef[ grt`R[dn,dn] ][grtC`symmetry] =grtF`symsym2;
ObjDef[ grt`R[dn,dn] ][grtC`depends] ={grt`Chr[dn,dn,up], grt`g[up,up]};

grtF`calcRicciD[ object_, index_] :=
Module[ {a,b,s},

 s = 0;
 Do[
   If[ a2 =!= a ,
      s = s + D[ Chrdndnup[metricName,a1,a2,a],xup[metricName,a] ]
	     - D[ Chrdndnup[metricName,a,a1,a],xup[metricName,a2] ];
   ];
   Do[
      s =s + Chrdndnup[metricName,a1,a2,b] * Chrdndnup[metricName,a,b,a]
	     - Chrdndnup[metricName,a1,a,b] * Chrdndnup[metricName,b,a2,a],
    {b, Ndim[metricName]}
   ],
  {a, Ndim[metricName]}
 ];
 Return[s];
]

(*----------------------------
# R[dn,dn,dn,dn]
#----------------------------*)

grtF`addObject[ grt`R[dn,dn,dn,dn] ];
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`header] ="Covariant Riemann";
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`root] = Rdndndndn;
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`rootStr] ="R ";
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`indexList] = {dn,dn,dn,dn};
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`calcFn] =grtF`calcRiem;
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`symmetry] =grtF`symRiem;
ObjDef[ grt`R[dn,dn,dn,dn] ][grtC`depends] ={  grt`g[dn,dn,pdn],  grt`Chr[dn,dn,dn], g[up,up]};


grtF`calcRiem [ objectName_, index_] :=
Module[ {m,n,s},
 s =0;
 (* can save total of 12 diff calls here (n=4) *)
 If[ a1 === a3 && a2 === a4,
    s = D[ gdndnpdn[metricName,a1,a2,a2], xup[metricName,a1] ];
 ,(*else*)
    s = 1/2*( D[gdndnpdn[metricName,a1,a4,a2] , xup[metricName,a3] ]
	  +  D[ gdndnpdn[metricName,a2,a3,a1] , xup[metricName,a4] ] );
 ];
 If[ a1 === a4 && a3 === a2 ,
     s = s - D[ gdndnpdn[metricName,a1,a3,a3] , xup[metricName,a1] ]
 ,(*else*)
     s = s - 1/2*( D[ gdndnpdn[metricName,a1,a3,a2], xup[metricName,a4] ]
	    + D[ gdndnpdn[metricName,a2,a4,a1], xup[metricName,a3] ] );
 ];
 Do[
   Do[
     s = s + gupup[metricName,m,n]*
	  (Chrdndndn[metricName,a1,a4,m] * Chrdndndn[metricName,a2,a3,n]
	  - Chrdndndn[metricName,a1,a3,m] * Chrdndndn[metricName,a2,a4,n]),
    { n, Ndim[metricName]}
   ],
  { m, Ndim[metricName] }
 ];
 Return[s];
]

(*----------------------------
# R[up,up,dn,dn]
# since this is used explicitly in invars, define
# it explicitly
#----------------------------*)

grtF`addObject[ grt`R[up,up,dn,dn] ];
ObjDef[ grt`R[up,up,dn,dn] ][grtC`header] ="Mixed Riemann";
ObjDef[ grt`R[up,up,dn,dn] ][grtC`root] = Rupupdndn;
ObjDef[ grt`R[up,up,dn,dn] ][grtC`rootStr] ="R ";
ObjDef[ grt`R[up,up,dn,dn] ][grtC`indexList] = {up,up,dn,dn};
ObjDef[ grt`R[up,up,dn,dn] ][grtC`calcFn] = grtF`calcsum2;
(* raise indices *)
ObjDef[ grt`R[up,up,dn,dn] ][grtC`calcFnParms] =
  Hold[ gupup[metricName,s1,a1] * gupup[metricName,s2,a2] *
  Rdndndndn[metricName,s1,s2,a3,a4] ];
ObjDef[ grt`R[up,up,dn,dn] ][grtC`symmetry] =grtF`symMRiem;
ObjDef[ grt`R[up,up,dn,dn] ][grtC`depends] ={ grt`g[up,up],  grt`R[dn,dn,dn,dn]};


(*----------------------------
# Ricciscalar
#----------------------------*)

grtF`addObject[ grt`Ricciscalar ];
ObjDef[ grt`Ricciscalar ][grtC`header] ="Ricci scalar";
ObjDef[ grt`Ricciscalar ][grtC`root] = scalarR;
ObjDef[ grt`Ricciscalar ][grtC`rootStr] ="R ";
ObjDef[ grt`Ricciscalar ][grtC`indexList] = {};
ObjDef[ grt`Ricciscalar ][grtC`calcFn] =grtF`calcscalarR;
ObjDef[ grt`Ricciscalar ][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`Ricciscalar ][grtC`depends] ={ grt`g[dn,dn],  grt`R[dn,dn]};

grtF`calcscalarR[ object_, index_] :=
Module[ {a, b, s},
  s = 0;
  Do[
    s = s + gupup[metricName, a, a] * Rdndn[metricName, a, a];
    Do[
       s = s + 2 * gupup[metricName, a, b] * Rdndn[metricName, a, b],
     { b, a+1,  Ndim[metricName], 1}
    ],
   { a, Ndim[metricName]}
  ];
  Return[s];
]



(*----------------------------
# RicciSq
#----------------------------*)

grtF`addObject[ grt`RicciSq ];
ObjDef[ grt`RicciSq ][grtC`header] ="Full Contraction of Ricci";
ObjDef[ grt`RicciSq ][grtC`root] = RicciSq;
ObjDef[ grt`RicciSq ][grtC`rootStr] ="RicciSq ";
ObjDef[ grt`RicciSq ][grtC`indexList] = {};
ObjDef[ grt`RicciSq ][grtC`calcFn] =grtF`calcRicciSq; # uses objectName to do WeylSq also!
ObjDef[ grt`RicciSq ][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`RicciSq ][grtC`depends] ={ grt`R[up,dn]};

grtF`calcRicciSq[ object_, index_] :=
 Module[ {i,j,s},
   s =0;
   Do[
      s = s + Rupdn[metricName,i,i] * Rupdn[metricName,i,i];
      Do[
	   s = s + 2 * Rupdn[metricName,i,j] * Rupdn[metricName,j,i],
       {j, i+1, Ndim[metricName], 1}
      ],
     {i, Ndim[metricName]}
   ];
   Return[s];
]

(*----------------------------
# RiemSq
#----------------------------*)

grtF`addObject[ grt`RiemSq ];
ObjDef[ grt`RiemSq ][grtC`header] ="Full Contraction of Riemann";
ObjDef[ grt`RiemSq ][grtC`root] = RiemSq;
ObjDef[ grt`RiemSq ][grtC`rootStr] ="K ";
ObjDef[ grt`RiemSq ][grtC`indexList] = {};
ObjDef[ grt`RiemSq ][grtC`calcFn] =grtF`calcRiemSq;
ObjDef[ grt`RiemSq ][grtC`calcFnParms] = {Rupupdndn,Rupupdndn};
ObjDef[ grt`RiemSq ][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`RiemSq ][grtC`depends] = { grt`R[up,up,dn,dn]};

grtF`calcRiemSq[ object_, index_] :=
Module[ {a,b,c,d,s,root1, root2},
(*
 * use two seperate roots to allow for the cmscalars W1R and W1I
 *)
 root1 = ObjDef[ object][grtC`calcFnParms][[1]];
 root2 = ObjDef[ object][grtC`calcFnParms][[2]];

 If[ Ndim[metricName] === 4 && fnCode === grtC`calc ,

 (* explicit formula for N=4 & calc *)
 s =
  4*(root1[metricName,1,2,1,2]*root2[metricName,1,2,1,2]+
  root1[metricName,1,2,1,3]*root2[metricName,1,3,1,2] +
  root1[metricName,1,2,1,4]*root2[metricName,1,4,1,2] +
  root1[metricName,1,2,2,3]*root2[metricName,2,3,1,2] +
  root1[metricName,1,2,2,4]*root2[metricName,2,4,1,2] +
  root1[metricName,1,2,3,4]*root2[metricName,3,4,1,2] +
  root1[metricName,1,3,1,2]*root2[metricName,1,2,1,3] +
  root1[metricName,1,3,1,3]*root2[metricName,1,3,1,3] +
  root1[metricName,1,3,1,4]*root2[metricName,1,4,1,3] +
  root1[metricName,2,4,1,2]*root2[metricName,1,2,2,4] +
  root1[metricName,2,4,1,3]*root2[metricName,1,3,2,4] +
  root1[metricName,2,4,1,4]*root2[metricName,1,4,2,4] +
  root1[metricName,2,4,2,3]*root2[metricName,2,3,2,4] +
  root1[metricName,2,4,2,4]*root2[metricName,2,4,2,4] +
  root1[metricName,1,3,2,3]*root2[metricName,2,3,1,3] +
  root1[metricName,1,3,2,4]*root2[metricName,2,4,1,3] +
  root1[metricName,1,3,3,4]*root2[metricName,3,4,1,3] +
  root1[metricName,1,4,1,2]*root2[metricName,1,2,1,4] +
  root1[metricName,1,4,1,3]*root2[metricName,1,3,1,4] +
  root1[metricName,1,4,1,4]*root2[metricName,1,4,1,4] +
  root1[metricName,1,4,2,3]*root2[metricName,2,3,1,4] +
  root1[metricName,1,4,2,4]*root2[metricName,2,4,1,4] +
  root1[metricName,1,4,3,4]*root2[metricName,3,4,1,4] +
  root1[metricName,2,3,1,2]*root2[metricName,1,2,2,3] +
  root1[metricName,2,3,1,3]*root2[metricName,1,3,2,3] +
  root1[metricName,2,3,1,4]*root2[metricName,1,4,2,3] +
  root1[metricName,2,3,2,3]*root2[metricName,2,3,2,3] +
  root1[metricName,2,3,2,4]*root2[metricName,2,4,2,3] +
  root1[metricName,2,3,3,4]*root2[metricName,3,4,2,3] +
  root1[metricName,2,4,3,4]*root2[metricName,3,4,2,4] +
  root1[metricName,3,4,1,2]*root2[metricName,1,2,3,4] +
  root1[metricName,3,4,1,3]*root2[metricName,1,3,3,4] +
  root1[metricName,3,4,1,4]*root2[metricName,1,4,3,4] +
  root1[metricName,3,4,2,3]*root2[metricName,2,3,3,4] +
  root1[metricName,3,4,2,4]*root2[metricName,2,4,3,4] +
  root1[metricName,3,4,3,4]*root2[metricName,3,4,3,4])

  ,(* else *)
    (* use this for calcAlter or Ndim != 4 *)
	s =0;
	Do[
	  Do[
	    Do[
	      Do[
		s =s + 4 * root1[metricName,a,b,c,d]* root2[metricName,c,d,a,b],
	      {d, c+1, Ndim[metricName], 1}],
	    {c, Ndim[metricName]-1}];
	    If[ fnCode === grtC`calcalter,
		s = simpHow[ s]
	    ],
	  { b , a+1, Ndim[metricName], 1}],
	{a, Ndim[metricName]-1}]
    ];
    (* W1I requires a division of the result *)
    If[ Length[ ObjDef[ object][grtC`calcFnParms] ] === 3 ,
       s = s/ObjDef[ object][grtC`calcFnParms][[3]]
    ];
    Return[s];
]

(*----------------------------
# S[dn,dn]
# - trace free Ricci
#----------------------------*)

grtF`addObject[ grt`S[dn,dn] ];
ObjDef[ grt`S[dn,dn] ][grtC`header] ="Covariant trace-free Ricci";
ObjDef[ grt`S[dn,dn] ][grtC`root] = Sdndn;
ObjDef[ grt`S[dn,dn] ][grtC`rootStr] ="S ";
ObjDef[ grt`S[dn,dn] ][grtC`indexList] = {dn,dn};
ObjDef[ grt`S[dn,dn] ][grtC`calcFn] = grtF`calcsum0;
ObjDef[ grt`S[dn,dn] ][grtC`calcFnParms] =	Hold[ Rdndn[metricName,a1,a2] -
	(1/4)* gdndn[metricName,a1,a2] * scalarR[metricName] ];
ObjDef[ grt`S[dn,dn] ][grtC`symmetry] =grtF`symsym2;
ObjDef[ grt`S[dn,dn] ][grtC`depends] ={ grt`R[dn,dn],  grt`g[dn,dn],  grt`Ricciscalar};

(*----------------------------
# Weyl[dn,dn,dn,dn]
#----------------------------*)

grtF`addObject[ grt`Weyl[dn,dn,dn,dn] ];
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`header] ="Covariant Weyl";
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`root] = Weyldndndndn;
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`rootStr] ="Weyl ";
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`indexList] = {dn,dn,dn,dn};
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`calcFn] =grtF`calcWeyl;
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`symmetry] =grtF`symRiem;
ObjDef[ grt`Weyl[dn,dn,dn,dn] ][grtC`depends] ={grt`R[dn,dn,dn,dn], grt`Ricciscalar};

grtF`calcWeyl [ objectName_, index_] :=
 Module[ {s},
   s = (Rdndndndn[metricName,a1,a2,a3,a4]
     -1/(Ndim[metricName]-2)*
		 ( gdndn[metricName,a1,a3] * Rdndn[metricName,a4,a2] -
		   gdndn[metricName,a1,a4] * Rdndn[metricName,a3,a2] -
		   gdndn[metricName,a2,a3] * Rdndn[metricName,a4,a1] +
		   gdndn[metricName,a2,a4] * Rdndn[metricName,a3,a1]) +
	    1/((Ndim[metricName] - 1)*(Ndim[metricName] - 2)) *
	    scalarR[metricName]*
	    ( gdndn[metricName,a1,a3] * gdndn[metricName,a4,a2] -
	      gdndn[metricName,a1,a4] * gdndn[metricName,a3,a2] )
	);
 Return[s];
]

(*----------------------------
# Weyl[dn,dn,up,up]
# since this is used explicitly in invars,
# define it explicitly (slightly faster)
#----------------------------*)

grtF`addObject[ grt`Weyl[dn,dn,up,up] ];
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`header] ="Mixed Weyl";
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`root] = Weyldndnupup;
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`rootStr] ="Weyl ";
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`indexList] = {up,up,dn,dn};
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`calcFn] =grtF`calcsum2;
(* raise indices *)
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`calcFnParms] = Hold[gupup[metricName,s1,a3]
    * gupup[metricName,s2,a4] *  Weyldndndndn[metricName,a1,a2,s1,s2] ];
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`symmetry] =grtF`symMRiem;
ObjDef[ grt`Weyl[dn,dn,up,up] ][grtC`depends] ={grt`g[up,up], grt`Weyl[dn,dn,dn,dn]};

(*----------------------------
# WeylSq
#----------------------------*)

grtF`addObject[ grt`WeylSq ];
ObjDef[ grt`WeylSq ][grtC`header] ="Full Contraction of Weyl";
ObjDef[ grt`WeylSq ][grtC`root] = WeylSq;
ObjDef[ grt`WeylSq ][grtC`rootStr] ="W ";
ObjDef[ grt`WeylSq ][grtC`indexList] = {};
ObjDef[ grt`WeylSq ][grtC`calcFn] = grtF`calcRiemSq; (* uses parms to do WeylSq also *)
ObjDef[ grt`WeylSq ][grtC`calcFnParms] = {Weyldndnupup,Weyldndnupup};
ObjDef[ grt`WeylSq ][grtC`symmetry] =grtF`symscalar;
ObjDef[ grt`WeylSq ][grtC`depends] ={ grt`Weyl[dn,dn,up,up]};

(*----------------------------
# x[up]
#----------------------------*)

grtF`addObject[ grt`x[up] ];
ObjDef[ grt`x[up] ][grtC`header] ="Coordinates";
ObjDef[ grt`x[up] ][grtC`root] = xup;
ObjDef[ grt`x[up] ][grtC`rootStr] ="x ";
ObjDef[ grt`x[up] ][grtC`indexList] = {up};
ObjDef[ grt`x[up] ][grtC`calcFn] =grtF`calcNULL; # assigned by grload
ObjDef[ grt`x[up] ][grtC`symmetry] = grtF`symvector;
ObjDef[ grt`x[up] ][grtC`depends] = {};



End[]; (* end the global scope *)
