(* 
* GRTensorM (c) 1996-98 by Peter Musgrave, Denis Pollney and Kayll Lake
* All rights Reserved. Update by Mustapha Ishak B.
* GRTensorM is provided free of charge on the condition that it is
* acknowledged that we accept no liability for software performance,
* continued maintenance, or damage to data. The authors retain any and
* all rights to the software known as GRTensor, GRTensorII and
* GRTensorM their documentation and distribution. 
*) 

Begin["grtG`"]; (* globals are default (Private) context *)

allSymmetries = {};

(*----------------------------------------------------------
* grtF`findSymmetry(permSet,numIndices)
*
* - given a set of permutations, see if any of the defined
* symmetries match. If not then assign a general routine
* based on the number of indices.
*----------------------------------------------------------*)

grtF`findSymmetry[ permSet_ ,num_] :=
 Module[ {a,b,c, found},
  Which[
    num === 0 ,
       Return[grtF`symscalar],
    num === 1 ,
       Return[grtF`symvector],
    True,
       (* scan the existing symmetries (as listed in allSymmetries)  *)
       Do[
	 (*
	  * We have found a symmetry routine matching the permSet
	  * provided that:
	  *     permSet - Candidate = empty set
	  *     Candidate - permSet = empty set
	  *)

	 If[ (Complement[ permSet, symmetry[ allSymmetries[[a]] ] ] === {} )
	     && (Complement[ symmetry[ allSymmetries[[a]] ], permSet] === {}),
	   found = allSymmetries[[a]];
	   Break[];
	 ],
	{ a, Length[ allSymmetries] }
       ];
       (* didn't get a match so assign the general routine *)
       If[ !ValueQ[ found],
	 found = ToExpression[ ToString[ SequenceForm[
	      grtF`symnosym, num ]]]
       ];
       Return[ found];

   ];
 ]


(**** scalar ***)

symmetry[grtF`symscalar] = {{}};
grtF`addSymmetry[ grtF`symscalar];



grtF`symscalar[objectName_, root_, calcFn_] :=
  Module[ {},
	If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ],
	   root[metricName] = calcFn[objectName,{}];
	];
	If[ simp ,
	   root[metricName] = simpHow[ root[metricName] ];
	];
	If[ callComp ,
	   grtF`component[ objectName,{},root[metricName] ];
	];
	Return[NULL];
  ] (* return NULL *)

(**** vector ***)
symmetry[grtF`symvector] = {{}};

grtF`addSymmetry[ grtF`symvector];
(* need this name for newindices and makeD *)
symmetry[grtF`symnosym1] = {{}};
grtF`symnosym1[ a_, b_, c_] := grtF`symvector[a,b,c];
grtF`addSymmetry[ grtF`symnosym1];

grtF`symvector [objectName_, root_, calcFn_] :=

  Module[ {seq},
  (* 2 index symmetric object *)

 (* for aesthetic reasons, display the coordinates on a single line *)
 If[ (grtG`fnCode === grtC`display) && (objectName === x[up]),
    (* display coord on a single line *)
    seq = Sequence[ " x", Superscript[1], " = ",grtG`xup[metricName,1] ];
    Do[
       seq = Sequence[ seq, ",  x", Superscript[a1], " = ",grtG`xup[metricName,a1] ],
    {a1, 2, Ndim[metricName]}];
    Write[grt`stream, seq];
    grtG`displayZero = False;
 , (* else *)
  Do[
      If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ],
	 root[metricName, a1] = calcFn[objectName,{a1}];
      ];
      If[ simp ,
	root[metricName,a1] =
		simpHow[ root[ metricName,a1] ];
      ];
      If[ callComp ,
	 grtF`component[objectName,{a1},root[metricName,a1] ];
      ],
     { a1, Ndim[metricName] }
  ];

 ];
 Return[NULL];

 ] (* return NULL *)

(**** 2 index symmetric ***)
symmetry[grtF`symsym2] = {{{2,1},1}};
grtF`addSymmetry[ grtF`symsym2];

grtF`symsym2 [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	   (* assignments here reflect symmetry *)
	   If[ a1 =!= a2 ,
	      root[metricName,a2,a1] =
		    root[metricName,a1,a2];
	   ];
	   (* assignement *)
	   root[metricName,a1,a2] =
		   calcFn[objectName,{a1,a2}];
	];
	If[ simp ,
	   root[metricName,a1,a2] =
		simpHow[ root[metricName,a1,a2] ];
	];
	If[ callComp ,
	   grtF`component[ objectName,{a1,a2},root[metricName,a1,a2] ];
	],
      { a2, a1, Ndim[metricName], 1 }],
    { a1, Ndim[metricName] }];
    Return[NULL];
] (* return NULL *)

(**** two index - no symmetry ***)

symmetry[grtF`symnosym2] = {{}};
grtF`addSymmetry[ grtF`symnosym2];

grtF`symnosym2 [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	  root[metricName,a1,a2] =
		calcFn[objectName,{a1,a2}];
	];
	If[ simp ,
	  root[metricName,a1,a2] =
		simpHow[  root[metricName,a1,a2] ];
	];
	If[ callComp ,
	   grtF`component[ objectName,{a1,a2}, root[metricName,a1,a2] ];
	],
      { a2, Ndim[metricName] }],
    { a1, Ndim[metricName] }];
    Return[NULL];
  ]

(**** three index, symmetric in first two *** d1metric, Chr1 etc. *)
symmetry[grtF`symChr1] = {{{2,1,3},1}};
grtF`addSymmetry[ grtF`symChr1];

grtF`symChr1 [objectName_, root_, calcFn_] :=
 Module[ {},
  (* three index objects, sym w.r.t first two *)
    Do[
      Do[
	Do[
	  If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	     (* assignments here reflect symmetry *)
	     If[ a1 =!= a2 ,
		root[metricName,a2,a1,a3] =
			root[metricName,a1,a2,a3];
	     ];
	     (* assignment *)
	     root[metricName,a1,a2,a3] =
		calcFn[objectName,{a1,a2,a3}];
	  ];
	  If[ simp ,
	     root[metricName,a1,a2,a3] =
		simpHow[ root[metricName,a1,a2,a3] ];
	  ];
	  If[ callComp ,
	     grtF`component[ objectName,{a1,a2,a3}, root[metricName,a1,a2,a3] ];
	  ],
	{ a3, Ndim[metricName] }],
      { a2, a1, Ndim[metricName], 1}],
    { a1, Ndim[metricName] }];
    Return[NULL];
  ] (* return NULL *)

(**** 3 indices, anti-symmetric in 12 ***)
grtG`symmetry[ grtF`sym3a12] = {{{2,1,3},-1}};
grtF`addSymmetry[ grtF`sym3a12];

grtF`sym3a12[objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	If[ calc,
		root[metricName,a1,a1,a3] = 0
        ];
	Do[
	  If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
		root[metricName,a2,a1,a3] = -root[metricName,a1,a2,a3];
	        root[metricName,a1,a2,a3] =
		     calcFn[objectName,{a1,a2,a3}] 
	  ];
	  If[ simp ,
	     root[metricName,a1,a2,a3] =
		simpHow[  root[metricName,a1,a2,a3] ];
	  ];
	  If[ callComp ,
	     grtF`component[ objectName,{a1,a2,a3}, root[metricName,a1,a2,a3] ];
	  ],
	{ a2, a1+1, Ndim[metricName] }],
      { a1, Ndim[metricName] }],
    { a3, Ndim[metricName] }];
    Return[NULL];
 ] (* return NULL *)

(**** 3 indices, anti-symmetric in 13 ***)

grtG`symmetry[ grtF`sym3a13] = {{{3,2,1},-1}};

grtF`sym3a13[objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	If[ calc,
		root[metricName,a1,a2,a1] = 0
        ];
	Do[
	  If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
		root[metricName,a3,a2,a1] = -root[metricName,a1,a2,a3];
	        root[metricName,a1,a2,a3] =
		     calcFn[objectName,{a1,a2,a3}] 
	  ];
	  If[ simp ,
	     root[metricName,a1,a2,a3] =
		simpHow[  root[metricName,a1,a2,a3] ];
	  ];
	  If[ callComp ,
	     grtF`component[ objectName,{a1,a2,a3}, root[metricName,a1,a2,a3] ];
	  ],
	{ a3, a1+1, Ndim[metricName] }],
      { a1, Ndim[metricName] }],
    { a2, Ndim[metricName] }];
    Return[NULL];
 ] (* return NULL *)

(**** 3 indices, no symmetry ***)
grtG`symmetry[ grtF`symnosym3] = {{}};
grtF`addSymmetry[ grtF`symnosym3];

grtF`symnosym3 [objectName_, root_, calcFn_] :=
  (* three index objects, sym w.r.t first two *)
  Module[ {},
    Do[
      Do[
	Do[
	  If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	     root[metricName,a1,a2,a3] =
		calcFn[objectName,{a1,a2,a3}];
	  ];
	  If[ simp ,
	     root[metricName,a1,a2,a3] =
		simpHow[  root[metricName,a1,a2,a3] ];
	  ];
	  If[ callComp ,
	     grtF`component[ objectName,{a1,a2,a3}, root[metricName,a1,a2,a3] ];
	  ],
	{ a3, Ndim[metricName] }],
      { a2, Ndim[metricName] }],
    { a1, Ndim[metricName] }];
    Return[NULL];
 ] (* return NULL *)

(**** four index, symetric in first pair and second pair *** d2metric *)
symmetry[grtF`symd2] = { {{2,1,3,4},1}, {{1,2,4,3},1} };
grtF`addSymmetry[ grtF`symd2];

grtF`symd2 [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	Do[
	  Do[
	    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	      (* assignments here reflect symmetry *)
	      If[ a1 =!= a2 ,
		 root[metricName,a2,a1,a3,a4] =
			root[metricName,a1,a2,a3,a4];
		 If[ a3 =!= a4 ,
		    root[metricName,a2,a1,a4,a3] =
			root[metricName,a1,a2,a3,a4];
		 ];
	      ];
	      If[ a3 =!= a4 ,
		 root[metricName,a1,a2,a4,a3] =
			root[metricName,a1,a2,a3,a4];
	      ];
	      (* assignment *)
	      root[metricName,a1,a2,a3,a4] =
		calcFn[objectName,{a1,a2,a3,a4}];
	    ];
	    If[ simp ,
	      root[metricName,a1,a2,a3,a4] =
		      simpHow[ root[metricName,a1,a2,a3,a4] ];
	    ];
	    If[ callComp ,
	      grtF`component[ objectName,{a1,a2,a3,a4},
		root[metricName,a1,a2,a3,a4] ];
	    ],
	  { a4, a3,  Ndim[metricName], 1 }],
	{ a3, Ndim[metricName] }],
      { a2, a1, Ndim[metricName], 1}],
    { a1, Ndim[metricName] }];
    Return[NULL];
  ]  (* return NULL *)

(**** four index - symmetry like the Riemann tensor ***)

symmetry[grtF`symRiem] = { {{2,1,3,4},-1}, {{1,2,4,3},-1},{{3,4,1,2},1} };
grtF`addSymmetry[ grtF`symRiem];

grtF`symRiem [objectName_, root_, calcFn_] :=
 Module[ {x},
    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ],
    (* by symmetry we know a number of terms are zero *)
      Do[
	Do[
	  Do[
	       root[metricName,a1,a1,a2,a3] = 0;
	       root[metricName,a1,a2,a3,a3] = 0,
	  { a3, Ndim[metricName] }],
	{ a2, Ndim[metricName] }],
      { a1, Ndim[metricName] }];
    ];
    Do[
      Do[
	Do[
	  If[ a3 === a1 , x = a2 , x = a3+1 ];
	  Do[
	      If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
		 (* assignments here reflect symmetries *)
		 root[metricName,a2,a1,a3,a4] =
			-root[metricName,a1,a2,a3,a4];
		 root[metricName,a1,a2,a4,a3] =
			-root[metricName,a1,a2,a3,a4];
		 root[metricName,a2,a1,a4,a3] =
			root[metricName,a1,a2,a3,a4];
		 (* now with first & second pair exchanged *)
		 root[metricName,a3,a4,a1,a2] =
			root[metricName,a1,a2,a3,a4];
		 root[metricName,a3,a4,a2,a1] =
			-root[metricName,a1,a2,a3,a4];
		 root[metricName,a4,a3,a1,a2] =
			-root[metricName,a1,a2,a3,a4];
		 root[metricName,a4,a3,a2,a1] =
			root[metricName,a1,a2,a3,a4];
		(*
		 * put assigment AFTER sym ,so when simplification occurs
		 * the symmetrized objects don't need reassigment
		 *)
		 root[metricName,a1,a2,a3,a4] =
			calcFn[objectName,{a1,a2,a3,a4}];
	      ];
	      If[ simp ,
		 root[metricName,a1,a2,a3,a4] =
		     simpHow[  root[metricName,a1,a2,a3,a4] ];
	      ];
	      If[ callComp ,
		 grtF`component[ objectName,{a1,a2,a3,a4},
			root[metricName,a1,a2,a3,a4] ];
	      ],
	  { a4, x, Ndim[metricName], 1 }],
	{ a3, a1, Ndim[metricName]-1, 1 }],
      { a2, a1+1, Ndim[metricName], 1 }],
    { a1, Ndim[metricName]-1 }];
 Return[NULL];
 ] (* return NULL *)

(**** four index - symmetry of Natural (1,3) Riemann tensor ***)

symmetry[grtF`symNRiem] = { {{1,2,4,3},-1} };
grtF`addSymmetry[ grtF`symNRiem];

grtF`symNRiem [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	Do[
	  Do[
	    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	      (* zero by symmetry *)
	      root[metricName,a1,a2,a3,a3] = 0;

	      (* assignments here reflect symmetry *)

	      root[metricName,a1,a2,a4,a3] =
		-root[metricName,a1,a2,a3,a4];
	      (* assignment *)
	      root[metricName,a1,a2,a3,a4] =
		calcFn[objectName,{a1,a2,a3,a4}];
	    ];
	    If[ simp ,
	       root[metricName,a1,a2,a3,a4] =
		   simpHow[  root[metricName,a1,a2,a3,a4] ];
	    ];
	    If[ callComp ,
	       grtF`component[ objectName,{a1,a2,a3,a4},
		   root[metricName,a1,a2,a3,a4] ];
	    ],
	  { a4, a3, Ndim[metricName], 1 }],
	{ a3, Ndim[metricName] }],
      { a2, Ndim[metricName] }],
    { a1, Ndim[metricName] }];
   Return[NULL];
  ] (* return NULL *)

(**** four index symmetry of (2,2) Riemann tensor ****)

symmetry[grtF`symMRiem] = { {{2,1,3,4},-1}, {{1,2,4,3},-1} };
grtF`addSymmetry[ grtF`symMRiem];

grtF`symMRiem[objectName_, root_, calcFn_] :=
  Module[ {},
    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ],
      (* by symmetry we know a number of terms are zero *)
      Do[
	Do[
	  Do[
	       root[metricName,a1,a1,a2,a3] = 0;
	       root[metricName,a1,a2,a3,a3] = 0,
	  { a3, Ndim[metricName] }],
	{ a2, Ndim[metricName] }],
      { a1, Ndim[metricName] }];
    ];
    Do[
      Do[
	Do[
	  Do[
	    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	       (* assignments here reflect symmetry *)
	       root[metricName,a1,a2,a4,a3] =
		-root[metricName,a1,a2,a3,a4];
	       root[metricName,a2,a1,a3,a4] =
		-root[metricName,a1,a2,a3,a4];
	       root[metricName,a2,a1,a4,a3] =
		root[metricName,a1,a2,a3,a4];
	       (* assignment *)
	       root[metricName,a1,a2,a3,a4] =
		calcFn[objectName,{a1,a2,a3,a4}];
	    ];
	    If[ simp ,
	       root[metricName,a1,a2,a3,a4] =
		  simpHow[ root[metricName,a1,a2,a3,a4] ];
	    ];
	    If[ callComp ,
	       grtF`component[ objectName,{a1,a2,a3,a4},
		root[metricName,a1,a2,a3,a4] ];
	    ],
	  { a4, a3+1, Ndim[metricName], 1 }],
	{ a3, Ndim[metricName]-1 }],
      { a2, a1+1, Ndim[metricName], 1 }],
    { a1, Ndim[metricName]-1 }];
    Return[NULL];
] (* return NULL *)

(**** four index - no symmetry ***)
grtG`symmetry[ grtF`symnosym4 ] = {{}};
grtF`addSymmetry[ grtF`symnosym4];

grtF`symnosym4 [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	Do[
	  Do[
	    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
	      root[metricName,a1,a2,a3,a4] =  calcFn[objectName,{a1,a2,a3,a4}];
	    ];
	    If[ simp ,
	       root[metricName,a1,a2,a3,a4] =
		    simpHow[  root[metricName,a1,a2,a3,a4] ];
	    ];
	    If[ callComp ,
	       grtF`component[ objectName,{a1,a2,a3,a4},
		 root[metricName,a1,a2,a3,a4] ];
	    ],
	  { a4, Ndim[metricName] }],
	{ a3, Ndim[metricName] }],
      { a2, Ndim[metricName] }],
    { a1, Ndim[metricName] }];
   Return[NULL];
] (* return NULL *)

grtG`symmetry[ grtF`symnosym5 ] = {{}};
grtF`addSymmetry[ grtF`symnosym5];

grtF`symnosym5 [objectName_, root_, calcFn_] :=
  Module[ {},
    Do[
      Do[
	Do[
	  Do[
	    Do[
    If[ calc && ValueQ[ ObjDef[ objectName][ grtC`calcFn] ] ,
            root[metricName,a1,a2,a3,a4,a5] = 
		calcFn[objectName,{a1,a2,a3,a4,a5}];
	    ];
	    If[ simp ,
	       root[metricName,a1,a2,a3,a4,a5] =
		    simpHow[  root[metricName,a1,a2,a3,a4,a5] ];
	    ];
	    If[ callComp ,
	       grtF`component[ objectName,{a1,a2,a3,a4,a5},
		 root[metricName,a1,a2,a3,a4,a5] ];
	    ],
	    { a5, Ndim[metricName] }],
	  { a4, Ndim[metricName] }],
	{ a3, Ndim[metricName] }],
      { a2, Ndim[metricName] }],
    { a1, Ndim[metricName] }];
   Return[NULL];
] (* return NULL *)



grtF`symLevC3[objectName_, root_, calcFn_] :=
(*
 * all other terms are just sign changes of  e
 *                                            1234
 *)
 Module[ {},
  If[ simp ,
      root[metricName,1,2,3] =
      simpHow[  root[metricName,1,2,3] ];
  ];
  If[ callComp ,
      grtF`component[ objectName,{1,2,3}, root[metricName,1,2,3] ];
  ];
 Return[NULL];
 ]

symmetry[grtF`symLevC4] = {{}};

grtF`symLevC4[objectName_, root_, calcFn_] :=
(*
 * all other terms are just sign changes of  e
 *                                            1234
 *)
 Module[ {},
  If[ simp ,
      root[metricName,1,2,3,4] =
      simpHow[  root[metricName,1,2,3,4] ];
  ];
  If[ callComp ,
      grtF`component[ objectName,{1,2,3,4}, root[metricName,1,2,3,4] ];
  ];
 Return[NULL];
 ]


End[]; (* close the global context *)

